/****************************************************************************
   Copyright & License:
   ====================
   
   Copyright 2009 - 2010 gAGE/UPC & ESA
   
   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at
   
   http://www.apache.org/licenses/LICENSE-2.0
   
   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 ***************************************************************************/

/****************************************************************************
 * Copyright: gAGE/UPC & ESA
 * Project: EDUNAV GNSS Lab Tool 
 * Author: Pere Ramos-Bosch (group of Astronomy and GEomatics - gAGE/UPC)
 * Maintainer: Adria Rovira-Garcia ( gAGE/UPC )
 *             glab.gage @ upc.edu     
 * File: dataHandling.c
 * Code Management Tool File Version: 2.0  Revision: 0 
 * Date: 2010/11/30
 ***************************************************************************/

/****************************************************************************
 * MODULE DESCRIPTION
 *
 * Name: dataHandling
 * Language: C
 *
 * Purpose:
 *  The purpose of this module is to centralize all the data structures of the 
 *  modules and allow an easy interface to them.
 *
 * Function:
 *  This class is in charge of declaring all the data structures used in all the
 *  modules. It defines all the requiered functions to access the information 
 *  inside the data structures, and a group of general functions.
 *
 * Dependencies:
 *  None
 *
 * Files modified:
 *  None
 *
 * Files read:
 *  None
 *
 * Resources usage:
 *  See Design Document
 ****************************************************************************/

/****************************
 *       RELEASE_HISTORY
 * -------------
 *          gLAB v1.3.1
 * Release: 2010/01/22
 * Change Log: First public version.
 * -------------
 *          gLAB v1.3.2
 * Release: 2010/02/23
 * Change Log: Corrected problem when processing with PCC (PC computed with C1).
 *             changeBase modified to be consistent with orientation matrix.
 *             Inertial/Non-inertial velocity switch included.
 *             minArcLength changed from 3 to 4.
 *             Included DCB data sources.
 *             Included SINEX, DCBs and GPS_Receiver_Type data files.
 *             Included strict radome option to enable/disable a relaxed matching
 *                  on ANTEX files.
 *             Corrected a bug when using PCC combination.
 *             When the receiver is of type 2 (no P1), the C1 will be provided instead
 *             Forward/Backward processing included.
 *             Corrected bug in t2doy function affecting the 31st of December and 
 *                  1st of January.
 * -------------
 *          gLAB v1.4.0
 * Release: 2010/06/21
 * Change Log: C1/P1 flexibility option is added to allow flexible handling of C1 and P1
 *             Included usable measurement index in TMeasOrder structure to select/deselect
 *                  individual measurements.
 *             Included prealignCP option to set/unset carrier-phase prealignment.
 *             Corrected a bug in t2doy function.
 * -------------
 *          gLAB v1.4.1
 * Release: 2010/07/09
 * Change Log: No changes in this file.
 * -------------
 *          gLAB v1.4.2
 * Release: 2010/07/31
 * Change Log: No changes in this file.
 * -------------
 *          gLAB v1.4.3
 * Release: 2010/08/31
 * Change Log: Maximum number of filter iterations added.
 *             Increased robustness to non sequencialy transmitted ephemerides present
 *                  in RINEX Navigation Message files.
 * -------------
 *          gLAB v1.4.4
 * Release: 2010/09/22
 * Change Log: Included new DCB management.
 * -------------
 *          gLAB v1.4.5
 * Release: 2010/10/31
 * Change Log: Tropospheric correction option is extended to select independently the nominal and mapping used.
 *             Dilution Of Precision extension: (GDOP, PDOP, TDOP, VDOP, HDOP) new variables.
 *             Default precise orbit interpolation degree changed from 9 to 10.
 *             SIGMA_INF splitted into different SIGMA_INF_DR and SIGMA_INF_DT: receiver coordinates and clock.
 * -----------
 *          gLAB v2.0.0
 * Release: 2010/11/30
 * Change Log: gLAB version released together with the Educational Book:
 *             "Global Navigation Satellite Systems: Fundamentals, Algorithms and Laboratory Exercises"
 *             Authors: J. Sanz, JM. Juan and M. Hernandez-Pajares.
 * -----------
 *       END_RELEASE_HISTORY
 *****************************/

/* External classes */
#include "dataHandling.h"

/**************************************
 * Declarations of internal operations
 **************************************/

/*****************************************************************************
 * Name        : initOptions
 * Description : Initialises to the defaults a TOptions structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TOptions  *options              O  N/A  TOptions structure
 *****************************************************************************/
void initOptions (TOptions *options) {
	int	i,j;
	
	// Verbose options
	options->printModel = 1;
	options->printPrefit = 1;
	options->printPostfit = 1;
	options->printFilterSolution = 1;
	options->printInfo = 1;
	options->printCycleslips = 1;
	options->printSatellites = 1;
	options->printInput = 1;
	options->printMeas = 0;
	options->printOutput = 1;
	options->printSatSel = 0;
	options->printSatDiff = 1;
	options->printSatStat = 1;
	options->printSatStatTot = 1;
	options->printSatPvt = 1;

	options->satVel = svITRF;

	// Processing and modeling options
	options->workMode = wmDOPROCESSING;
	options->receiverPositionSource = rpRINEX;
	options->strictradome = 0;
	options->antennaData = adNONE;
	options->ARPData = arpNONE;
	for (j=0;j<3;j++) {
		options->receiverARP[j] = 0;
		for (i=0;i<MAX_FREQUENCIES_PER_GNSS;i++) {
			options->receiverPCO[i][j] = 0;
		}
	}
	options->prealignCP = 1;
	options->flexibleC1P1 = 1;
	options->csLi = 1;
	options->csLimin = 0.034;
	options->csLimax = 0.08;
	options->csLit = 60;
	options->csBW = 1;
	options->csBWmin = 0.9;
	options->csBWmax = 18;
	options->csBWslope = 9;
	options->csL1C1 = 0;
	options->csL1C1slope = 5;
	options->csL1C1window = 100;
	options->csL1C1max = 15;
	options->timeTrans = 1;
	options->earthRotation = 1;
	options->satellitePhaseCenter = 1;
	options->relativisticCorrection = 1;
	options->windUpCorrection = 1;
	options->gravitationalCorrection = 1;
	options->solidTidesCorrection = 1;
	options->satelliteClockCorrection = 1;
	options->ionoModel = NoIono;
	options->troposphericCorrection = 1;
	options->tropNominal = UNB3Nominal;
	options->tropMapping = SimpleNominal;
        options->p1c1dcbModel = p1c1FLEXIBLE;
// Found: options->p1p2dcbModel = p1c1NONE;
	options->p1p2dcbModel = p1p2NONE;
	options->decimate = 300;
	options->orbitInterpolationDegree = 10;
	options->clockInterpolationDegree = 0;
	options->satelliteHealth = 1;
	options->discardEclipsedSatellites = 0;
        options->maxKalmanIterations = 20;

	// Filtering options
	for (i=0;i<MAX_FILTER_MEASUREMENTS;i++) {
		options->fixedMeasurementWeight[i] = 1;
		options->elevationMeasurementWeight[i][0] = 0;
		options->elevationMeasurementWeight[i][1] = 10*d2r;
	}
	options->totalFilterMeasurements = 2;
	options->measurement[0] = PC;
	options->measurement[1] = LC;
	options->smoothMeas[0] = NA;
	options->smoothMeas[1] = NA;
	options->smoothEpochs = 0;
	options->fixedMeasurementWeight[0] = 1;
	options->fixedMeasurementWeight[1] = 0.01;
	options->minArcLength = 4;
	options->maxdr2 = 10000;
	options->estimateTroposphere = 1;
	options->usePhase = 1;
	options->filterParams[PHI_PAR][DR_UNK] = 1;
	options->filterParams[PHI_PAR][DT_UNK] = 0;
	options->filterParams[PHI_PAR][TROP_UNK] = 1;
	options->filterParams[PHI_PAR][BIAS_UNK] = 1;
	options->filterParams[Q_PAR][DR_UNK] = 0;
	options->filterParams[Q_PAR][DT_UNK] = SIGMA_INF_DT*SIGMA_INF_DT;
	options->filterParams[Q_PAR][TROP_UNK] = 1e-4;
	options->filterParams[Q_PAR][BIAS_UNK] = 0;
	options->filterParams[P0_PAR][DR_UNK] = SIGMA_INF_DR*SIGMA_INF_DR;
	options->filterParams[P0_PAR][DT_UNK] = SIGMA_INF_DT*SIGMA_INF_DT;
	options->filterParams[P0_PAR][TROP_UNK] = 0.5*0.5;
	options->filterParams[P0_PAR][BIAS_UNK] = 20*20;
	options->elevationMask = 5*d2r;
	options->filterIterations = 1;
	
	
	for (i=0;i<MAX_GNSS;i++) {
		for (j=0;j<MAX_SATELLITES_PER_GNSS;j++) 
			options->includeSatellite[i][j] = 1;
		for (j=0;j<MAX_FREQUENCIES_PER_GNSS;j++)
			options->usableFreq[i][j] = 1;
	}
}

/*****************************************************************************
 * Name        : initSatellite
 * Description : Initialise a TSatellite struct
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  IO N/A  TEpoch struct
 * int  satIndex                   I  N/A  Pointing to the sat[satIndex] 
 *                                         TSatellite structure that should be 
 *                                         initialised
 *****************************************************************************/
void initSatellite (TEpoch *epoch, int satIndex) {
	int	i;

	epoch->sat[satIndex].arcLength = 1;
	epoch->sat[satIndex].hasBeenCycleslip = 1;
	
	for (i=0;i<MAX_FILTER_MEASUREMENTS;i++) {
		epoch->sat[satIndex].smoothedMeas[i] = 0;
	}

	for (i=0;i<epoch->measOrder[epoch->sat[satIndex].GNSS].nDiffMeasurements;i++) {
		epoch->sat[satIndex].phasePrealign[i] = 0;
		epoch->sat[satIndex].windUpRadAccumReceiver = 0;
		epoch->sat[satIndex].windUpRadAccumSatellite = 0;
	}
}

/*****************************************************************************
 * Name        : initEpoch
 * Description : Initialise a TEpoch struct
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  O  N/A  TEpoch struct
 *****************************************************************************/
void initEpoch(TEpoch *epoch) {
	int i,j;

	epoch->numSatellites = 0;
	epoch->prevNumSatellites = 0;
	epoch->numGNSS=0;
	epoch->receiver.recType = rtUNKNOWN;
	epoch->receiver.interval = -1;
	epoch->receiver.name[0] = '\0';
	epoch->receiver.type[0] = '\0';
	epoch->receiver.aproxPosition[0] = EARTH_RADIUS;
//	epoch->receiver.aproxPosition[0] = 0.01;
	epoch->receiver.aproxPosition[1] = 0;
	epoch->receiver.aproxPosition[2] = 0;
	epoch->receiver.aproxPositionError = 1e10;
	epoch->receiver.equivalentC1P1 = 0;
	epoch->prealign = 0;
	epoch->ZTD = 0;
	
	for (i=0;i<3;i++) {
		epoch->solidTideDisplacement[i] = 0;
		epoch->receiver.ARP[i] = 0;
		for (j=0;j<MAX_FREQUENCIES_PER_GNSS;j++) {
			epoch->receiver.antenna.PCO[j][i] = 0;
		}
	}

	for (i=0;i<MAX_SATELLITES_IN_VIEW;i++) {
		epoch->sat[i].arcLength = 0;
		for (j=0;j<MAX_MEASUREMENTS_PER_SATELLITE;j++) {
			epoch->sat[i].phasePrealign[j] = 0;
		}
	}

	for (j=0;j<MAX_GNSS;j++) {
		for (i=0;i<MAX_MEASUREMENTS;i++) {
			epoch->measOrder[j].meas2Ind[i] = -1;
			epoch->measOrder[j].conversionFactor[i] = 0;
			epoch->measOrder[j].usable[i] = 1;
		}
		for (i=0;i<MAX_SATELLITES_PER_GNSS;i++) {
			epoch->satInfo[j][i].lastEclipse.SoD = 0;
			epoch->satInfo[j][i].lastEclipse.MJDN = 0;
		}
	}
}

/*****************************************************************************
 * Name        : initGNSSproducts
 * Description : Initialise a TGNSSproducts struct
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TGNSSproducts*  products        O  N/A  TGNSSproducts struct
 *****************************************************************************/
void initGNSSproducts(TGNSSproducts *products) {
	products->BRDC = NULL;
	products->SP3 = NULL;
}

/*****************************************************************************
 * Name        : initBRDCproducts
 * Description : Initialise a TBRDCproducts struct
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TBRDCproducts  *BRDC            O  N/A  TBRDCproducts struct
 *****************************************************************************/
void initBRDCproducts(TBRDCproducts *BRDC) {
	int i,j;

	BRDC->numsats = 0;
	BRDC->block = NULL;
	BRDC->numblocks = NULL;

	for (i=0;i<3;i++) {
	  BRDC->ionA[i] = 0;
	  BRDC->ionB[i] = 0;
	}
	for (i=0;i<MAX_GNSS;i++) {
		for (j=0;j<MAX_SATELLITES_PER_GNSS;j++) {
			BRDC->index[i][j] = -1;
		}
	}
}

/*****************************************************************************
 * Name        : initSP3products
 * Description : Initialise a TSP3products struct
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TSP3products  *SP3              O  N/A  TSP3products struct
 *****************************************************************************/
void initSP3products(TSP3products *SP3) {
	int i,j;

	SP3->orbits.numSatellites = 0;
	SP3->orbits.block = NULL;
	SP3->orbits.accuracy = NULL;
	SP3->orbits.numRecords = 0;
	SP3->orbits.interval = 0;
	SP3->orbits.velocityAvailable = 0;
	SP3->orbits.startTime.MJDN = 0;
	SP3->orbits.startTime.SoD = 0;

	SP3->clocks.numSatellites = 0;
	SP3->clocks.block = NULL;
	SP3->clocks.numRecords = 0;
	SP3->clocks.vsizeblock = 0;
	SP3->clocks.interval = 0;
	SP3->clocks.startTime.MJDN = 0;
	SP3->clocks.startTime.SoD = 0;

	for (i=0;i<MAX_GNSS;i++) {
		for (j=0;j<MAX_SATELLITES_PER_GNSS;j++) {
			SP3->clocks.index[i][j] = -1;
			SP3->orbits.index[i][j] = -1;
		}
	}
}

/*****************************************************************************
 * Name        : initSP3orbitblock
 * Description : Initialise a TSP3orbitblock struct
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TSP3orbitblock  *SP3            O  N/A  TSP3orbitblock struct
 *****************************************************************************/
void initSP3orbitblock(TSP3orbitblock *SP3orbitblock) {
	SP3orbitblock->v[0] = 0;
	SP3orbitblock->v[1] = 0;
	SP3orbitblock->v[2] = 0;
	SP3orbitblock->x[0] = 0;
	SP3orbitblock->x[1] = 0;
	SP3orbitblock->x[2] = 0;
}

/*****************************************************************************
 * Name        : initSP3clockblock
 * Description : Initialise a TSP3clockblock struct
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TSP3clockblock  *SP3            O  N/A  TSP3clockblock struct
 *****************************************************************************/
void initSP3clockblock(TSP3clockblock *SP3clockblock) {
	SP3clockblock->clock = INVALID_CLOCK;
	SP3clockblock->clockrate = 0;
	SP3clockblock->clockdriftrate = 0;
}

/*****************************************************************************
 * Name        : initIONEX
 * Description : Initialise a TIONEX structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TIONEX  *IONEX                  O  N/A  TIONEX struct
 *****************************************************************************/
void initIONEX(TIONEX *IONEX) {
	int i,j;

	IONEX->numMaps = 0;
	IONEX->conversionFactor = 0.1; // Default conversion factor
	IONEX->exponent = -1;

	for (i=0;i<MAX_GNSS;i++) {
		for (j=0;j<MAX_SATELLITES_PER_GNSS;j++) {
			IONEX->DCB[i][j] = 0;
			IONEX->DCBRMS[i][j] = 1e6;
		}
	}
}

/*****************************************************************************
 * Name        : initIonoMap
 * Description : Initialise a TIonoMap structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TIonoMap  *ionoMap              O  N/A  TIonoMap struct
 *****************************************************************************/
void initIonoMap(TIonoMap *ionoMap) {
	ionoMap->dhgt = -1;
	ionoMap->dlat = -1;
	ionoMap->dlon = -1;

	ionoMap->RMS = NULL;
	ionoMap->TEC = NULL;
}

/*****************************************************************************
 * Name        : initFilterSolution
 * Description : Initialise a TFilterSolution structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TFilterSolution  *solution      O  N/A  TFilterSolution struct
 *****************************************************************************/
void initFilterSolution (TFilterSolution *solution) {
	int	i;

	solution->prevNumSatellites = 0;

	for (i=0;i<MAX_UNK;i++) {
		solution->parameters[i] = 0;
	}

	for (i=0;i<MAX_VECTOR_UNK;i++) {
		solution->correlations[i] = 0;
	}
}

/*****************************************************************************
 * Name        : initConstellation
 * Description : Initialise a TConstellation structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TConstellation  *constellation  O  N/A  TConstellation struct
 *****************************************************************************/
void initConstellation(TConstellation *constellation) {
	constellation->numSatellites = 0;
	constellation->sat = NULL;
}

/*****************************************************************************
 * Name        : initConstellationElement
 * Description : Initialise a TConstellationElement structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TConstellationElement  *sat     O  N/A  TConstellationElement struct
 *****************************************************************************/
void initConstellationElement(TConstellationElement *sat) {
	int		i;
	
	sat->GNSS = GPS;
	sat->block = UNKNOWN_BLOCK;
	for (i=0;i<MAX_FREQUENCIES_PER_GNSS;i++)
		sat->PCO[i][0] = sat->PCO[i][1] = sat->PCO[i][2] = 0;
	sat->PRN = 0;
	sat->SVN = 0;
	sat->tDecommissioned.MJDN = 62502;  // 1st January 2030
	sat->tDecommissioned.SoD = 0;
	sat->tLaunch.MJDN = 0;
	sat->tLaunch.SoD = 0;
}

/*****************************************************************************
 * Name        : initAntennaList
 * Description : Initialise a TAntennaList structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TAntennaList  *antennaList      O  N/A  TAntennaList struct
 *****************************************************************************/
void initAntennaList(TAntennaList *antennaList) {
	antennaList->numAnt = 0;
	antennaList->ant = NULL;
}

/*****************************************************************************
 * Name        : initAntenna
 * Description : Initialise a TAntenna structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TAntenna  *antenna              O  N/A  TAntenna struct
 *****************************************************************************/
void initAntenna(TAntenna *antenna) {
	int		i;
	
	for (i=0;i<MAX_FREQUENCIES_PER_GNSS;i++) 
		antenna->PCO[i][0] = antenna->PCO[i][1] = antenna->PCO[i][2] = 0;
	antenna->type[0] = '\0';
	antenna->type_ant[0] = '\0';
	antenna->type_radome[0] = '\0';
}

/*****************************************************************************
 * Name        : initStat
 * Description : Initialise a TStat structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TStat  *stat                    O  N/A  TStat struct
 *****************************************************************************/
void initStat (TStat *stat) {
	stat->n = 0;
	stat->mean = 0;
	stat->mean2 = 0;
}

/*****************************************************************************
 * Name        : initTGD
 * Description : Initialise a TGDData structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TTGDdata  *tgdData              O  N/A  TTGDdata struct
 * enum P1C1DCBModel  p1c1dcbModel I  N/A  P1C1 DCB source (p1c1UNKNOWN if still unavailable)
 * enum P1P2DCBModel  p1p2dcbModel I  N/A  P1P2 DCB source (p1p2UNKNOWN if still unavailable)
 *****************************************************************************/
void initTGD (TTGDdata *tgdData, enum P1C1DCBModel p1c1dcbModel, enum P1P2DCBModel p1p2dcbModel) {
	tgdData->BRDC = NULL;
	tgdData->DCB = NULL;
	tgdData->p1c1dcbModel = p1c1dcbModel;
	tgdData->p1p2dcbModel = p1p2dcbModel;

//	if (tgdData->p1c1dcbModel==p1c1STRICT ||tgdData->p1p2dcbModel==p1p2DCB) {
		tgdData->DCB = malloc(sizeof(TDCBdata));
		initDCB(tgdData->DCB);
//	}
}

/*****************************************************************************
 * Name        : initDCB
 * Description : Initialise a TDCBdata structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TDCBdata  *dcbData              O  N/A  TDCBdata struct
 *****************************************************************************/
void initDCB (TDCBdata *dcbData) {
	int i,j,k;
	
	dcbData-> n = 0;
	
	for (i=0;i<MAX_DCB_TYPES;i++) {
		dcbData->block[i].startTime.MJDN = 0;
		dcbData->block[i].startTime.SoD = 0;
		dcbData->block[i].endTime.MJDN = 0;
		dcbData->block[i].endTime.SoD = 0;
		dcbData->block[i].measFrom = NA;
		dcbData->block[i].measTo = NA;
		for (j=0;j<MAX_GNSS;j++) {
			for (k=0;k<MAX_SATELLITES_PER_GNSS;k++) {
				dcbData->block[i].dcb[j][k] = 0;
				dcbData->block[i].available[j][k] = 0;
			}
		}
	}
}

/*****************************************************************************
 * Name        : initReceiverList
 * Description : Initialise a TReceiverList structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TReceiverList  *recList         O  N/A  TReceiverList struct
 *****************************************************************************/
void initReceiverList (TReceiverList *recList) {
	recList->n = 0;
	recList->rec = NULL;
}

/*****************************************************************************
 * Name        : initStationList
 * Description : Initialise a TStationList structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TStationList  *stationList      O  N/A  TStationList struct
 *****************************************************************************/
void initStationList (TStationList *stationList) {
	stationList->n = 0;
	stationList->station = NULL;
}

/*****************************************************************************
 * Name        : changeBase
 * Description : Changes a 3D vector to the specified base
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double  *base                   I  N/A  Base (vectors of the base in the rows)
 * double  *vector		           IO N/A  Vector to be changed (output 
 *                                        overwrites this vector)
 *****************************************************************************/
void changeBase (double base[3][3], double vector[3]) {
	double	res[3];
	int		i,j;

	for (i=0;i<3;i++) {
		res[i] = 0;
		for (j=0;j<3;j++) {
			res[i] += base[i][j] * vector[j];
		}
	}

	memcpy (vector,res,sizeof(double)*3);
}

/*****************************************************************************
 * Name        : min
 * Description : Return the minimum between two numbers
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double  a                       I  N/A  First number
 * double  b		               I  N/A  Second number 
 *                                        overwrites this vector)
 * Returned value (double)         O  N/A  Minimum between a and b
 *****************************************************************************/
double min (double a, double b) {
	if (a<b) return a;
	else return b;
}

/*****************************************************************************
 * Name        : max
 * Description : Return the maximum between two numbers
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double  a                       I  N/A  First number
 * double  b		               I  N/A  Second number 
 *                                        overwrites this vector)
 * Returned value (double)         O  N/A  Maximum between a and b
 *****************************************************************************/
double max (double a, double b) {
	if (a<b) return b;
	else return a;
}

/*****************************************************************************
 * Name        : modulo
 * Description : Does the modulo mod of a number
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double  a                       I  N/A  Number
 * double  mod		               I  N/A  Modulo base
 * Returned value (double)         O  N/A  Modulo
 *****************************************************************************/
double modulo (double a, double mod) {
	return a - ((int)(a/mod))*mod;
}

/*****************************************************************************
 * Name        : unitary
 * Description : Makes a vector unitary
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double  *vect                   IO N/A  3 dimension vector
 *****************************************************************************/
void unitary(double *vect) {
	double	module;

	module = sqrt(vect[0]*vect[0] + vect[1]*vect[1] + vect[2]*vect[2]);

	vect[0] /= module;
	vect[1] /= module;
	vect[2] /= module;
}

/*****************************************************************************
 * Name        : vectProd
 * Description : Makes the vectorial product between two vectors
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double  *a                      I  N/A  First vector
 * double  *b                      I  N/A  Second vector
 * double  *res                    O  N/A  Result vector [res = a x b]
 *****************************************************************************/
void vectProd(double *a, double *b, double *res) {
	res[0] = a[1]*b[2] - a[2]*b[1];
	res[1] = a[2]*b[0] - a[0]*b[2];
	res[2] = a[0]*b[1] - a[1]*b[0];
}

/*****************************************************************************
 * Name        : scalarProd
 * Description : Makes the scalar product between two vectors
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double  *a                      I  N/A  First vector
 * double  *b                      I  N/A  Second vector
 * Returned value (double)         O  N/A  Result vector [res = a · b]
 *****************************************************************************/
double scalarProd(double *a, double *b) {
	return a[0]*b[0] + a[1]*b[1] + a[2]*b[2];
}

/*****************************************************************************
 * Name        : trim
 * Description : Trim a string (remove preceding and succeding spaces)
 * Parameters  :
 * Name                           |Da|Unit|Description
 * char  *line                     IO N/A  String to trim
 * Returned value (char*)          O  N/A  Pointer to *line
 *****************************************************************************/
char *trim(char *line) {
	int i;
	int pos=0;
	int lastnotspace=0;
	int state=0;
		// 0 -> Trimming first spaces
		// 1 -> Waitting to end string

	for (i=0;i<strlen(line);i++){
		if (state==0) {
			if (line[i]!=' ') {
				state = 1;
				line[pos] = line[i];
				pos++;
				lastnotspace = pos;
			}
		} else if (state==1) {
			line[pos] = line[i];
			pos++;
			if (line[i]!=' ' && line[i]!='\n') {
				lastnotspace = pos;
			}
		}
	}
	line[lastnotspace]='\0';
	return &line[0];
}

/*****************************************************************************
 * Name        : getstr
 * Description : Get a substring from a string and trim it
 * Parameters  :
 * Name                           |Da|Unit|Description
 * char  *line                     I  N/A  Input string
 * char  *out                      O  N/A  Substring
 * int  ini                        I  N/A  Initial character position
 * int  length                     I  N/A  Length of the substring
 * Returned value (char*)          O  N/A  Pointer to *out
 *****************************************************************************/
char *getstr(char *out,char *line,int ini, int length) {
	int end;

	end = strlen(line);
	if (ini>=end) {
		out[0]='\0';
		return &out[0];
	}
	strncpy(out,&line[ini],length);
	out[length]='\0';
	trim(out);
	return &out[0];
}

/*****************************************************************************
 * Name        : whatIs
 * Description : Check whether an observation is carrier phase, pseudorange, 
 *               SNR or doppler
 * Parameters  :
 * Name                            |Da|Unit|Description
 * MeasurementType  meas            I  N/A  Measurement type (C1P, P1P...)
 * Returned value (MeasurementKind) O  N/A  Measurement kind (Pseudorange, CarrierPhase...)
 *****************************************************************************/
enum MeasurementKind whatIs(enum MeasurementType meas) {

	if (meas == NA) return UNKNOWN_MEAS;
	else if (meas>=ENDMEAS) {
		// Combination
		if (meas == LW || meas == LI || meas == LC || meas == G1C || meas == G1P || meas == G2C || meas == G2P) return CarrierPhase;
		else if (meas == PW || meas == PI || meas == PC || meas == PCC) return Pseudorange;
		else return UNKNOWN_MEAS;
	} else if (((meas-1)%4)==0) return Pseudorange;
	else if (((meas-2)%4)==0) return CarrierPhase;
	else if (((meas-3)%4)==0) return Doppler;
	else if (((meas)%4)==0) return SNR;
	else return UNKNOWN_MEAS;
}

/*****************************************************************************
 * Name        : getLambda
 * Description : Get the lambda corresponding to the GNSS and measurement
 * Parameters  :
 * Name                           |Da|Unit|Description
 * GNSSystem  GNSS                 I  N/A  GNSS system
 * MeasurementType  meas           I  N/A  Measurement type
 * Returned value (double)         O  m    Lambda on the measurement
 *****************************************************************************/
double getLambda(enum GNSSystem GNSS, enum MeasurementType meas) {
	if (GNSS == GPS) {
		if (meas <= S1N) return GPSl1;
		else if (meas <= S2N) return GPSl2;
		else if (meas <= S5X) return GPSl5;
	} else if (GNSS == Galileo) {
		if (meas <= S1N) return GALl1;
		else if (meas > S2N && meas <= S5X) return GALl5a;
		else if (meas <= S7X) return GALl5b;
		else if (meas <= S8X) return GALl5;
		else if (meas <= S6Z) return GALl6;
	} else if (GNSS == GLONASS) {
		return 1;
	} else if (GNSS == GEO) {
		if (meas <= S1N) return SBASl1;
		else if (meas > S2N && meas <= S5X) return SBASl5;
	}

	return 1;
}

/*****************************************************************************
 * Name        : getFrequency
 * Description : Get the frequency corresponding to the GNSS and measurement
 * Parameters  :
 * Name                           |Da|Unit|Description
 * GNSSystem  GNSS                 I  N/A  GNSS system
 * MeasurementType  meas           I  N/A  Measurement type
 * Returned value (double)         O  Hz   Frequency of the measurement
 *****************************************************************************/
double getFrequency(enum GNSSystem GNSS, enum MeasurementType meas) {
	return c/getLambda (GNSS,meas);
}

/*****************************************************************************
 * Name        : getFrequencyInt
 * Description : Get the integer identifier value for the frequency 
 *               corresponding to the measurement
 * Parameters  :
 * Name                           |Da|Unit|Description
 * MeasurementType  meas           I  N/A  Measurement type
 * Returned value (int)            O  N/A  Frequency identifier of the measurement:
 *							GPS:		'G01' - 1 - L1
 *										'G02' - 2 - L2 
 *										'G05' - 5 - L5
 *							GLONASS:	'R01' - 1 - G1
 *										'R02' - 2 - G2
 *							Galileo:	'E01' - 1 - E1
 *										'E05' - 5 - E5a
 *										'E07' - 7 - E5b
 *										'E08' - 8 - E5 (E5a-E5b)
 *										'E06' - 6 - E56   
 *****************************************************************************/
int getFrequencyInt(enum MeasurementType meas) {
	if (meas <= S1N) return 1;
	else if (meas <= S2N) return 2;
	else if (meas <= S5X) return 5;
	else if (meas <= S7X) return 7;
	else if (meas <= S8X) return 8;
	else if (meas <= S6Z) return 6;
	else return 0;
}

/*****************************************************************************
 * Name        : measstr2meastype
 * Description : Transform from string to internal Measurement type
 * Parameters  :
 * Name                            |Da|Unit|Description
 * char  *str                       I  N/A  Measurement string
 * Returned value (MeasurementType) O  N/A  Measurement type
 *****************************************************************************/
enum MeasurementType measstr2meastype(char *str) {
	// Possible options:
	//    1st digit: C,L,D,S,P,G,B   [7]
	//    2nd digit: 1,2,5,6,7,8,W,I,C    [9]
	//    3rd digit: <empty>,A,B,C,D,I,L,M,N,P,Q,S,W,X,Y,Z   [16]

	static int	initialized = 0;
	static enum MeasurementType	s2t[7][9][16];
	int 		a1,a2,a3;

	if (!initialized) {
		for (a1=0;a1<7;a1++) {
			for (a2=0;a2<9;a2++) {
				for (a3=0;a3<16;a3++) {
					s2t[a1][a2][a3] = NA;
				}
			}
		}

		// Combinations
		s2t[1][6][0] = LW;
		s2t[4][6][0] = PW;
		s2t[6][6][0] = BW;
		s2t[1][7][0] = LI;
		s2t[4][7][0] = PI;
		s2t[1][8][0] = LC;
		s2t[4][8][0] = PC;
		s2t[4][8][3] = PCC;
		s2t[5][0][3] = G1C;
		s2t[5][0][9] = G1P;
		s2t[5][1][3] = G2C;
		s2t[5][1][9] = G2P;
		
		////// Automatically generated code block /////
		// RINEX 3.00
		s2t[0][0][1] = C1A; s2t[1][0][1] = L1A; s2t[2][0][1] = D1A; s2t[3][0][1] = S1A; s2t[0][0][2] = C1B; s2t[1][0][2] = L1B; s2t[2][0][2] = D1B; s2t[3][0][2] = S1B; s2t[0][0][3] = C1C; s2t[1][0][3] = L1C; s2t[2][0][3] = D1C; s2t[3][0][3] = S1C; s2t[0][0][9] = C1P; s2t[1][0][9] = L1P; s2t[2][0][9] = D1P; s2t[3][0][9] = S1P; s2t[0][0][12] = C1W; s2t[1][0][12] = L1W; s2t[2][0][12] = D1W; s2t[3][0][12] = S1W; s2t[0][0][13] = C1X; s2t[1][0][13] = L1X; s2t[2][0][13] = D1X; s2t[3][0][13] = S1X; s2t[0][0][14] = C1Y; s2t[1][0][14] = L1Y; s2t[2][0][14] = D1Y; s2t[3][0][14] = S1Y; s2t[0][0][15] = C1Z; s2t[1][0][15] = L1Z; s2t[2][0][15] = D1Z; s2t[3][0][15] = S1Z; s2t[0][0][7] = C1M; s2t[1][0][7] = L1M; s2t[2][0][7] = D1M; s2t[3][0][7] = S1M; s2t[1][0][8] = L1N; s2t[2][0][8] = D1N; s2t[3][0][8] = S1N;
		s2t[0][1][3] = C2C; s2t[1][1][3] = L2C; s2t[2][1][3] = D2C; s2t[3][1][3] = S2C; s2t[0][1][4] = C2D; s2t[1][1][4] = L2D; s2t[2][1][4] = D2D; s2t[3][1][4] = S2D; s2t[0][1][11] = C2S; s2t[1][1][11] = L2S; s2t[2][1][11] = D2S; s2t[3][1][11] = S2S; s2t[0][1][6] = C2L; s2t[1][1][6] = L2L; s2t[2][1][6] = D2L; s2t[3][1][6] = S2L; s2t[0][1][13] = C2X; s2t[1][1][13] = L2X; s2t[2][1][13] = D2X; s2t[3][1][13] = S2X; s2t[0][1][9] = C2P; s2t[1][1][9] = L2P; s2t[2][1][9] = D2P; s2t[3][1][9] = S2P; s2t[0][1][12] = C2W; s2t[1][1][12] = L2W; s2t[2][1][12] = D2W; s2t[3][1][12] = S2W; s2t[0][1][14] = C2Y; s2t[1][1][14] = L2Y; s2t[2][1][14] = D2Y; s2t[3][1][14] = S2Y; s2t[0][1][7] = C2M; s2t[1][1][7] = L2M; s2t[2][1][7] = D2M; s2t[3][1][7] = S2M; s2t[1][1][8] = L2N; s2t[2][1][8] = D2N; s2t[3][1][8] = S2N;
		s2t[0][2][5] = C5I; s2t[1][2][5] = L5I; s2t[2][2][5] = D5I; s2t[3][2][5] = S5I; s2t[0][2][10] = C5Q; s2t[1][2][10] = L5Q; s2t[2][2][10] = D5Q; s2t[3][2][10] = S5Q; s2t[0][2][13] = C5X; s2t[1][2][13] = L5X; s2t[2][2][13] = D5X; s2t[3][2][13] = S5X;
		s2t[0][4][5] = C7I; s2t[1][4][5] = L7I; s2t[2][4][5] = D7I; s2t[3][4][5] = S7I; s2t[0][4][10] = C7Q; s2t[1][4][10] = L7Q; s2t[2][4][10] = D7Q; s2t[3][4][10] = S7Q; s2t[0][4][13] = C7X; s2t[1][4][13] = L7X; s2t[2][4][13] = D7X; s2t[3][4][13] = S7X;
		s2t[0][5][5] = C8I; s2t[1][5][5] = L8I; s2t[2][5][5] = D8I; s2t[3][5][5] = S8I; s2t[0][5][10] = C8Q; s2t[1][5][10] = L8Q; s2t[2][5][10] = D8Q; s2t[3][5][10] = S8Q; s2t[0][5][13] = C8X; s2t[1][5][13] = L8X; s2t[2][5][13] = D8X; s2t[3][5][13] = S8X;
		s2t[0][3][1] = C6A; s2t[1][3][1] = L6A; s2t[2][3][1] = D6A; s2t[3][3][1] = S6A; s2t[0][3][2] = C6B; s2t[1][3][2] = L6B; s2t[2][3][2] = D6B; s2t[3][3][2] = S6B; s2t[0][3][3] = C6C; s2t[1][3][3] = L6C; s2t[2][3][3] = D6C; s2t[3][3][3] = S6C; s2t[0][3][13] = C6X; s2t[1][3][13] = L6X; s2t[2][3][13] = D6X; s2t[3][3][13] = S6X; s2t[0][3][15] = C6Z; s2t[1][3][15] = L6Z; s2t[2][3][15] = D6Z; s2t[3][3][15] = S6Z;
		////// End of automatically generated code block /////

		// RINEX 2.11
		s2t[0][0][0] = C1C; // C1
		s2t[4][0][0] = C1P; // P1
		s2t[1][0][0] = L1P; // L1
		s2t[2][0][0] = D1P; // D1
		s2t[3][0][0] = S1P; // S1
		
		s2t[0][1][0] = C2C;	// C2
		s2t[4][1][0] = C2P; // P2
		s2t[1][1][0] = L2P;	// L2
		s2t[2][1][0] = D2P;	// D2
		s2t[3][1][0] = S2P; // S2
		
		s2t[0][2][0] = C5X; // C5
		s2t[1][2][0] = L5X; // L5
		s2t[2][2][0] = D5X; // D5
		s2t[3][2][0] = S5X; // S5
		
		s2t[0][3][0] = C6X; // C6
		s2t[1][3][0] = L6X; // L6
		s2t[2][3][0] = D6X; // D6
		s2t[3][3][0] = S6X; // S6
		
		s2t[0][4][0] = C7X; // C7
		s2t[1][4][0] = L7X; // L7
		s2t[2][4][0] = D7X; // D7
		s2t[3][4][0] = S7X; // S7
		
		s2t[0][5][0] = C8X; // C8
		s2t[1][5][0] = L8X; // L8
		s2t[2][5][0] = D8X; // D8
		s2t[3][5][0] = S8X; // S8

		initialized = 1;
	}

	if (str[0]=='C') a1=0;
	else if (str[0]=='L') a1=1;
	else if (str[0]=='D') a1=2;
	else if (str[0]=='S') a1=3;
	else if (str[0]=='P') a1=4;
	else if (str[0]=='G') a1=5;
	else if (str[0]=='B') a1=6;
	else return NA;

	if (str[1]=='1') a2=0;
	else if (str[1]=='2') a2=1;
	else if (str[1]=='5') a2=2;
	else if (str[1]=='6') a2=3;
	else if (str[1]=='7') a2=4;
	else if (str[1]=='8') a2=5;
	else if (str[1]=='W') a2=6;
	else if (str[1]=='I') a2=7;
	else if (str[1]=='C') a2=8;
	else return NA;

	if (str[2]=='\0') a3=0;
	else if (str[2]=='A') a3=1;
	else if (str[2]=='B') a3=2;
	else if (str[2]=='C') a3=3;
	else if (str[2]=='D') a3=4;
	else if (str[2]=='I') a3=5;
	else if (str[2]=='L') a3=6;
	else if (str[2]=='M') a3=7;
	else if (str[2]=='N') a3=8;
	else if (str[2]=='P') a3=9;
	else if (str[2]=='Q') a3=10;
	else if (str[2]=='S') a3=11;
	else if (str[2]=='W') a3=12;
	else if (str[2]=='X') a3=13;
	else if (str[2]=='Y') a3=14;
	else if (str[2]=='Z') a3=15;
	else return NA;
	
	return s2t[a1][a2][a3];
}

/*****************************************************************************
 * Name        : meastype2measstr
 * Description : Transform from internal Measurement type to string
 * Parameters  :
 * Name                           |Da|Unit|Description
 * MeasurementType  meas           I  N/A  Measurement type
 * Returned value (char*)          O  N/A  Measurement string
 *****************************************************************************/
char *meastype2measstr(enum MeasurementType meas) {
	static char str[10];
	static int	initialized = 0;
	static char	l[34];
	char		type;
	int			aux;
	
	strcpy(str,"NAN");

	if (!initialized) {
		initialized = 1;
		l[0]='A'; l[1]='B'; l[2]='C'; l[3]='P'; l[4]='W'; l[5]='X'; l[6]='Y'; l[7]='Z'; l[8]='M'; l[9]='N';
		l[10]='C'; l[11]='D'; l[12]='S'; l[13]='L'; l[14]='X'; l[15]='P'; l[16]='W'; l[17]='Y'; l[18]='M'; l[19]='N';
		l[20]='I'; l[21]='Q'; l[22]='X'; l[23]='I'; l[24]='Q'; l[25]='X'; l[26]='I'; l[27]='Q'; l[28]='X'; l[29]='A';
		l[30]='B'; l[31]='C'; l[32]='X'; l[33]='Z';
	}

	if (meas<ENDMEAS) {
		switch (whatIs(meas)) {
			case Pseudorange:
				type='C';
				break;
			case CarrierPhase:
				type='L';
				break;
			case Doppler:
				type='D';
				break;
			case SNR:
				type='S';
				break;
			default:
				return str;
				break;
		}
		aux = getFrequencyInt(meas);
		sprintf(str,"%1c%1d%1c",type,aux,l[(int)((meas-1)/4)]);
	} else {
		switch (meas) {
			case 138:
				strcpy(str,"LW");
				break;
			case 139:
				strcpy(str,"BW");
				break;
			case 140:
				strcpy(str,"PW");
				break;
			case 141:
				strcpy(str,"LI");
				break;
			case 142:
				strcpy(str,"PI");
				break;
			case 143:
				strcpy(str,"LC");
				break;
			case 144:
				strcpy(str,"PC");
				break;
			case 145:
				strcpy(str,"PCC");
				break;
			case 146:
				strcpy(str,"G1C");
				break;
			case 147:
				strcpy(str,"G1P");
				break;
			case 148:
				strcpy(str,"G2C");
				break;
			case 149:
				strcpy(str,"G2P");
				break;
			default:
				return str;
				break;
		}
	}

	return str;
}

/*****************************************************************************
 * Name        : gnsschar2gnsstype
 * Description : Transform from string to internal GNSS System type
 * Parameters  :
 * Name                           |Da|Unit|Description
 * char  *system                   I  N/A  GNSS System string
 * Returned value (GNSSystem)      O  N/A  GNSSystem type
 *****************************************************************************/
enum GNSSystem gnsschar2gnsstype(char system) {
	if (system=='G' || system==' ') return GPS;
	else if (system=='E')return Galileo;
	else if (system=='R') return GLONASS;
	else if (system=='S') return GEO;

	return GPS; // Default value
}

/*****************************************************************************
 * Name        : gnsstype2gnssstr
 * Description : Transform from internal GNSS System type to string
 * Parameters  :
 * Name                           |Da|Unit|Description
 * GNSSystem  GNSS                 I  N/A  GNSSystem type
 * Returned value (char*)          O  N/A  GNSS System string
 *****************************************************************************/
char *gnsstype2gnssstr(enum GNSSystem GNSS) {
	static char str[10];

	strcpy(str,"NAN");

	switch (GNSS) {
		case GPS:
			strcpy(str,"GPS");
			break;
		case Galileo:
			strcpy(str,"GAL");
			break;
		case GLONASS:
			strcpy(str,"GLO");
			break;
		case GEO:
			strcpy(str,"GEO");
			break;
		default:
			strcpy(str,"NAN");
	}

	return str;
}

/*****************************************************************************
 * Name        : gnsstype2char
 * Description : Transform from internal GNSS System type to char
 * Parameters  :
 * Name                           |Da|Unit|Description
 * GNSSystem  GNSS                 I  N/A  GNSSystem type
 * Returned value (char)           O  N/A  GNSS System char
 *****************************************************************************/
char gnsstype2char(enum GNSSystem GNSS) {
	
	switch (GNSS) {
		case GPS:
			return 'G';
			break;
		case Galileo:
			return 'E';
			break;
		case GLONASS:
			return 'R';
			break;
		case GEO:
			return 'S';
			break;
		default:
			return 'N';
	}

	return 'N';
}

/*****************************************************************************
 * Name        : satblockstr2satblock
 * Description : Transform from a satellite block string to the internal 
 *               satellite block
 * Parameters  :
 * Name                           |Da|Unit|Description
 * char  *str                      I  N/A  Satellite block string
 * Returned value (SatelliteBlock) O  N/A  Internal satellite block
 *****************************************************************************/
enum SatelliteBlock satblockstr2satblock (char *str) {
	if (strcmp(str,"BLOCK I")==0) {
		return GPS_BLOCK_I;
	} else if (strcmp(str,"BLOCK II")==0) {
		return GPS_BLOCK_II;
	} else if (strcmp(str,"BLOCK IIA")==0) {
		return GPS_BLOCK_IIA;
	} else if (strcmp(str,"BLOCK IIR")==0) {
		return GPS_BLOCK_IIR;
	} else if (strcmp(str,"BLOCK IIR-A")==0) {
		return GPS_BLOCK_IIR;
	} else if (strcmp(str,"BLOCK IIR-B")==0) {
		return GPS_BLOCK_IIR;
	} else if (strcmp(str,"BLOCK IIR-M")==0) {
		return GPS_BLOCK_IIRM;
	} else if (strcmp(str,"GLONASS")==0) {
		return GLONASS_BLOCK;
	} else if (strcmp(str,"GLONASS-M")==0) {
		return GLONASS_BLOCK_M;
	}
	
	return UNKNOWN_BLOCK;
}

/*****************************************************************************
 * Name        : satblockstr2satblock
 * Description : Transform from an internal satellite block to a satellite  
 *               block string
 * Parameters  :
 * Name                           |Da|Unit|Description
 * SatelliteBlock  satBlock        I  N/A  Internal satellite block
 * Returned value (char *)         O  N/A  Satellite block string
 *****************************************************************************/
char *satblock2satblockstr (enum SatelliteBlock satBlock) {
	static char str[20];

	strcpy(str,"NAN");

	switch (satBlock) {
		case GPS_BLOCK_I:
			strcpy(str,"BLOCK I");
			break;
		case GPS_BLOCK_II:
			strcpy(str,"BLOCK II");
			break;
		case GPS_BLOCK_IIA:
			strcpy(str,"BLOCK IIA");
			break;
		case GPS_BLOCK_IIR:
			strcpy(str,"BLOCK IIR");
			break;
		case GPS_BLOCK_IIRM:
			strcpy(str,"BLOCK IIR-M");
			break;
		case GLONASS_BLOCK:
			strcpy(str,"GLONASS");
			break;
		case GLONASS_BLOCK_M:
			strcpy(str,"GLONASS-M");
			break;
		default:
			strcpy(str,"NAN");
	}

	return str;
}

/*****************************************************************************
 * Name        : getMeasModelValue
 * Description : Get the measurement and model value of a particular satellite
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  TEpoch structure
 * GNSSystem  system               I  N/A  GNSS system of the satellite
 * int  PRN                        I  N/A  PRN identifier of the satellite
 * MeasurementType  measType       I  N/A  Measurement type
 * double  *measurement            O  m    Measurement
 * double  *model                  O  m    Model
 * Returned value (int)            O  N/A  Status of the function
 *                                         1 => Measurement found
 *                                         0 => Not found
 *                                        -1 => Found, but not usable
 *****************************************************************************/
int getMeasModelValue(TEpoch *epoch, enum GNSSystem system, int PRN, enum MeasurementType measType, double *measurement, double *model) {
	int		ind;
	int		i;
	double	rawValue;
	double	conversionFactor;
	double	align;
	double	meas1,meas2;
	double	model1,model2;
	int		res1,res2;
	int		usable = 1;

	// Check if measurement is usable
	if (!epoch->measOrder[system].usable[measType]) usable = 0;

	if (measType<ENDMEAS) { // It is a measurement and NOT a combination
		// If it is a receiver type, that does not has P1, use C1 instead
		if (epoch->receiver.recType==rtNOP1 && measType==C1P) { 
			measType = C1C;
		}
		
		// Obtain the stored index of the measurement (-1 if not found)
		ind = epoch->measOrder[system].meas2Ind[measType];
		
		// If the measurement is not found, and it can make equivalences between C1 and P1
		if (ind==-1 && epoch->receiver.equivalentC1P1) { 
			if (measType==C1C) {
				measType=C1P;
				ind = epoch->measOrder[system].meas2Ind[measType];
			} else if (measType==C1P) {
				measType=C1C;
				ind = epoch->measOrder[system].meas2Ind[measType];
			} 
		}
		if (ind==-1) return 0; // Measurement not found
	} else if (measType==PW) {
		res1 = getMeasModelValue(epoch,system,PRN,C1P,&meas1,&model1);
		res2 = getMeasModelValue(epoch,system,PRN,C2P,&meas2,&model2);
		if (res1 && res2) {
			if (measurement!=NULL) *measurement = (GPSmf1*meas1 + GPSmf2*meas2) / (GPSmf1 + GPSmf2);
			if (model!=NULL) *model =  (GPSmf1*model1 + GPSmf2*model2) / (GPSmf1 + GPSmf2);
			if (usable) return 1;
			else return -1;
		} else {
			return 0;
		}
	} else if (measType==LW) {
		res1 = getMeasModelValue(epoch,system,PRN,L1P,&meas1,&model1);
		res2 = getMeasModelValue(epoch,system,PRN,L2P,&meas2,&model2);
		if (res1 && res2) {
			if (measurement!=NULL) *measurement = (GPSmf1*meas1 - GPSmf2*meas2) / (GPSmf1 - GPSmf2);
			if (model!=NULL) *model =  (GPSmf1*model1 - GPSmf2*model2) / (GPSmf1 - GPSmf2);
			if (usable) return 1;
			else return -1;
		} else {
			return 0;
		}
	} else if (measType==BW) {
		res1 = getMeasModelValue(epoch,system,PRN,LW,&meas1,&model1);
		res2 = getMeasModelValue(epoch,system,PRN,PW,&meas2,&model2);
		if (res1 && res2) {
			if (measurement!=NULL) *measurement = meas1 - meas2;
			if (model!=NULL) *model =  model1 - model2;
			if (usable) return 1;
			else return -1;
		} else {
			return 0;
		}
	} else if (measType==PI) {
		res1 = getMeasModelValue(epoch,system,PRN,C2P,&meas1,&model1);
		res2 = getMeasModelValue(epoch,system,PRN,C1P,&meas2,&model2);
		if (res1 && res2) {
			if (measurement!=NULL) *measurement = meas1 - meas2;
			if (model!=NULL) *model =  model1 - model2;
			if (usable) return 1;
			else return -1;
		} else {
			return 0;
		}
	} else if (measType==LI) {
		res1 = getMeasModelValue(epoch,system,PRN,L1P,&meas1,&model1);
		res2 = getMeasModelValue(epoch,system,PRN,L2P,&meas2,&model2);
		if (res1 && res2) {
			if (measurement!=NULL) *measurement = meas1 - meas2;
			if (model!=NULL) *model =  model1 - model2;
			if (usable) return 1;
			else return -1;
		} else {
			return 0;
		}
	} else if (measType==PC) {
		res1 = getMeasModelValue(epoch,system,PRN,C1P,&meas1,&model1);
		res2 = getMeasModelValue(epoch,system,PRN,C2P,&meas2,&model2);
		if (res1 && res2) {
			if (measurement!=NULL) *measurement = (GPSmf1*GPSmf1*meas1 - GPSmf2*GPSmf2*meas2) / (GPSmf1*GPSmf1 - GPSmf2*GPSmf2);
			if (model!=NULL) *model =  (GPSmf1*GPSmf1*model1 - GPSmf2*GPSmf2*model2) / (GPSmf1*GPSmf1 - GPSmf2*GPSmf2);
			if (usable) return 1;
			else return -1;
		} else {
			return 0;
		}
	} else if (measType==PCC) {
		res1 = getMeasModelValue(epoch,system,PRN,C1C,&meas1,&model1);
		res2 = getMeasModelValue(epoch,system,PRN,C2P,&meas2,&model2);
		if (res1 && res2) {
			if (measurement!=NULL) *measurement = (GPSmf1*GPSmf1*meas1 - GPSmf2*GPSmf2*meas2) / (GPSmf1*GPSmf1 - GPSmf2*GPSmf2);
			if (model!=NULL) *model =  (GPSmf1*GPSmf1*model1 - GPSmf2*GPSmf2*model2) / (GPSmf1*GPSmf1 - GPSmf2*GPSmf2);
			if (usable) return 1;
			else return -1;
		} else {
			return 0;
		}
	} else if (measType==LC) {
		res1 = getMeasModelValue(epoch,system,PRN,L1P,&meas1,&model1);
		res2 = getMeasModelValue(epoch,system,PRN,L2P,&meas2,&model2);
		if (res1 && res2) {
			if (measurement!=NULL) *measurement = (GPSmf1*GPSmf1*meas1 - GPSmf2*GPSmf2*meas2) / (GPSmf1*GPSmf1 - GPSmf2*GPSmf2);
			if (model!=NULL) *model =  (GPSmf1*GPSmf1*model1 - GPSmf2*GPSmf2*model2) / (GPSmf1*GPSmf1 - GPSmf2*GPSmf2);
			if (usable) return 1;
			else return -1;
		} else {
			return 0;
		}
	} else if (measType==G1C) {
		res1 = getMeasModelValue(epoch,system,PRN,C1C,&meas1,&model1);
		res2 = getMeasModelValue(epoch,system,PRN,L1P,&meas2,&model2);
		if (res1 && res2) {
			if (measurement!=NULL) *measurement = (meas1 + meas2) / 2.0;
			if (model!=NULL) *model =  (model1 + model2) / 2.0;
			if (usable) return 1;
			else return -1;
		} else {
			return 0;
		}
	} else if (measType==G1P) {
		res1 = getMeasModelValue(epoch,system,PRN,C1P,&meas1,&model1);
		res2 = getMeasModelValue(epoch,system,PRN,L1P,&meas2,&model2);
		if (res1 && res2) {
			if (measurement!=NULL) *measurement = (meas1 + meas2) / 2.0;
			if (model!=NULL) *model =  (model1 + model2) / 2.0;
			if (usable) return 1;
			else return -1;
		} else {
			return 0;
		}
	} else if (measType==G2C) {
		res1 = getMeasModelValue(epoch,system,PRN,C2C,&meas1,&model1);
		res2 = getMeasModelValue(epoch,system,PRN,L2P,&meas2,&model2);
		if (res1 && res2) {
			if (measurement!=NULL) *measurement = (meas1 + meas2) / 2.0;
			if (model!=NULL) *model =  (model1 + model2) / 2.0;
			if (usable) return 1;
			else return -1;
		} else {
			return 0;
		}
	} else if (measType==G2P) {
		res1 = getMeasModelValue(epoch,system,PRN,C2P,&meas1,&model1);
		res2 = getMeasModelValue(epoch,system,PRN,L2P,&meas2,&model2);
		if (res1 && res2) {
			if (measurement!=NULL) *measurement = (meas1 + meas2) / 2.0;
			if (model!=NULL) *model =  (model1 + model2) / 2.0;
			if (usable) return 1;
			else return -1;
		} else {
			return 0;
		}
	} else {
		return 0; // Unknown measurement
	}

	// Seek satellite
	for (i=0;i<epoch->numSatellites;i++) {
		if (epoch->sat[i].GNSS == system && epoch->sat[i].PRN == PRN) {
			// Satellite found
			rawValue = epoch->sat[i].meas[ind].value;
			conversionFactor = epoch->measOrder[system].conversionFactor[measType];
			align = epoch->sat[i].phasePrealign[ind];
			if (measurement!=NULL) *measurement =  (rawValue+align) * conversionFactor;
			if (model!=NULL) *model = epoch->sat[i].meas[ind].model;
			
			if (measurement!=NULL && rawValue == -1) {
				*measurement=-1;
				return 0;
			}
			if (usable) return 1;
			else return -1;
		}
	}
	return 0;
}

/*****************************************************************************
 * Name        : getModelValue
 * Description : Get the model value of a particular satellite
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  TEpoch structure
 * GNSSystem  system               I  N/A  GNSS system of the satellite
 * int  PRN                        I  N/A  PRN identifier of the satellite
 * MeasurementType  measType       I  N/A  Measurement type
 * Returned value (double)         O  m    Model value	(-1 invalid model)
 *****************************************************************************/
double getModelValue(TEpoch *epoch, enum GNSSystem system, int PRN, enum MeasurementType measType) {
	double	model;
	int		res;

	res = getMeasModelValue(epoch,system,PRN,measType,NULL,&model);

	if (res==1) {
		return model;
	} else {
		return -1;
	}
}

/*****************************************************************************
 * Name        : getMeasurementValue
 * Description : Get the model value of a particular satellite
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  TEpoch structure
 * GNSSystem  system               I  N/A  GNSS system of the satellite
 * int  PRN                        I  N/A  PRN identifier of the satellite
 * MeasurementType  measType       I  N/A  Measurement type
 * Returned value (double)         O  m    Measurement value  (-1 invalid measurement)
 *****************************************************************************/
double getMeasurementValue(TEpoch *epoch, enum GNSSystem system, int PRN, enum MeasurementType measType) {
	double	measurement;
	int		res;

	res = getMeasModelValue(epoch,system,PRN,measType,&measurement,NULL);

	if (res==1) {
		return measurement;
	} else {
		return -1;
	}
}

/*****************************************************************************
 * Name        : satWas
 * Description : Check if a satellite was in the former epoch
 * Parameters  :
 * Name                           |Da|Unit|Description
 * GNSSystem  system               I  N/A  GNSS system of the satellite
 * int  PRN                        I  N/A  PRN identifier of the satellite
 * TSatellite  *prevSat            I  N/A  Former satellite vector
 * int  prevNumSatellites          I  N/A  Former size of satellite vector
 * Returned value (int)            O  N/A  Status of the function
 *                                         -1  => New appearance of the satellite
 *                                         0-N => Satellite was in this index in
 *                                                the former epoch 
 *****************************************************************************/
int satWas(enum GNSSystem GNSS,int PRN,TSatellite *prevSat,int prevNumSatellites) {
	int i;
	
	for (i=0;i<prevNumSatellites;i++) {
		if (prevSat[i].GNSS == GNSS && prevSat[i].PRN == PRN) {
			return i; // Satellite found
		}
	}
	return -1; // Satellite not found
}

/*****************************************************************************
 * Name        : getProductsFirstEpochBRDC
 * Description : Return the first epoch TTime in a TGNSSproducts broadcast structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TGNSSproducts  *products        I  N/A  TGNSSproducts structure
 * Returned value (TTime)          O  N/A  Time of first observation in file
 *****************************************************************************/
TTime getProductsFirstEpochBRDC (TGNSSproducts *products){
	TTime	t;
	int		i;
	
	t.MJDN = 0;
	t.SoD = 0;
	
	if (products->type!=BRDC) return t;
	
	// Initialise with the first transmission time of first satellite
	memcpy(&t,&products->BRDC->block[0][0].TtransTime,sizeof(TTime));
	// Get lower transmission time
	for (i=0;i<products->BRDC->numsats;i++) {
		if (tdiff(&products->BRDC->block[i][0].TtransTime,&t)<0) {
			memcpy(&t,&products->BRDC->block[i][0].TtransTime,sizeof(TTime));
		}
	}
	return t;
}

/*****************************************************************************
 * Name        : getProductsFirstEpochSP3
 * Description : Return the first epoch TTime in a TGNSSproducts SP3 structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TGNSSproducts  *products        I  N/A  TGNSSproducts structure
 * Returned value (TTime)          O  N/A  Time of first observation in file
 *****************************************************************************/
TTime getProductsFirstEpochSP3 (TGNSSproducts *products) {
	TTime	t;
	
	t.MJDN = 0;
	t.SoD = 0;
	
	if (products->type!=SP3) return t;
	
	return products->SP3->orbits.startTime; // Header start time
}

/*****************************************************************************
 * Name        : getProductsFirstEpoch
 * Description : Return the first epoch TTime in a TGNSSproducts structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TGNSSproducts  *products        I  N/A  TGNSSproducts structure
 * Returned value (TTime)          O  N/A  Time of first observation in file
 *****************************************************************************/
TTime getProductsFirstEpoch (TGNSSproducts *products) {
	if (products->type == SP3) {
		return getProductsFirstEpochSP3 (products);
	} else { // (products->type == BRDC)
		return getProductsFirstEpochBRDC (products);
	}
}

/*****************************************************************************
 * Name        : getProductsLastEpochBRDC
 * Description : Return the last epoch TTime in a TGNSSproducts broadcast structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TGNSSproducts  *products        I  N/A  TGNSSproducts structure
 * Returned value (TTime)          O  N/A  Time of last observation in file
 *****************************************************************************/
TTime getProductsLastEpochBRDC (TGNSSproducts *products) {
	TTime	t;
	int		i;
	
	t.MJDN = 0;
	t.SoD = 0;
	
	if (products->type!=BRDC) return t;
	
	// Initialise with the last toe time of first satellite
	memcpy(&t,&products->BRDC->block[0][products->BRDC->numblocks[0]-1].Ttoe,sizeof(TTime));
	// Get lower transmission time
	for (i=0;i<products->BRDC->numsats;i++) {
		if (tdiff(&products->BRDC->block[i][products->BRDC->numblocks[i]-1].Ttoe,&t)>0) {
			memcpy(&t,&products->BRDC->block[i][products->BRDC->numblocks[i]-1].Ttoe,sizeof(TTime));
		}
	}
	// Broadcast can be valid for longer periods (up to 2 hours later from Toe)
	t = tdadd(&t,7200);
	return t;
}

/*****************************************************************************
 * Name        : getProductsLastEpochSP3
 * Description : Return the last epoch TTime in a TGNSSproducts SP3 structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TGNSSproducts  *products        I  N/A  TGNSSproducts structure
 * Returned value (TTime)          O  N/A  Time of last observation in file
 *****************************************************************************/
TTime getProductsLastEpochSP3 (TGNSSproducts *products) {
	TTime	t;
	
	t.MJDN = 0;
	t.SoD = 0;
	
	if (products->type!=SP3) return t;

	return products->SP3->orbits.block[0][products->SP3->orbits.numRecords - 1].t;
}

/*****************************************************************************
 * Name        : getProductsLastEpoch
 * Description : Return the last epoch TTime in a TGNSSproducts structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TGNSSproducts  *products        I  N/A  TGNSSproducts structure
 * Returned value (TTime)          O  N/A  Time of last observation in file
 *****************************************************************************/
TTime getProductsLastEpoch (TGNSSproducts *products) {
	if (products->type == SP3) {
		return getProductsLastEpochSP3 (products);
	} else { // (products->type == BRDC)
		return getProductsLastEpochBRDC (products);
	}
}

/*****************************************************************************
 * Name        : selectBlock
 * Description : Select the BRDC block with the following conditions
 *                  - Transmission time equal or lower than reference time
 *                  - Closest toe to reference time
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TBRDCproducts  *products        I  N/A  TBRDCproducts structure used as base to return PVT
 * TTime  *t                       I  N/A  Reference time used
 * GNSSystem  GNSS                 I  N/A  GNSS system of the satellite
 * int  PRN                        I  N/A  PRN identifier of the satellite
 * TOptions  *options              I  N/A  TOptions structure
 * Returned value (TBRDCblock*)    O  N/A  Pointer to the corresponding block
 *****************************************************************************/
TBRDCblock *selectBRDCblock (TBRDCproducts *products, TTime *t, enum GNSSystem GNSS, int PRN, TOptions *options) {
	int			ind;
	int			i;
	TBRDCblock	*block;
	double		diff;
	double		diffTransmitted;
	double		diffToe;
	int			selected = -1;
	double		diffLastTransmitted = 9e9;

	ind = products->index[GNSS][PRN];

	if (ind==-1) return NULL;

	for (i=0;i<products->numblocks[ind];i++) {
		block = &products->block[ind][i];
		diffTransmitted = tdiff(t,&block->TtransTime);
		diffToe = tdiff(t,&block->Ttoe);
		if (diffTransmitted>=0 && diffTransmitted<diffLastTransmitted && diffToe<=7200 && diffToe>=-7200) {
			if (block->SVhealth==0 || options->satelliteHealth==0) {
				diffLastTransmitted = diffTransmitted;
				selected = i;
			}
		}
	}

	if (selected==-1) return NULL;

	// Check if selected is inside the fit interval
	block = &products->block[ind][selected];
	diff = tdiff(t,&block->Ttoe);
	if (diff<=7200 && diff>=-7200) return block;
	else return NULL;
}

/*****************************************************************************
 * Name        : MJDN
 * Description : Get the modified julian day from YY/MM/DD
 * Parameters  :
 * Name                           |Da|Unit|Description
 * struct tm  tm                   I  N/A  Date
 * Returned value (int)            O  N/A  Integer days in Modified Julian Date
 *****************************************************************************/
int MJDN(struct tm *tm) {
	int y,m;

	y = tm->tm_year + 1900;
	m = tm->tm_mon + 1;
	if (m<=2) {
		y--;
		m+=12;
	}

	return (int)(365.25*y) + (int)(30.6001*(m+1)) + tm->tm_mday - 679019;
}

/*****************************************************************************
 * Name        : t2tm
 * Description : Get the calendar time from a TTime structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TTime  *t                       I  N/A  TTime structure
 * tm  *tm                         O  N/A  tm structure
 * double  *seconds                O  N/A  Seconds of minute in double format
 *****************************************************************************/
void t2tm(TTime *t, struct tm *tm, double *seconds) {
	int		year;
	double	doy;
	time_t	tt;
	
	t2doy(t,&year,&doy);
	tm->tm_year = year - 1900;
	tm->tm_mday = (int)(doy);
	tm->tm_mon = 0;
	tm->tm_hour = (int)(t->SoD/3600);
	tm->tm_min = (int)(modulo(t->SoD,3600)/60);
	tm->tm_sec = (int)(modulo(t->SoD,60));
	tm->tm_isdst = -1;
	*seconds =  modulo(t->SoD,60);
	tt = mktime(tm);
	//localtime_r(&tt,tm);
	memcpy(tm,localtime(&tt),sizeof(tm));
}

/*****************************************************************************
 * Name        : t2doy
 * Description : Get the Year/doy from a TTime structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TTime  *t                       I  N/A  TTime structure
 * int  *year                      O  N/A  Year
 * int  *doy                       O  N/A  Day of year
 *****************************************************************************/
void t2doy(TTime *t, int *year, double *doy) {
	const int MJD_1980 = 44239;
	int     day;    
	int     ref = 1980; 
	int     day_aux;
	int     iy;
	
	day = t->MJDN - MJD_1980;
	
	day_aux = modulo(day,1461);
	if (day_aux>365) {
		day_aux -= 366; 
		*doy = modulo (day_aux,365) + 1;
		iy = day_aux/365 + 1;
	} else {
		*doy = day_aux + 1;
		iy = 0; 
	}
	*year = ref + 4*(day/1461) + iy;
	*doy += t->SoD/86400;
}

/*****************************************************************************
 * Name        : t2doystr
 * Description : Get a string with the Year/DoY/SoD from a TTime structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TTime  *t                       I  N/A  TTime structure
 * Returned value (char*)          O  N/A  String with Year / Doy / SoD (length=17)
 *****************************************************************************/
char *t2doystr (TTime *t) {
	static char	str[100];
	int			year;
	double		doy;
	
	t2doy(t,&year,&doy);
	sprintf(str,"%04d %03d %8.2lf",year,(int)(doy),(doy-(int)(doy))*86400.0);
	return str;
}

/*****************************************************************************
 * Name        : gpsws2ttime
 * Description : Transforms between GPSweek and Seconds of Week to TTime structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * int  GPSweek                    I  N/A  GPS Week
 * double  SoW                     I  s    Seconds of week
 * Returned value (TTime)          O  N/A  TTime structure
 *****************************************************************************/
TTime gpsws2ttime(int GPSweek, double SoW) {
	TTime		t;
	int			DoW;

	// GPSweek = 0 and SoW = 0 corresponds to MJDN = 44244
	DoW = (int)(SoW/86400.0);
	t.MJDN = (int)(44244 + GPSweek * 7.0 + DoW);
	t.SoD = SoW - DoW*86400.0;

	return t;
}

/*****************************************************************************
 * Name        : cal2t
 * Description : Transforms between Calendar time (year, month, day of month, 
 *               hour, minute and second) to TTime structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * int  year                       I  y    Year (can be of 4 or 2 digits)
 * int  month                      I  m    Month
 * int  day                        I  d    Day of month
 * int  hour                       I  h    Hours of day
 * int  minute                     I  m    Minutes in hour
 * double  second                  I  s    Seconds of minute
 * Returned value (TTime)          O  N/A  TTime structure
 *****************************************************************************/
TTime cal2t (int year, int month, int day, int hour, int minute, double second) {
	struct tm	tm;
	TTime		t;
	
	if (year<100) { // 2 digits
		if (year<=70) { // 20XX
			tm.tm_year = year + 100;
		} else { // 19XX
			tm.tm_year = year;
		}
	} else { // 4 digits
		tm.tm_year = year - 1900;
	}
	
	tm.tm_mon = month - 1;
	tm.tm_mday = day;
	tm.tm_hour = hour;
	tm.tm_min = minute;
	tm.tm_sec = (int)second;
	t.MJDN = MJDN (&tm);
	t.SoD = tm.tm_hour*3600 + tm.tm_min*60 + second;
	
	return t;
}

/*****************************************************************************
 * Name        : tdiff
 * Description : Difference between two internal time structures in seconds, 
 *               equivalent to t1-t2
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TTime  *a                       I  N/A  Time 1
 * TTime  *b                       I  N/A  Time 2
 * Returned value (double)         O  s    Seconds between t1 and t2
 *****************************************************************************/
double tdiff(TTime *a, TTime *b) {
	return (a->MJDN-b->MJDN)*86400.0 + (a->SoD-b->SoD);
}

/*****************************************************************************
 * Name        : tdadd
 * Description : Add a double (in seconds) from a TTime structure 
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TTime  *a                       I  N/A  Reference time
 * double  b                       I  s    Time to add
 * Returned value (TTime)          O  N/A  a+b
 *****************************************************************************/
TTime tdadd(TTime *a, double b) {
	TTime		t;
	int			MJDNdiff;

	t.MJDN = a->MJDN;
	t.SoD = a->SoD + b;

	MJDNdiff = (int)(t.SoD/86400);

	t.MJDN += MJDNdiff;
	t.SoD -= MJDNdiff*86400;

	if (t.SoD<0) {
		t.MJDN--;
		t.SoD += 86400;
	}

	return t;
}

/*****************************************************************************
 * Name        : numSteps
 * Description : Return the number of increments of da between values a2 and a1
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double  a1                      I  N/A  First value
 * double  a2                      I  N/A  Second value
 * double  da                      I  N/A  Increment between values
 * Returned value (int)            O  N/A  Number of increments
 *****************************************************************************/
int numSteps(double a1, double a2, double da) {
	if (da==0) return 1;
	if (da==-1) return 0; // Uninitialised

	return (int)((a2-a1)/da+1); // +1 is to take into account that both a1 and a2 are included in the distance
}

/*****************************************************************************
 * Name        : allocateIonoMap
 * Description : Allocate memory for ionospheric map. The following fields of
 *               the structure must be already prepared for this function:
 *                    nhgt, nlat, nlon
 *               As they will be used to allocate memory in the members:
 *                    TEC & RMS
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TIonoMap  *ionoMap              IO N/A  Ionospheric map structure
 *****************************************************************************/
void allocateIonoMap (TIonoMap *ionoMap) {
	int		totalSize;

	totalSize = ionoMap->nhgt * ionoMap->nlat * ionoMap->nlon;

	ionoMap->TEC = (double*)(realloc(ionoMap->TEC,sizeof(double)*totalSize));
	ionoMap->RMS = (double*)(realloc(ionoMap->RMS,sizeof(double)*totalSize));
}

/*****************************************************************************
 * Name        : E
 * Description : Return the value 10^exponent withouth using pow function
 * Parameters  :
 * Name                           |Da|Unit|Description
 * int  exponent                   I  N/A  Exponent
 * Returned value (double)         O  N/A  10^exponent
 *****************************************************************************/
double E(int exponent) {
	double	value = 1;
	int		i;

	if (exponent<0) {
		for (i=0;i>exponent;i--) value /= 10;
	} else if (exponent>0) {
		for (i=0;i<exponent;i++) value *= 10;
	}

	return value;
}

/*****************************************************************************
 * Name        : conv3Dto1D
 * Description : Transforms a 3 dimensional index to a single dimension index
 * Parameters  :
 * Name                           |Da|Unit|Description
 * int  i                          I  N/A  First dimension  
 * int  j                          I  N/A  Second dimension
 * int  k                          I  N/A  Third dimension
 * int  ni                         I  N/A  Size of first dimension
 * int  nj                         I  N/A  Size of second dimension
 * Returned value (int)            O  N/A  1D index
 *****************************************************************************/
int	conv3Dto1D (int i, int j, int k, int ni, int nj) {
	if (ni==0 || nj==0) return 0;

	return k*ni*nj + j*ni + i;
}

/*****************************************************************************
 * Name        : ionoConv3Dto1D
 * Description : Transforms a 3 dimensional index to a single dimension index
 *               obtaining the data from a TIonoMap structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TIonoMap  *ionoMap              I  N/A  Ionospheric map structure
 * int  ihgt                       I  N/A  Height index  
 * int  ilat                       I  N/A  Latitude index
 * int  ilon                       I  N/A  Longitude index
 * Returned value (int)            O  N/A  1D index
 *****************************************************************************/
int ionoConv3Dto1D (TIonoMap *ionoMap, int ihgt, int ilat, int ilon) {
	return conv3Dto1D(ihgt,ilat,ilon,ionoMap->nhgt,ionoMap->nlat);
}

/*****************************************************************************
 * Name        : ionoHLLto1D
 * Description : Gives the 1D index of the double values of hgt, lat and lon
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TIonoMap  *ionoMap              I  N/A  Ionospheric map structure
 * double  hgt                     I  km   Height value  
 * double  lat                     I  deg  Latitude value
 * double  lon                     I  deg  Longitude value
 * Returned value (int)            O  N/A  1D index
 *****************************************************************************/
int ionoHLLto1D (TIonoMap *ionoMap, double hgt, double lat, double lon) {
	int		ilat, ilon, ihgt;

	if (ionoMap->nhgt==1) {
		ihgt=0;
	} else {
		ihgt = (int)((hgt-ionoMap->hgt1)/ionoMap->dhgt);
	}

	if (ionoMap->nlat==1) {
		ilat=0;
	} else {
		ilat = (int)((lat-ionoMap->lat1)/ionoMap->dlat);
	}

	if (ionoMap->nlon==1) {
		ilon=0;
	} else {
		ilon = (int)((lon-ionoMap->lon1)/ionoMap->dlon);
	}

	return ionoConv3Dto1D(ionoMap, ihgt, ilat, ilon);
}

/*****************************************************************************
 * Name        : setIonoValue
 * Description : Sets the TEC or RMS value of a ionospheric map
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TIonoMap  *ionoMap              IO N/A  Ionospheric map structure
 * int  type                       I  N/A  TEC or RMS (0 or 1)
 * double  hgt                     I  km   Height value  
 * double  lat                     I  deg  Latitude value
 * double  lon                     I  deg  Longitude value
 * double  value                   I  TECU Value to fill
 * Returned value (int)            O  N/A  Status of function
 *                                         0 => Failed to set the value
 *                                         1 => Value properly set
 *****************************************************************************/
int setIonoValue(TIonoMap *ionoMap, int type, double hgt, double lat, double lon, double value) {
	int ind;

	ind = ionoHLLto1D(ionoMap,hgt,lat,lon);

	if (ind<0) return 0;

	if (type==0) { // TEC
		ionoMap->TEC[ind] = value;
	} else if (type==1) { // RMS
		ionoMap->RMS[ind] = value;
	} else {
		return 0;
	}

	return 1;
}

/*****************************************************************************
 * Name        : getIonoValue
 * Description : Returns the TEC or RMS value of a ionospheric map
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TIonoMap  *ionoMap              I  N/A  Ionospheric map structure
 * int  type                       I  N/A  TEC or RMS (0 or 1)
 * double  hgt                     I  km   Height value  
 * double  lat                     I  deg  Latitude value
 * double  lon                     I  deg  Longitude value
 * Returned value (double)         O  TECU Value of the cell (in TECUs).
 *                                    9999 is used for invalid values,
 *                                    as defined in IONEX format)
 *****************************************************************************/
double getIonoValue(TIonoMap *ionoMap, int type, double hgt, double lat, double lon) {
	int ind;

	ind = ionoHLLto1D(ionoMap,hgt,lat,lon);

	if (ind<0) return 9999;

	if (type==0) { // TEC
		return ionoMap->TEC[ind];
	} else if (type==1) { // RMS
		return ionoMap->RMS[ind];
	}

	return 9999;
}

/*****************************************************************************
 * Name        : getIonoMapTotalSize
 * Description : Returns the total size of a map
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TIonoMap  *ionoMap              I  N/A  Ionospheric map structure
 * Returned value (int)            O  N/A  Map total size
 *****************************************************************************/
int getIonoMapTotalSize(TIonoMap *ionoMap) {
	return ionoMap->nhgt * ionoMap->nlat * ionoMap->nlon;
}

/*****************************************************************************
 * Name        : getConstellationElement
 * Description : Returns the TConstellationElement of a satellite for a given
 *               time
 * Parameters  :
 * Name                           |Da|Unit|Description
 * GNSSystem  GNSS                 I  N/A  GNSS enumerator of the satellite
 * int  PRN                        I  N/A  PRN of the satellite
 * TTime  *t                       I  N/A  Reference time
 * TConstellation  *constellation  I  N/A  TConstellation structure
 * Returned value (TConstellationElement*)
 *                                 O  N/A  Pointer to the TConstellationElement
 *                                         selected
 *                                         NULL => Satellite not found
 *****************************************************************************/
TConstellationElement* getConstellationElement(enum GNSSystem GNSS, int PRN, TTime *t,TConstellation* constellation) {
	int i;

	for (i=0;i<constellation->numSatellites;i++) {
		if (constellation->sat[i].GNSS == GNSS && constellation->sat[i].PRN == PRN) {
			if (tdiff(t,&constellation->sat[i].tLaunch)>0 && tdiff(t,&constellation->sat[i].tDecommissioned)<0) {
				// If t is beyond tLaunch but lower than tDecommissioned
				return &constellation->sat[i];
			}
		}
	}
	return NULL;
}

/*****************************************************************************
 * Name        : getStation
 * Description : Get the station structure from its name (as a string)
 * Parameters  :
 * Name                           |Da|Unit|Description
 * char  *name                     I  N/A  Name of the station
 * TStationList  *stationList      I  N/A  List of stations to look for
 * Returned value (*TStatoin)      O  N/A  Pointer to the antenna inside TStationList
 *                                         If it is NULL, the station has not been
 *                                         found
 *****************************************************************************/
TStation *getStation(char *name, TStationList *stationList) {
	int		i;
	
	for (i=0;i<stationList->n;i++) {
		if (strcasecmp(name,stationList->station[i].name)==0) return &stationList->station[i];
	}
	
	return NULL;
}

/*****************************************************************************
 * Name        : addStat
 * Description : Add a value to a TStat structure, computing the mean and square
 *               mean of all the values that have been inserted
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TStat  *stat                    IO N/A  TStat structure to operate
 * double  value                   I  N/A  Value to include to stat
 *****************************************************************************/
void addStat (TStat *stat, double value) {
	stat->mean = (stat->n*stat->mean + value)/(stat->n+1);
	stat->mean2 = (stat->n*stat->mean2 + value*value)/(stat->n+1);
	stat->n++;
}
