/****************************************************************************
   Copyright & License:
   ====================
   
   Copyright 2009 - 2010 gAGE/UPC & ESA
   
   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at
   
   http://www.apache.org/licenses/LICENSE-2.0
   
   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 ***************************************************************************/

/****************************************************************************
 * Copyright: gAGE/UPC & ESA
 * Project: EDUNAV GNSS Lab Tool 
 * Author: Pere Ramos-Bosch (group of Astronomy and GEomatics - gAGE/UPC)
 * Maintainer: Adria Rovira-Garcia ( gAGE/UPC )
 *             glab.gage @ upc.edu        
 * File: gLAB.c
 * Code Management Tool File Version: 2.0  Revision: 0 
 * Date: 2010/11/30
 ***************************************************************************/

/****************************************************************************
 * MODULE DESCRIPTION
 *
 * Name: gLAB
 * Language: C
 *
 * Purpose:
 *  The purpose of this module is to create the main executable and parse all the
 *  command-line options
 *
 * Function:
 *  This class is in charge of organising the rest of the modules, creating the 
 *  chain process, and to parse all the command-line options.
 *
 * Dependencies:
 *  dataHandling, input, preprocessing, model, filter, output
 *
 * Files modified:
 *  None
 *
 * Files read:
 *  None
 *
 * Resources usage:
 *  See Design Document
 ****************************************************************************/

/****************************
 *       RELEASE_HISTORY
 * -------------
 *          gLAB v1.3.1
 * Release: 2010/01/22
 * Change Log: First public version
 * -------------
 *          gLAB v1.3.2
 * Release: 2010/02/02
 * Change Log: Corrected problem with the Orbit comparison
 *             Exported SATDIFF, SATSTAT and SATSTATTOT printing to output module
 *             For orbit comparison mode, it can now disconnect the satellite phase center
 *             For orbit comparison mode, when satellite phase center correction is active,
 *                  it will now require an ANTEX or Constellation file
 *             Included SHOWPRODUCT mode
 *             Added --output parameter
 *             Enhanced parsing of antenna in ANTEX file
 *             Included show product mode
 *             '-input:SP3' parameter can now be provided as '-input:sp3'
 *             Corrected some typos in the help
 *             Changed '-output' to '-output:file'
 *             Included the '--output:file' option to redirect to stdout 
 *             Changed the message help to be consistent with the new updates
 *             minArcLength changed from 3 to 4 to be sure that the Li cycle-slip detector 
 *                  has properly worked
 *             Included DCB data sources
 *             Forward/Backward processing included
 *             SATSTAT message will now provide the GNSS system it refers to
 *             '-pre:setrectype' option is added to force a specific receiver type
 * -------------
 *          gLAB v1.4.0
 * Release: 2010/06/21
 * Change Log: '-pre:flexibleC1P1' option is added to allow flexible handling of C1 and P1
 *             '-pre:prealign' option to set/unset carrier-phase prealignment
 *             Cycle-slip detector changed to specify when it does not have enough measurements
 *             Added a check to see if the RINEX observation file has times inside the products
 *                  file provided
 * -------------
 *          gLAB v1.4.1
 * Release: 2010/07/09
 * Change Log: No changes in this file.
 * -------------
 *          gLAB v1.4.2
 * Release: 2010/07/31
 * Change Log: No changes in this file.
 * -------------
 *          gLAB v1.4.3
 * Release: 2010/08/31
 * Change Log: Maximum number of filter iterations added.
 * --------------
 *          gLAB v1.4.4
 * Release: 2010/09/22
 * Change Log: Included new DCB management parsing options.
 * -------------
 *          gLAB v1.4.5
 * Release: 2010/10/31
 * Change Log: Tropospheric correction '-model:trop' option is extended to '-model:trop' to activate/deactivate
 *             '-model:trop:nominal <nominal>' to select the nominal used '-model:trop:mapping <mapping>' to 
 *             select which mapping is used. This permits cross combinations among mappings and nominals.
 * 	       Default precise orbit interpolation degree changed from 9 to 10.
 *             Filter 'inf' value for clock and coordinates noise process and initial covariance substituted
 *                  to real floating quantities. SIGMA_INF splitted into different SIGMA_INF_DR and SIGMA_INF_DT.
 * -----------
 *          gLAB v2.0.0
 * Release: 2010/11/30
 * Change Log: gLAB version released together with the Educational Book:
 *             "Global Navigation Satellite Systems: Fundamentals, Algorithms and Laboratory Exercises"
 *             Authors: J. Sanz, JM. Juan and M. Hernandez-Pajares.
 *
 *             SATSEL debugging message included in the help messages.
 * -----------
 *       END_RELEASE_HISTORY
 *****************************/

/* System modules */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* External classes */
#include "input.h"
#include "preprocessing.h"
#include "model.h"
#include "filter.h"
#include "dataHandling.h"
#include "output.h"

/* Internal types, constants and data */

#define VERSION 			"2.0.0"
#define MAX_WORD_LENGTH		300
#define MAX_WORDS			200

// Enumerator for Navigation kinematics
enum NK {
	NKND,
	NKstatic,
	NKkinematic
};

// Enumerator for Navigation mode
enum NM {
	NMND,
	NMppp,
	NMstandalone
};

// Enumerator for Navigation measurements
enum NO {
	NOND,
	NOpseudorange,
	NOcarrierphase
};

// Struct for support options
typedef struct {
	enum	NK			navKinematics;
	enum	NM			navMode;
	enum	NO			navObservables;
	enum	IonoModel		ionoModel;
	enum    TropModelNominal        tropNominal; 
	enum    TropModelMapping	tropMapping;
	enum	P1C1DCBModel   p1c1dcbModel;
	enum	P1P2DCBModel   p1p2dcbModel;
	int					estimateTroposphere;
	int					troposphericCorrection;
	char				cfgFile[MAX_INPUT_LINE];
	char				obsFile[MAX_INPUT_LINE];
	char				navFile[2][MAX_INPUT_LINE];
	int					numNav;
	char				SP3File[2][MAX_INPUT_LINE];
	int					numSP3;
	char				orbFile[2][MAX_INPUT_LINE];
	int					numOrb;
	char				clkFile[2][MAX_INPUT_LINE];
	int					numClk;
	char				conFile[MAX_INPUT_LINE];
	char				outFile[MAX_INPUT_LINE];
	char				antFile[MAX_INPUT_LINE];
	char				klbFile[MAX_INPUT_LINE];
	char				tgdFile[2][MAX_INPUT_LINE];
	char				recFile[MAX_INPUT_LINE];
	char				snxFile[MAX_INPUT_LINE];
	int					numTGD;
        int                             numP1C1DCB;
        int                             numP1P2DCB;
        int                             numP1P2RINEX;
	int					csLi;
	int					csBW;
	int					csL1C1;
	int					totalFilterMeasurements;
	enum MeasurementType	measurement[MAX_FILTER_MEASUREMENTS];
	enum MeasurementType	smoothWith[MAX_FILTER_MEASUREMENTS];
	double				fixedMeasurementWeight[MAX_FILTER_MEASUREMENTS];
	double				elevationMeasurementWeight[MAX_FILTER_MEASUREMENTS][2];
	double				filterParams[3][MAX_PAR];
						// filterParams is the data for Phi, Q and P0, the first dimension of the matrix
						// accounts for: 0->Phi,  1->Q  and  2->P0. To access it: PHI_PAR, Q_PAR and P0_PAR
						// The second dimension accounts for the unknown types, in particular, to access it
						// it should be used: DR_UNK, DT_UNK, TROP_UNK and BIAS_UNK
	int					discardEclipsedSatellites;
	enum ReceiverType	setRecType;
} TSupportOptions;

/**************************************
 * Declarations of internal operations
 **************************************/

/*****************************************************************************
 * Name        : initSupportOptions
 * Description : Initialises support options
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TSupportOptions  *sopt          IO N/A  TSupportOptions structure to initialise
 *****************************************************************************/
void initSupportOptions (TSupportOptions *sopt) {
	int	i,j;
	
	sopt->navKinematics = NKND;
	sopt->navMode = NMND;
	sopt->navObservables = NOND;
	sopt->cfgFile[0] = '\0';
	sopt->obsFile[0] = '\0';
	sopt->conFile[0] = '\0';
	sopt->recFile[0] = '\0';
	sopt->snxFile[0] = '\0';
	sopt->navFile[0][0] = '\0';
	sopt->SP3File[0][0] = '\0';
	sopt->orbFile[0][0] = '\0';
	sopt->clkFile[0][0] = '\0';
	sopt->navFile[1][0] = '\0';
	sopt->SP3File[1][0] = '\0';
	sopt->orbFile[1][0] = '\0';
	sopt->clkFile[1][0] = '\0';
	sopt->tgdFile[0][0] = '\0';
	sopt->tgdFile[1][0] = '\0';
	sopt->numNav = 0;
	sopt->numSP3 = 0;
	sopt->numOrb = 0;
	sopt->numClk = 0;
	sopt->numTGD = 0;
	sopt->numP1C1DCB = 0;
	sopt->numP1P2DCB = 0;
	sopt->numP1P2RINEX = 0;
	sopt->outFile[0] = '\0';
	sopt->antFile[0]= '\0';
	sopt->klbFile[0] = '\0';
	sopt->ionoModel = UNKNOWN_IM;
	sopt->troposphericCorrection = -1;
	sopt->tropNominal = UNKNOWN_TMNominal; 
	sopt->tropMapping = UNKNOWN_TMMapping;
	sopt->p1c1dcbModel = p1c1UNKNOWN;
	sopt->p1p2dcbModel = p1p2UNKNOWN;
	sopt->csLi = -1;
	sopt->csBW = -1;
	sopt->csL1C1 = -1;
	sopt->estimateTroposphere = -1;
	sopt->totalFilterMeasurements = -1;
	sopt->discardEclipsedSatellites = -1;
	for (i=0;i<MAX_FILTER_MEASUREMENTS;i++) {
		sopt->measurement[i] = NA;
		sopt->smoothWith[i] = NA;
		sopt->fixedMeasurementWeight[i] = -1;
		sopt->elevationMeasurementWeight[i][0] = -1;
		sopt->elevationMeasurementWeight[i][1] = -1;
	}
	for (i=0;i<3;i++) {
		for (j=0;j<MAX_PAR;j++) {
			sopt->filterParams[i][j] = -1;
		}
	}
	sopt->setRecType = rtNA;
}

/*****************************************************************************
 * Name        : showHelp
 * Description : Shows the program help in the standard output
 * Parameters  : None
 *****************************************************************************/
void showHelp () {
	printf("gLAB version v%s\n",VERSION);
	printf("GNSS LABORATORY. Tool to analyze GNSS data\n\n");
	printf("Author:\n");
	printf("    Pere Ramos-Bosch        group of Astronomy and GEomatics (gAGE/UPC)\n\n");
	printf("Maintainer:\n");
	printf("    Adria Rovira-Garcia     group of Astronomy and GEomatics (gAGE/UPC)\n");
	printf("                            Please send any bugs/comments to:  glab.gage @ upc.edu   \n\n");
	printf("Website:\n");
	printf("    www.gage.es/gLAB\n\n");     
	printf("Usage:\n");
	printf("    gLAB [options]\n");
	printf("Options:\n");
	printf("    -help                   Shows this help\n");
	printf("    -examples               Shows some examples to call the program\n");
	printf("    -config                 Shows some info on the configuration file\n");
	printf("    -messages               Shows the fields of each output message (see below, 'VERBOSE OPTIONS')\n");
	printf("\n  INPUT OPTIONS\n");
	printf("    -input:cfg <file>       Sets the input configuration file\n");
	printf("    -input:obs <file>       Sets the input RINEX observation file\n");
	printf("    -input:nav <file>       Sets the input RINEX navigation message file\n");
	printf("    -input:sp3 <file>       Sets the input SP3 orbits and clocks file\n");
	printf("    -input:orb <file>       Sets the input SP3 orbits\n");
	printf("    -input:clk <file>       Sets the input clock file\n");
	printf("    -input:con <file>       Sets the input constellation status file (default GPSConstellationStatus.txt)\n");
	printf("    -input:ant <file>       Sets the input ANTEX satellite/receiver antenna data file\n");
	printf("    -input:klb <file>       Sets the input RINEX navigation file for Klobuchar corrections, if this parameter\n");
	printf("                            is avoided, it is used the -input:nav file\n");
	printf("    -input:dcb <file>       Sets the input DCB source as a .DCB file or a RINEX navigation file\n");
	printf("    -input:rec <file>       Sets the input GPS receiver types file (default GPS_Receiver_Types) for DCB usage\n");
	printf("    -input:snx <file>       Sets the input SINEX file for receiver position\n");
	printf("    The use of '-input:nav' file will preconfigure the parameters to work in the Standard Point Positioning (SPP).\n");
	printf("    The use of '-input:SP3'/'-input:orb'/'-input:clk' will preconfigure the parameters to work in the Precise Point\n");
	printf("    Positioning (PPP) approach. All the values can be overridden by specifying the parameters. See below for more details\n");
	printf("    on which are the defaults for each mode\n");
	printf("\n  PREPROCESSING OPTIONS   (use -pre:... to activate, --pre:... to deactivate, when applicable)\n\n");
	printf("    -pre:dec #              # = number              Decimate input data by # seconds [default 300]\n");
	printf("                            In product comparison mode (see below, in 'WORK MODES'), it sets the time step of the\n");
	printf("                            comparisons\n");
	printf("    -pre:sat g#             Exclude satellite from processing\n");
	printf("                            g = character determining GNSS system (G->GPS)\n");
	printf("                            # = PRN number\n");
	printf("    -pre:elevation <val>    Elevation mask. Satellites below this threshold will be discarded (in degrees) [default 5]\n");
	printf("    -pre:eclipse            Discard satellites under Earth eclipse [default: enabled for PPP, disabled for SPP]\n");
	printf("    -pre:availf g#          Mark frequencies available [default all]\n");
	printf("                            g = character determining GNSS system (G->GPS)\n");
	printf("                            # = frequencies available\n");
	printf("                            Sample: '-pre:availf G12' Frequencies 1 and 2 of GPS available\n");
	printf("                                    '-pre:availf G1'  Frequency 1 of GPS available, 2 unavailable\n");
	printf("    -pre:smooth <val>       Number of epochs of smoothing [default 0, which means disabled]\n");
	printf("    -pre:smoothMeas <n> <meas>         Smooth measurement in filter <n> with measurement type <meas>\n");
	printf("                            In general, <n> is a pseudorange and <meas> a carrier phase, see option '-filter:select' for more\n");
	printf("                            information in the selection of measurements to be used in the filter\n");
	printf("                            Sample: '-pre:smoothMeas 1 LC'\n");
	printf("    -pre:setrecpos <val>    <val> = RINEX           Set the receiver a priori position as the one specified in the RINEX\n");
	printf("                                                    observation file [default]\n");
	printf("                            <val> = SINEX           Set the receiver a priori position to be read from a SINEX file (to be\n");
	printf("                                                    specified by the '-input:snx' parameter\n");
	printf("                            <val> = calculate       The receiver a priori position will be calculated by the program.\n");
	printf("                                                    This is especially usefun when processing moving receivers (trajectories)\n");
	printf("                                                    or when the approximate receiver position is not known. With this option\n");
	printf("                                                    activated, the differential fields of the OUTPUT message will be zero.\n");
	printf("                            <val> = <x> <y> <z>     Specify the receiver a priori position in meters\n");
	printf("                                                    Sample: '-pre:recpossrc 4789032.7143 176594.9690 4195013.2268'\n");
	printf("    -pre:setrectype <val>   <val> = gpsrt           Set the receiver type as the one specified in the GPS_Receiver_Types\n");
	printf("                                                    file (provided by '-input:rec') [default if '-input:rec' provided]\n");
	printf("                            <val> = 0               Set the receiver type as 'Unknown' [default]\n");
	printf("                            <val> = 1               Set the receiver type as 'Cross-correlated'\n");
	printf("                                                    In this mode, P2 will be corrected with the DCB of P1-C1\n");
	printf("                            <val> = 2               Set the receiver type as 'No P1'\n");
	printf("                                                    In this mode, C1 will be used instead of P1\n");
	printf("                            <val> = 3               Set the receiver type as 'Consistent measurements'\n");
	printf("    -pre:prealign           Prealign carrier phase measurements with its corresponding pseudorange [default on]\n");
	printf("    -pre:cs:li              Use the carrier phase geometric-free combination (Li) for cycle-slip detection (2 frequencies) [default in PPP]\n");
	printf("    --pre:cs:li             Do not use carrier phase geometric-free combination (Li) for cycle-slip detection (2 frequencies) [default in SPP]\n");
	printf("                            Li combination is a slow varying function which only has ambiguity and ionosphere\n");
	printf("                            Each epoch and satellite an expected Li value is computed and compared against the\n");
 	printf("                            measured one, if this difference is above a certain threshold, a cycle-slip is marked.\n");
 	printf("                            This threshold is (max + (min-max)*exp(-dt/t)), being dt the time between epochs\n");
	printf("    -pre:cs:li:min <val>    Minimum threshold between estimated and measured Li values [default 0.034]\n");
	printf("    -pre:cs:li:max <val>    Maximum threshold between estimated and measured Li values [default 0.08]\n"); 
	printf("    -pre:cs:li:t0 <val>     Time constant to set the threshold between maximum and minimum [default 60]\n");
	printf("    -pre:cs:bw              Use Melbourne-Wübbena (BW) for cycle-slip detection (2 frequencies) [default in PPP]\n");
	printf("    --pre:cs:bw             Do not use Melbourne-Wübbena (BW) for cycle-slip detection (2 frequencies) [default in SPP]\n");
	printf("                            BW combination is a mixed combination between carrier phase and pseudoranges.\n");
	printf("                            This combination is free of ionosphere and geometry, thence constant, but\n");
	printf("                            has a high noise (due to the pseudorange measurements used). If this noise is\n");
	printf("                            low enough, it is straightforward to detect cycle-slips, but in noisy environments\n");
	printf("                            BW is not able to detect cycle-slips. The algorithm using BW computes the mean and\n");
	printf("                            the standard deviation of the last epochs in order to obtain an estimated value\n"); 
	printf("                            and the noise level of the combination. The estimated value is compared against the\n");
	printf("                            measured value, and the noise level is used for the threshold of this difference.\n");
	printf("                            Difference must be lower than standard deviation multiplied by a number (slope).\n");
	printf("                            This threshold has minimum (min) and maximum (max) saturation values.\n");
	printf("    -pre:cs:min <val>       Minimum threshold between estimated and measured BW values in relation to its standard deviation [default 0.9]\n");
	printf("    -pre:cs:max <val>       Maximum threshold between estimated and measured BW values in relation to its standard deviation [default 18]\n");
	printf("    -pre:cs:slope <val>     Relation between estimated and measured BW values in relation to its standard deviation [default 9]\n");
	printf("    -pre:cs:l1c1            Use the L1-C1 combination for cycle-slip detection (1 frequency) [default in PPP]\n");
	printf("    --pre:cs:l1c1           Do not use the L1-C1 combination for cycle-slip detection (1 frequency) [default in SPP]\n");
	printf("                            The L1-C1 averages the difference between carrier phase and pseudorange measurements in F1 for several epochs.\n");
	printf("                            This cycle-slip detection method is very useful for single-frequency receivers, as it only requires\n");
	printf("                            measurements from one frequency. As a counterpart, the ionospheric term is different for C1 and L1\n");
	printf("                            thence, this combination will tend to diverge. It becomes necessary to set a smoothing window\n");
	printf("                            to limit this divergence (-pre:cs:l1c1:window). This makes that this method becomes a bit limited\n");
	printf("                            with data rates too low. Ideally this method should be used with rates of 1Hz\n");
	printf("    -pre:cs:l1c1:slope <val>       Relation between estimated and measured L1-C1 values in relation to its standard deviation [default 5]\n");
	printf("    -pre:cs:l1c1:window <val>      Number of epochs to limit the L1-C1 ionosphere divergence [default 100]\n");
	printf("    -pre:cs:l1c1:max <val>         Maximum standard deviation for the threshold calculation in the L1-C1 [default 15] (m) \n");
	printf("\n  MODELLING OPTIONS   (use -model:... to activate, --model:... to deactivate)\n");
	printf("    -model:iono <val>       <val> = no              Do not correct ionosphere [default in PPP] (equivalent to '--model:iono')\n");
	printf("                                  = Klobuchar       Correct measurements with Klobuchar model [default in SPP]\n");
	printf("    -model:trop             Correct troposphere delay [default on] \n");
	printf("    --model:trop            Do not correct troposphere delay \n");
	printf("    -model:trop:nominal <val>   <val> = Simple     Compute a simple tropospheric nominal depending on receiver's height over the sea level. [default in PPP]\n");
	printf("                                <val> = UNB3       Troposphere nominals are calculated from the receiver’s height and estimates of five meteorological\n");
	printf("                                                   parameters: pressure, temperature, water vapour pressure, temperature lapse rate and water vapour lapse rate.\n");
	printf("                                                   It is adopted by SBAS systems (see RTCA-MOPS, 2006). [default in SPP]\n");
	printf("    -model:trop:mapping <val>   <val> = Simple     Compute the mapping as the obliquity factor described in Black and Eisner, 1984. This mapping\n");
	printf("                                                   only depends on satellite elevation and it is common for wet and dry components.  [default in SPP]\n");
	printf("                                <val> = Niell      Compute the mapping described in  A.E. Niell, 1996. This mapping  considers different obliquity\n");
	printf("                                                   factors for the wet and dry components [default in PPP]\n");	
	printf("    -model:satclocks        Correct the measurements with the satellite clock offsets estimations [default on]\n");
	printf("    -model:relclock         Correct the measurements with the relativistic clock model [default on]\n");
	printf("    -model:satmovinflight   Consider satellite movement during signal flight time [default on]\n");
	printf("    -model:earthrotinflight Consider Earth rotation during signal flight time [default on]\n");
	printf("    -model:satphasecenter   Correct satellite phase center to mass center corrections [default on in PPP, off in SPP]\n");
	printf("    -model:recphasecenter <val>    <val> = no                Do not correct antenna receiver phase center [default in SPP]\n");
	printf("                            <val> = ANTEX                    Use the ANTEX file to correct the antenna phase center [default in PPP]\n");
	printf("                            <val> = <nfreq> <dN> <dE> <dU>   Set dN, dE, dU as the antenna phase center for the frequency <nfreq>\n");
	printf("                                       <nfreq> = Frequency number (typical for GPS, 1 and 2)\n");
	printf("                                       <dN>    = North correction [in meters]\n");
	printf("                                       <dE>    = East correction [in meters]\n");
	printf("                                       <dU>    = Up correction [in meters]\n");
	printf("                            Sample: '-model:recphasecenter 1 0 0 0.11 -model:recphasecenter 2 0 0 0.128'\n");
	printf("                                    Typical AOAD/M_T antenna phase center correction, 11 cm and 12.8 cm in the\n");
	printf("                                    Up component for frequencies 1 and 2\n");
	printf("    -model:radomestrict     When using ANTEX to correct the receiver phase center, this option will force a perfect match between\n");
	printf("                            the antenna name and radome reported in the receiver RINEX file and the ones in the ANTEX file. If this option\n");
	printf("                            is disabled, and the radome is not found, the corrections will be used considering the radome \"NONE\" [default off]\n");   
	printf("    -model:arp <val>        <val> = no              Do not apply any Antenna Referece Point (ARP) correction [default in SPP] (equivalent to '--model:arp')\n");
	printf("                            <val> = RINEX           Use as ARP the information on the RINEX file (field 'ANTENNA: DELTA H/E/N') [default in PPP]\n");
	printf("                            <val> = <dN> <dE> <dU>  Set dN, dE, dU as the ARP\n");
	printf("                                       <dN>    = North correction [in meters]\n");
	printf("                                       <dE>    = East correction [in meters]\n");
	printf("                                       <dU>    = Up correction [in meters]\n");
	printf("    -model:dcb:p1c1 <val>   <val> = no        Do not correct Differential Code Biases between P1 and C1 codes (equivalent to '--model:dcb:p1c1')\n");
        printf("                                              Do not identify C1 and P1 never\n");
	printf("                            <val> = flexible  Do not correct Differential Code Biases between P1 and C1 codes [default on]\n");
        printf("                                              Identify C1 and P1 when one of them is missing\n");
	printf("                            <val> = strict    Correct  Differential Code Biases between P1 and C1 codes\n");
        printf("                                              Identify C1 and P1 when applicable (defined by the receiver type)\n");
	printf("    -model:dcb:p1p2 <val>   <val> = no        Do not correct Differential Code Biases between codes P1 and P2 (equivalent to '--model:dcb:p1p2')\n");
	printf("                            <val> = RINEX     Correct Differential Code Biases between codes P1 and P2 using TGDs from the RINEX navigation file [default in SPP, off in PPP]\n");
	printf("                            <val> = DCB       Correct Differential Code Biases between codes P1 and P2 from a precise DCB file\n");
        printf("                                              Identify C1 and P1 when applicable (defined by the receiver type)\n");
	printf("    -model:windup           Correct the wind up term for carrier phase measurements [default]\n");
	printf("    -model:solidtides       Correct the Earth surface deformation due to solid tides [default on in PPP, off in SPP]\n");
	printf("    -model:relpath          Correct the path range delay term due to the gravitational gradient between receiver and transmitter\n"); 
	printf("                            [default on in PPP, off in SPP]\n");
	printf("    -model:orbit:deg <val>  Precise orbit interpolation degree [default 10]\n");
	printf("    -model:clock:deg <val>  Precise clock interpolation degree [default 0 - no interpolation]\n");
	printf("    -model:satellitehealth  Only valid when using broadcast products. Use the healthy flag of the navigation message [default on]\n");
	printf("\n  FILTERING OPTIONS\n");
	printf("    -filter:trop            Estimate the troposphere of the station [default in PPP with carrier phase] ('--filter:trop' to disable it)\n");
	printf("    -filter:nav <nav>       <nav> = static          Process supposing a static receiver [default in PPP]\n");
	printf("                                  = kinematic       Process supposing a moving receiver [default in SPP]\n");
	printf("    -filter:meas <meas>     <meas> = pseudorange    Use only pseudorange for positioning\n");
	printf("                                   = carrierphase   Use pseudorange and carrier phase for positioning [default]\n");
	printf("    -filter:select <num> <meas1> <meas2> ...        Select the measurements or combinations to include in the filtering\n");
	printf("                            <num> = Number of measurements/combinations\n");
	printf("                            <measN> = List of measurements/combinations\n");
	printf("                            [Defaults]: PPP - Pseudorange      => 1 PC\n");
	printf("                            [Defaults]: PPP - Carrierphase     => 2 PC LC\n");
	printf("                            [Defaults]: SPP - Pseudorange      => 1 C1C\n");
	printf("                            [Defaults]: SPP - Carrierphase     => 2 C1C L1P\n");
	printf("    -filter:fixedweight <n> <val>         Apply the specified standard deviation to the measurement 'n', to be used as weight in the filter.\n");
	printf("                            <n>    = Measurement number\n");
	printf("                            <val>  = Standard deviation of the measurement (m)\n");
	printf("                            In particular, the filter shall apply as weight = 1/(<val>^2)\n");
	printf("                            Sample: '-filter:fixedweight 1 2'    Set 2 meters of standard deviation to measurement 1 in filter\n");
	printf("                                    '-filter:fixedweight 2 0.01' Set 1 centimeter of standard deviation to measurement 2 in filter\n");
	printf("                            Defaults: PPP:         Pseudorange->1m  Carrierphase->0.01m\n");
	printf("                                      SPP:         Pseudorange->5m  Carrierphase->0.10m\n");
	printf("    -filter:elevweight <n> <a> <b> <c>    Apply the specified values to compute the standard deviation of the measurement 'n'\n");
	printf("                            std = a + b * e^(-elevation/c)\n");
	printf("                            <n>    = Measurement number\n");
	printf("                            <a>    = Minimum standard deviation of the measurement [weight at elevation 90º] (m)\n");
	printf("                            <b>    = Multiplier to e [standard deviation at elevation 0º] (m)\n");
	printf("                            <c>    = Elevation constant (degrees)\n");
	printf("                            Again, the filter shall apply as weight = 1/(std^2)\n");
	printf("                            Sample: '-filter:elevweight 1 0.3 8 10'   Apply the standard deviation: 0.3+8*e^(-elevation/10)\n");
	printf("                            Note: '-filter:elevweight 1 x 0 y' equals to '-filter:fixedweight 1 x'\n");
	printf("    -filter:phi:dr <val>    Specify the Phi value for position unknowns [defaults static:1 kinematic:0]\n");
	printf("    -filter:phi:clk <val>   Specify the Phi value for clock unknown [default 0]\n");
	printf("    -filter:phi:trop <val>  Specify the Phi value for troposphere unknown [default 1]\n");
	printf("    -filter:phi:amb <val>   Specify the Phi value for ambiguity unknowns [default 1]\n");
	printf("    -filter:q:dr <val>      Specify the Q noise value for position unknowns [defaults static:0 kinematic:1e8] (m^2)\n");
	printf("    -filter:q:clk <val>     Specify the Q noise value for clock unknown [default 9e10] (m^2)\n");
	printf("    -filter:q:trop <val>    Specify the Q noise variation value for troposphere unknown [default 1e-4] (m^2/h)\n");
	printf("    -filter:q:amb <val>     Specify the Q noise value for ambiguity unknowns [default 0] (m^2)\n");
	printf("    -filter:p0:dr <val>     Specify the P0 initial value for position unknowns [default 1e8] (m^2)\n");
	printf("    -filter:p0:clk <val>    Specify the P0 initial value for clock unknown [default 9e10] (m^2)\n");
	printf("    -filter:p0:trop <val>   Specify the P0 initial value for troposphere unknown [default 0.5^2] (m^2)\n");
	printf("    -filter:p0:amb <val>    Specify the P0 initial value for ambiguity unknowns (for prealigned carrier phases) [default 20^2] (m^2)\n");
	printf("    -filter:backward        Specify that the filter does a backward processing after the forward one is finished. This meas that it\n");
	printf("                            process the data backwards. The \"turn point\" is defined as the latest point where orbits and clocks\n");
	printf("                            are available or when the observation RINEX ends (whatever is first) ('--filter:backward' to disable it)\n");
	printf("                            [default disabled]\n");
	printf("\n  OUTPUT OPTIONS\n");
	printf("    -output:file <file>     Sets the output file [default stdout]\n");
	printf("    --output:file           Sets the output to stdout [default]\n");
	printf("    -output:satvel <val>    <val> = inertial  Prints the inertial velocity in the messages where satellite velocity is given\n");
	printf("                            <val> = ITRF      Prints the ITRF velocity in the messages where satellite velocity is given\n");
	printf("\n  VERBOSE OPTIONS   (use -print:... to activate, --print:... to deactivate)\n");
	printf("    -print:info             Print INFO messages [default on]\n");
	printf("    -print:cycleslips       Print CS messages [default on]\n");
	printf("    -print:input            Print INPUT messages [default on]\n");
	printf("    -print:meas             Print MEAS messages [default off]\n");
	printf("    -print:model            Print MODEL messages [default on]\n");
	printf("    -print:satellites       Print EPOCHSAT messages [default on]\n");
	printf("    -print:prefit           Print PREFIT messages [default on]\n");
	printf("    -print:postfit          Print POSTFIT messages [default on]\n");
	printf("    -print:filter           Print FILTER messages [default on]\n");
	printf("    -print:output           Print OUTPUT messages [default on]\n");
	printf("    -print:satdiff          Print SATDIFF messages in comparison mode (see below) [default on]\n");
	printf("    -print:satstat          Print SATSTAT messages in comparison mode (see below) [default on]\n");
	printf("    -print:satstattot       Print SATSTATTOT messages in comparison mode (see below) [default on]\n");
	printf("    -print:satpvt           Print SATPVT messages in show product mode (see below) [default on]\n");
	printf("    -print:satsel           Print satellite selection information [default off]\n");
	printf("    -print:all              Print all messages\n");
	printf("    -print:none             Do not print anything\n");
	printf("                            More information on print messages can be seen with the '-messages' option\n");
	printf("\n  WORK MODES\n");
	printf("    gLAB can work in four different modes:\n");
	printf("        - Positioning Mode: 'Standard' mode, where all the processing is done, and a solution for a\n");
	printf("                       receiver is provided as OUTPUT messages. The minimum parameters required for this\n");
	printf("                       mode are an input observation file ('-input:obs') and orbit and clock products\n");
	printf("                       ('-input:nav', '-input:SP3' or '-input:orb'/'-input:clk'). Using precise products\n");
	printf("                       will also require the use of an ANTEX file ('-input:ant')\n");
	printf("        - Show Input Mode: This mode only reads an input RINEX observation file and print its measurements.\n");
	printf("                       The parameter required for this mode is '-input:obs', and specifically, no orbit and\n");
	printf("                       clock products should be provided (if provided, gLAB will switch to Positioning Mode)\n"); 
	printf("        - Product Comparison Mode: This mode reads and compares two different sources of orbit and clock products.\n");
	printf("                       In order to use this mode, '-input:obs' must be avoided, and two different orbit and clock\n");
	printf("                       products should be provided. This mode outputs the SATDIFF, SATSTAT and STASTATTOT messages\n");
	printf("        - Show Product Mode: This mode reads a single source of orbit and clock products.\n");
	printf("                       In order to use this mode, '-input:obs' must be avoided, and a single orbit and clock\n");
	printf("                       product should be provided. This mode output SAT messages\n"); 
}

/*****************************************************************************
 * Name        : showExamples
 * Description : Shows the program examples in the standard output
 * Parameters  : None
 *****************************************************************************/
void showExamples () {
	printf("gLAB version v%s\n",VERSION);
	printf("GNSS LABORATORY. Tool to analyse GNSS data\n");
	printf("Write 'gLAB -help' for help on the options\n");
	printf("Usage:\n");
	printf("    gLAB [options]\n");
	printf("Examples:\n");
	printf("    o Typical PPP:\n");
	printf("        gLAB -input:obs madr2000.06o -input:sp3 igs13843.sp3 -input:ant igs_pre1400.atx -filter:meas carrierphase -filter:nav static\n");
	printf("    o Same as before using the defaults:\n");
	printf("        gLAB -input:obs madr2000.06o -input:sp3 igs13843.sp3 -input:ant igs_pre1400.atx\n");
	printf("    o Typical SPP:\n");
	printf("        gLAB -input:obs madr2000.06o -input:nav brdc2000.06n\n");
	printf("    o PPP in static mode with pseudorange only decimating to 900 seconds:\n");
	printf("        gLAB -input:obs madr2000.06o -input:sp3 igs13843.sp3 -input:ant igs_pre1400.atx -filter:meas pseudorange -filter:nav static -pre:dec 900\n");
	printf("    o Typical PPP without INFO messages:\n");
	printf("        gLAB -input:obs madr2000.06o -input:sp3 igs13843.sp3 -input:ant igs_pre1400.atx --print:info\n"); 
	printf("    o Typical PPP with some verbose options:\n");
	printf("        gLAB -input:obs madr2000.06o -input:sp3 igs13843.sp3 -input:ant igs_pre1400.atx -print:none -print:cycleslips -print:model -print:satellites\n");
	printf("    o Typical standalone using standard ionospheric model (Klobuchar) and not using it\n");
	printf("        gLAB -input:obs madr2000.06o -input:nav brdc2000.06n -filter:nav kinematic\n");
	printf("        gLAB -model:iono no -input:obs madr2000.06o -input:nav brdc2000.06n -filter:nav kinematic\n");
	printf("    o Save output to a file:\n");
	printf("        gLAB [options] > outputfile\n");
	printf("        gLAB [options] -output:file outputfile\n");
	printf("    o Print only INPUT messages wihout any decimation, do not process:\n");
	printf("      (no -input:sp3 nor -input:nav nor -input:orb/-input:clk)\n");
	printf("        gLAB -input:obs madr2000.06o -pre:dec 0 -print:none -print:input\n");
}

/*****************************************************************************
 * Name        : showConfig
 * Description : Shows information on the config file
 * Parameters  : None
 *****************************************************************************/
void showConfig () {
	printf("gLAB version v%s\n",VERSION);
	printf("GNSS LABORATORY. Tool to analyse GNSS data\n");
	printf("Write 'gLAB -help' for help on the options\n");
	printf("Usage:\n");
	printf("    gLAB [options]\n");
	printf("\n  CONFIG FILE\n");
	printf("    The config file can be specified by the '-input:cfg <file>' command-line parameter. This file is an alternative\n");
	printf("    way to configure the behaviour of the tool. In case that there are parameters in the config file and in the\n");
	printf("    command-line arguments, the command-line parameters will have priority over the config file ones. The order to\n");
	printf("    read the parameters is:\n");
	printf("        o Check if the '-input:cfg' parameter exist in the command-line arguments\n");
	printf("        o If it exists, parse config file parameters\n");
	printf("        o Parse command-line parameters\n");
	printf("    The config file is basically a mirror of the command-line parameters. The way to include an option is as follows:\n");
	printf("        <parameter> <options> of the parameter\n");
	printf("    For example, for the parameters '-input:obs upc42000_10s.06o' and '-filter:select 2 PC LC', the config file shall be:\n");
	printf("        -input:obs upc42000_10s.06o\n");
	printf("        -filter:select 2 PC LC\n");
	printf("    All lines beginning with # in the config file shall be interpreted as a comment, and ignored in the parsing process\n");
	printf("\n  SAMPLE CONFIG FILE\n");
	printf("        # INPUT\n");
	printf("        -input:obs madr2000.06o\n");
	printf("        -input:nav brdc2000.06n\n");
	printf("        -input:ant igs_pre1400.atx\n\n");
	printf("        # PREPROCESS\n");
	printf("        -pre:elevation 5\n\n");
	printf("        # MODEL\n");
	printf("        -model:recphasecenter ANTEX\n\n");
	printf("        # FILTER\n");
	printf("        -filter:meas carrierphase\n");
	printf("        -filter:nav kinematic\n\n");
	printf("        # OUTPUT\n");
	printf("        -print:all\n\n");
	printf("    This config file will be equivalent to the command-line execution:\n");
	printf("        gLAB -input:obs madr2000.06o -input:nav brdc2000.06n -input:ant igs_pre1400.atx -pre:elevation 5 -model:recphasecenter ANTEX -filter:meas carrierphase -filter:nav kinematic -print:all\n");
}

/*****************************************************************************
 * Name        : showMessages
 * Description : Shows the fields of each output message
 * Parameters  : None
 *****************************************************************************/
 void showMessages() {
 	printf("gLAB version v%s\n",VERSION);
	printf("GNSS LABORATORY. Tool to analyse GNSS data\n");
	printf("Write 'gLAB -help' for help on the options\n");
	printf("Usage:\n");
	printf("    gLAB [options]\n");
	printf("\n  MESSAGES FORMAT\n");
	printf("    INFO\n");
	printf("        Variable format, depending on the specific information that should be displayed\n");
	printf("    CS\n");
	printf("        Cycle-slip message information. It is shown when a cycle-slip is found by any detector.\n");
	printf("        The first 5 fields are fixed, afterwards, there are three possible blocks depending on\n");
	printf("        the cycle-slip detector activated.\n");
	printf("        Field 1:  'CS'\n");
	printf("        Field 2:  Year\n");
	printf("        Field 3:  Doy\n");
	printf("        Field 4:  Seconds of day\n");
	printf("        Field 5:  GNSS System (GPS, GAL, GLO or GEO)\n");
	printf("        Field 6:  PRN satellite identifier\n");
	printf("        Block 1:  Geometric-free carrier phase cycle-slip detector (Li)\n");
	printf("            Field 1: 'Li:x', being x, 0 or 1 (0-> Li did not detect the cycle-slip, 1-> Li did)\n");
	printf("            Field 2: Li value\n");
	printf("            Field 3: Li Threshold\n");
	printf("        Block 2:  Melbourne-Wübbena cycle-slip detector (Bw)\n");
	printf("            Field 1: 'Bw:x', being x, 0 or 1 (0-> Bw did not detect the cycle-slip, 1-> Bw did)\n");
	printf("            Field 2: Bw value\n");
	printf("            Field 3: Bw Threshold\n");
	printf("        Block 3:  L1-C1 average cycle-slip detector (L1C1)\n");
	printf("            Field 1: 'L1C1:x', being x, 0 or 1 (0-> L1C1 did not detect the cycle-slip, 1-> L1C1 did)\n");
	printf("            Field 2: L1C1 value\n");
	printf("            Field 3: L1C1 Threshold\n");
	printf("        Sample:\n");
	printf("            CS 2006 200  4500.00 GPS 27 Li:1 (0.066968 [0.052100]) Bw:0 (0.360430 [5.775316])\n");
	printf("        If all the measurements for a specific cycle-slip does not exist the number inside code\n");
	printf("        NOT_ENOUGH_MEASUREMENTS will be included inside the '()'\n");
	printf("    INPUT\n");
	printf("        Input data message. It is shown after an epoch is read (and decimated). It contains the\n");
	printf("        measurements for each satellite for this epoch\n");
	printf("        Field 1:  'INPUT'\n");
	printf("        Field 2:  Year\n");
	printf("        Field 3:  Doy\n");
	printf("        Field 4:  Seconds of day\n");
	printf("        Field 5:  GNSS System (GPS, GAL, GLO or GEO)\n");
	printf("        Field 6:  PRN satellite identifier\n");
	printf("        Field 7:  Arc length (number of undecimated epochs after the last cycle-slip)\n");
	printf("        For GPS:\n");
	printf("            Field 8:  C1 [C1C]\n");
	printf("            Field 9:  P1 [C1P]\n");
	printf("            Field 10: P2 [C2P]\n");
	printf("            Field 11: L1 [L1P] (prealigned, in meters)\n");
	printf("            Field 12: L2 [L2P] (prealigned, in meters)\n");
	printf("        For Galileo (GAL):\n");
	printf("            Field 8:  C1A\n");
	printf("            Field 9:  C1B\n");
	printf("            Field 10: C1C\n");
	printf("            Field 11: C7Q\n");
	printf("            Field 12: C8Q\n");
	printf("            Field 13: L1A (prealigned, in meters)\n");
	printf("            Field 14: L1B (prealigned, in meters)\n");
	printf("            Field 15: L1C (prealigned, in meters)\n");
	printf("            Field 16: L7Q (prealigned, in meters)\n");
	printf("            Field 17: L8Q (prealigned, in meters)\n");
	printf("        For GLONASS (GLO):\n");
	printf("            Field 8:  C1 [C1C]\n");
	printf("            Field 9:  C2 [C2C]\n");
	printf("            Field 10: L1 [L1P](prealigned, in meters)\n");
	printf("            Field 11: L2 [L2P](prealigned, in meters)\n");
	printf("        For GEO:\n");
	printf("            Field 8:  C1 [C1C]\n");
	printf("        Sample:\n");
	printf("            INPUT 2006 200     0.00 GPS 19      1  23119003.9020  23119002.6110  23119004.0750  23119002.7507  23119004.0925\n");
	printf("    MEAS\n");
	printf("        Information of the input measurements in a variable format, and azimuth/elevation of the satellite.\n");
	printf("        It is written after modelling the message.\n");
	printf("        Field 1:   'MEAS'\n");
	printf("        Field 2:   Year\n");
	printf("        Field 3:   Doy\n");
	printf("        Field 4:   Seconds of day\n");
	printf("        Field 5:   GNSS System (GPS, GAL, GLO or GEO)\n");
	printf("        Field 6:   PRN satellite identifier\n");
	printf("        Field 7:   Elevation of the satellite [degrees]\n");
	printf("        Field 8:   Azimuth of the satellite [degrees]\n");
	printf("        Field 9:   Number of measurements\n");
	printf("        Field 10:  List of measurements included\n");
	printf("        Field 11-: Each following column is the value of the measurement specified in Field 10\n");
	printf("        If there is no data available on azimuth and elevation, the fields 7 and 8 will be fixed to zero.\n");
	printf("        Sample:\n");
	printf("            MEAS     2006 200     0.00 GPS 19   26.66  -79.62  6  C1C:L1C:C1P:L1P:C2P:L2P  23119003.9020         0.0000  23119002.6110  23119002.7507  23119004.0750  23119004.0925\n");
	printf("    MODEL\n");
	printf("        Model break down message. It is shown when a model can be fully computed for each measurement.\n");
	printf("        Field 1:  'MODEL'\n");
	printf("        Field 2:  Year\n");
	printf("        Field 3:  Doy\n");
	printf("        Field 4:  Seconds of day\n");
	printf("        Field 5:  GNSS System (GPS, GAL, GLO or GEO)\n");
	printf("        Field 6:  PRN satellite identifier\n");
	printf("        Field 7:  Measurement identifier (as string)\n");
	printf("        Field 8:  Signal flight time [sec]\n");
	printf("        Field 9:  Measured value [m]\n");
	printf("        Field 10: Full model value [m]\n");
	printf("        Field 11: Satellite X position [m]\n");
	printf("        Field 12: Satellite Y position [m]\n");
	printf("        Field 13: Satellite Z position [m]\n");
	printf("        Field 14: Satellite X velocity [m]\n");
	printf("        Field 15: Satellite Y velocity [m]\n");
	printf("        Field 16: Satellite Z velocity [m]\n");
	printf("        Field 17: Satellite-receiver geometric distance [m]\n");
	printf("        Field 18: Satellite clock correction [m]\n");
	printf("        Field 19: Satellite phase center projection [m]\n");
	printf("        Field 20: Receiver phase center projection [m]\n");
	printf("        Field 21: Receiver Antenna Reference Point (ARP) projection [m]\n");
	printf("        Field 22: Relativity correction [m]\n");
	printf("        Field 23: Wind-up correction [m] (for carrier phase measurements)\n");
	printf("        Field 24: Troposphere nominal correction [m]\n");
	printf("        Field 25: Ionosphere correction [m]\n");
	printf("        Field 26: Gravitational delay correction [m]\n");
	printf("        Field 27: Total Group Delay (TGD) correction [m]\n");
	printf("        Field 28: Solid tides correction [m]\n");
        printf("        Field 29: Satellite elevation [degrees]\n");
        printf("        Field 30: Satellite azimuth [degrees]\n");
	printf("        The satellite coordinates (fields 11-16) are given in the reception epoch if the model of the\n");
	printf("        signal transmission time is enabled (if it is disabled, they are given in the transmission\n");
	printf("        time). The coordinates are relative to the antenna phase center or satellite mass center,\n");
	printf("        depending on the products input:\n");
	printf("			* Broadcast: Antenna Phase Center\n");
	printf("            * Precise: Satellite Mass Center\n");
	printf("        Field 9 is the direct measurement (as in the RINEX file), but scaled to meters for carrier phase\n");
	printf("        measurements. Field 10 is the model computed for this measurement. Field 10 is the direct sum\n");
	printf("        of fields 17 to 28.\n");
	printf("        Sample:\n");
	printf("            MODEL 2006 200     0.00 GPS 19 L1P     0.07712  23119002.7507  23119008.7501   8811456.7780 -21033910.1687  13675922.8867      1828.7339      2353.7679      2467.3576  23119457.7652     -456.31787   0.00000  -0.04936  -0.01140   2.32333   0.10671   4.85412  -0.00000   0.01544   0.00000   0.06394   9.164  79.274\n");
	printf("    EPOCHSAT\n");
	printf("        Message with the satellites used to compute the solution. It is shown when the filter is run,\n");
	printf("        and is given for each measurement\n");
	printf("        Field 1:  'EPOCHSAT'\n");
	printf("        Field 2:  Year\n");
	printf("        Field 3:  Doy\n");
	printf("        Field 4:  Seconds of day\n");
	printf("        Field 5:  Measurement identifier (as string)\n");
	printf("        Field 6:  Number of satellites in view\n");
	printf("        Field 7-: Each following column is the identifier of a satellite\n");
	printf("        Sample:\n");
	printf("            EPOCHSAT 2006 200   300.00  PC  7 15  3 19 16 18 21 22\n");
	printf("    SATSEL\n");
	printf("        Message with debug information of the reason why a satellite has been discarded (or selected) for processing.\n");
	printf("        Sample:\n");
	printf("            SATSEL  7 discarded: Elevation too low ( 3.04 )\n");
	printf("    PREFIT\n");
	printf("        Prefilter values message. It provides the measurement-model values. It is shown in each filter\n");
	printf("        execution.\n");
	printf("        Field 1:  'PREFIT' (if the satellite is used in the computation) or 'PREFIT*' (if it is not)\n");
	printf("        Field 2:  Year\n");
	printf("        Field 3:  Doy\n");
	printf("        Field 4:  Seconds of day\n");
	printf("        Field 5:  GNSS System (GPS, GAL, GLO or GEO)\n");
	printf("        Field 6:  PRN satellite identifier\n");
	printf("        Field 7:  Measurement identifier (as string)\n");
	printf("        Field 8:  Measurement-model value (prefit) [m]\n");
	printf("        Field 9:  Measurement value [m]\n");
	printf("        Field 10: Model value [m]\n");
	printf("        Field 11: X-partial derivative (X-component of the satellite line-of-sight vector)\n");
	printf("        Field 12: Y-partial derivative (Y-component of the satellite line-of-sight vector)\n");
	printf("        Field 13: Z-partial derivative (Z-component of the satellite line-of-sight vector)\n");
	printf("        Field 14: T-partial derivative \n");
	printf("        Field 15: Elevation of the satellite [degrees]\n");
	printf("        Field 16: Azimuth of the satellite [degrees]\n");
	printf("        Field 17: Standard deviation of the measurement  (for the filter) [m]\n");
	printf("        In general Field 8 = Field 9 - Field 10, but this is no longer true when using smoothing, as the\n");
	printf("        Field 9 is the raw measurement without smoothing, but the Field 8 computation takes smoothing into\n");
	printf("        account.\n");
	printf("        Sample:\n");
	printf("            PREFIT   2006 200   300.00 GPS 19  LC        -7.3029  22982271.7155  22982279.0184   28.28   0.3931  -0.4834   0.7822   1.0000   77.912   130.010\n");
	printf("    POSTFIT\n");
	printf("        Postfilter values message. It provides the corrected prefits with the filter estimation. It is\n");
	printf("        shown in each filter execution.\n");
	printf("        Field 1:  'POSTFIT'\n");
	printf("        Field 2:  Year\n");
	printf("        Field 3:  Doy\n");
	printf("        Field 4:  Seconds of day\n");
	printf("        Field 5:  GNSS System (GPS, GAL, GLO or GEO)\n");
	printf("        Field 6:  PRN satellite identifier\n");
	printf("        Field 7:  Measurement identifier (as string)\n");
	printf("        Field 8:  Measurement-corrected model value (postfit) [m]\n");
	printf("        Field 9:  Measurement value [m]\n");
	printf("        Field 10: Corrected model value with the filter estimations[m]\n");
	printf("        Field 11: Elevation of the satellite [degrees]\n");
	printf("        Field 12: Azimuth of the satellite [degrees]\n");
	printf("        Field 13: Only given for carrier phase measurements. It is the estimated carrier phase ambiguity. [m]\n");
	printf("        As with PREFIT messages, Field 8 is not necessary Field 9 - Field 10 when using smoothing.\n");
	printf("        Sample:\n");
	printf("            POSTFIT  2006 200   300.00 GPS 19  LC         0.0000  22982271.7155  22982271.7155   77.912   130.010   0.3029\n");
	printf("    FILTER\n");
	printf("        Filter solution message. This message provides direct information on the filter estimates. It is shown\n");
	printf("        in each filter execution.\n");
	printf("        Field 1:  'FILTER'\n");
	printf("        Field 2:  Year\n");
	printf("        Field 3:  Doy\n");
	printf("        Field 4:  Seconds of day\n");
	printf("        Field 5-: Filter estimates. The order is: 3D estimated position, clock, troposphere and ambiguities\n"); 
	printf("        The number of fields is variable in this message. With a full filter (troposphere and ambiguities\n");
	printf("        estimation), the fields are as follows:\n");
	printf("            Field 5:  Receiver X position [m]\n");
	printf("            Field 6:  Receiver Y position [m]\n");
	printf("            Field 7:  Receiver Z position [m]\n");
	printf("            Field 8:  Receiver clock [m]\n");
	printf("            Field 9:  Zenith Tropospheric Delay [m]\n");
	printf("            Field 10: Carrierphase ambiguities [m]\n");
	printf("        Sample:\n");
	printf("            FILTER 2006 200   300.00   4849203.0770   -360328.5730   4114913.9184        -7.4867         2.1946        -0.0001         1.1079        -1.1073         0.3029         0.4555         0.0897         0.0001         0.3845         0.2136\n");
	printf("    OUTPUT\n");
	printf("        Receiver solution message. This message provides the estimated receiver position. It is shown in\n");
	printf("        each filter execution.\n");
	printf("        Field 1:  'OUTPUT'\n");
	printf("        Field 2:  Year\n");
	printf("        Field 3:  Doy\n");
	printf("        Field 4:  Seconds of day\n");
	printf("        Field 5:  Square root of the sum of the covariance matrix. This is a measure of the convergence of the filter\n");
	printf("        Field 6:  Receiver X position [m]\n");
	printf("        Field 7:  Receiver Y position [m]\n");
	printf("        Field 8:  Receiver Z position [m]\n");
	printf("        Field 9:  Receiver X position - Nominal a priori X position [m]\n");
	printf("        Field 10: Receiver Y position - Nominal a priori Y position [m]\n");
	printf("        Field 11: Receiver Z position - Nominal a priori Z position [m]\n");
	printf("        Field 12: Receiver X formal error [m]\n");
	printf("        Field 13: Receiver Y formal error [m]\n");
	printf("        Field 14: Receiver Z formal error [m]\n");
	printf("        Field 15: Receiver latitude [degrees]\n");
	printf("        Field 16: Receiver longitude [degrees]\n");
	printf("        Field 17: Receiver height [m]\n");
	printf("        Field 18: Receiver North difference in relation to nominal a priori position [m]\n");
	printf("        Field 19: Receiver East difference in relation to nominal a priori position [m]\n");
	printf("        Field 20: Receiver Up difference in relation to nominal a priori position [m]\n");
	printf("        Field 21: Receiver formal error in North direction [m]\n");
	printf("        Field 22: Receiver formal error in East direction [m]\n");
	printf("        Field 23: Receiver formal error in Up direction [m]\n");
	printf("        Field 24: Geometric Dilution of Precision (GDOP)\n");
	printf("        Field 25: Positioning Dilution of Precision (PDOP)\n");
	printf("        Field 26: Time Dilution of Precision (TDOP)\n");
	printf("        Field 27: Horizontal Dilution of Precision (HDOP)\n");
	printf("        Field 28: Vertical Dilution of Precision (VDOP)\n");
	printf("        Field 29: Zenith Tropospheric Delay (including nominal value) [m]\n");
	printf("        Field 30: Zenith Tropospheric Delay (excluding nominal value) [m]\n");
	printf("        Field 31: Zenith Tropospheric Delay formal error [m]\n");
	printf("        The nominal a priori position is a prerequisite of the processing, and can be specified or read from the RINEX\n");
	printf("        input file. See the option '-pre:setrecpos' in the help ('-help') for more details.\n");
	printf("        Sample:\n");
	printf("            OUTPUT 2006 200   300.00    2.6219   4849203.1236   -360328.5229   4114913.9535     0.7693     0.4145     0.7580     1.9353     0.6998     1.6246   40.429162956   -4.249653155  830.480629026     0.0993     0.4704     1.0522     1.1365     0.6772     2.2637     0.9572     12.2743     5.0975     7.2205     3.5866     3.0880     0.0097     0.4995\n");
	printf("    SATDIFF [orbit and clock comparison]\n");
	printf("        Satellite difference message for products comparison mode. This message shows the difference between\n");
	printf("        both orbit and clocks for two different OC products fonts for a specific time. It is shown each time step.\n");
	printf("        The SISRE fields are the Signal-in-Space Range Error, it is a measure of how the error impact a general\n");
	printf("        receiver found in the Earth's surface. It is computed as:\n");
	printf("            SISRE = sqrt((radDiff-clkDiff)^2 + 1/49*(ctDiff^2 + atDiff^2))\n");
	printf("        Being:\n");
	printf("            radDiff, the difference of satellite positions in the radial component\n");
	printf("            clkDiff, the difference of satellite clock offsets\n");
	printf("            ctDiff, the difference of satellite positions in the cross-track component\n");
	printf("            atDiff, the difference of satellite positions in the along-track component\n");
	printf("        Field 1:   'SATDIFF'\n");
	printf("        Field 2:   Year\n");
	printf("        Field 3:   Doy\n");
	printf("        Field 4:   Seconds of day\n");
	printf("        Field 5:   GNSS System (GPS, GAL, GLO or GEO)\n");
	printf("        Field 6:   PRN satellite identifier\n");
	printf("        Field 7:   SISRE difference\n");
	printf("        Field 8:   SISRE orbit-only difference\n");
	printf("        Field 9:   3D orbit difference\n");
	printf("        Field 10:  Clock difference\n");
	printf("        Field 11:  Radial position difference\n");
	printf("        Field 12:  Along-track position difference\n");
	printf("        Field 13:  Cross-track position difference\n");
	printf("        Sample:\n");
	printf("            SATDIFF 2006 200     0.00 GPS  1     4.1082     0.5359     1.1019    -3.5883     0.5176     0.5497     0.8025\n");
	printf("    SATSTAT [orbit and clock comparison]\n");
	printf("        Satellite difference summary message for products comparison mode. This message shows the differences\n");
	printf("        between orbit and clocks for two different OC products fonts for a specific satellite. It is shown after\n");
	printf("        all the comparisons are done, and presents a summary for each satellite.\n");
	printf("        Field 1:      'SATSTAT'\n");
	printf("        Field 2:      GNSS System (GPS, GAL, GLO or GEO)\n");
	printf("        Field 3:      PRN satellite identifier\n");
	printf("        Field 4:      Number of samples\n");
	printf("        Field 5-7:    SISRE comparisons  [mean,RMS,stdDev]\n");
	printf("        Field 8-10:   SISRE orbit-only comparisons  [mean,RMS,stdDev]\n");
	printf("        Field 11-13:  3D orbit comparisons  [mean,RMS,stdDev]\n");
	printf("        Field 14-16:  Clock comparisons  [mean,RMS,stdDev]\n");
	printf("        Field 17-19:  Radial position comparisons  [mean,RMS,stdDev]\n");
	printf("        Field 20-22:  Along-track position comparisons  [mean,RMS,stdDev]\n");
	printf("        Field 23-25:  Cross-track position comparisons  [mean,RMS,stdDev]\n");
	printf("        Sample:\n");
	printf("            SATSTAT  1   95     1.5530     2.1112     1.4302     0.9279     1.0058     0.3881     1.4115     1.4714     0.4155    -0.3357     1.9280     1.8985     0.8988     0.9938     0.4239     0.6081     0.9604     0.7433     0.1717     0.5050     0.4750\n");
	printf("    SATSTATTOT [orbit and clock comparison]\n");
	printf("        Difference summary message for products comparison mode. This message shows the statistics of the\n");
	printf("        differences between orbit and clocks for two different OC products fonts. It is shown after all the\n");
	printf("        comparisons are done, and presents a summary for all of them.\n");
	printf("        Field 1:      'SATSTATTOT'\n");
	printf("        Field 2:      Number of samples\n");
	printf("        Field 3-5:    SISRE comparisons  [mean,RMS,stdDev]\n");
	printf("        Field 6-8:    SISRE orbit-only comparisons  [mean,RMS,stdDev]\n");
	printf("        Field 9-11:   3D orbit comparisons  [mean,RMS,stdDev]\n");
	printf("        Field 12-14:  Clock comparisons  [mean,RMS,stdDev]\n");
	printf("        Field 15-17:  Radial position comparisons  [mean,RMS,stdDev]\n");
	printf("        Field 18-20:  Along-track position comparisons  [mean,RMS,stdDev]\n");
	printf("        Field 21-23:  Cross-track position comparisons  [mean,RMS,stdDev]\n");
	printf("        Sample:\n");
	printf("            SATSTATTOT 2660     1.2971     1.6206     0.9714     1.0207     1.1478     0.5250     1.6623     1.7701     0.6083    -0.1909     1.3835     1.3702     0.9488     1.1312     0.6160     0.0514     1.1930     1.1918    -0.0345     0.6560     0.6551\n");
	printf("    SATPVT [show product mode]\n");
	printf("        Satellite data messages for show product mode. This message show the PVT data for a specific time\n");
	printf("        and satellite.\n");
	printf("        Field 1:      'SATPVT'\n");
	printf("        Field 2:      Year\n");
	printf("        Field 3:      Doy\n");
	printf("        Field 4:      Seconds of day\n");
	printf("        Field 5:      GNSS System (GPS, GAL, GLO or GEO)\n");
	printf("        Field 6:      PRN satellite identifier\n");
	printf("        Field 7:      Satellite X position [m]\n");
	printf("        Field 8:      Satellite Y position [m]\n");
	printf("        Field 9:      Satellite Z position [m]\n");
	printf("        Field 10:     Satellite X velocity [m/s]\n");
	printf("        Field 11:     Satellite Y velocity [m/s]\n");
	printf("        Field 12:     Satellite Z velocity [m/s]\n");
	printf("        Field 13:     Satellite clock offset [m]\n");
	printf("        Sample:\n");
	printf("            SATPVT 2006 199 79500.00 GPS  1  15340446.1662 -20066066.3778  -7929249.6479      1191.6479      2132.5615     -3023.0448     19826.4736\n");
}

/*****************************************************************************
 * Name        : parseOption
 * Description : Parse an option from a vector of strings into the options or
 *               support options
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TSupportOptions  *sopt          O  N/A  TSupportOptions structure
 * TOptions  *options              O  N/A  TOptions structure
 * char  *argv[]                   I  N/A  Vector of strings to be parsed
 * Returned value (int)            O  N/A  Status of the function
 *                                         -3 => Missing second parameter                                        
 *                                         -2 => Parameter value not valid
 *                                         -1 => Parameter name unknown
 *                                         0  => Help/examples message (end program)
 *                                         >0 => Number of strings parsed
 *****************************************************************************/
int parseOption (TSupportOptions *sopt, TOptions *options, char *argv[]) {
	int		aux1,aux2;
	double	auxd;
	int		i;
	char	auxstr[10];
	
	if (strcmp(argv[0],"-help")==0) {
		showHelp();
		return 0;
	} else if (strcmp(argv[0],"-examples")==0) {
		showExamples();
		return 0;
	} else if (strcmp(argv[0],"-config")==0) {
		showConfig();
		return 0;
	} else if (strcmp(argv[0],"-messages")==0) {
		showMessages();
		return 0;
	} else if (strcmp(argv[0],"-output:file")==0) {
		if (argv[1]==NULL) return -3;
		strcpy(sopt->outFile,argv[1]);
		return 2;
	} else if (strcmp(argv[0],"--output:file")==0) {
		sopt->outFile[0]='\0';
		return 1;
	} else if (strcmp(argv[0],"-output:satvel")==0) {
		if (argv[1]==NULL) return -3;
		if (strcasecmp(argv[1],"inertial")==0) {
			options->satVel = svINERTIAL;
			return 2;
		} else if (strcasecmp(argv[1],"itrf")==0) {
			options->satVel = svITRF;
			return 2;
		} else return -2;
	} else if (strcmp(argv[0],"-input:cfg")==0) {
		strcpy(sopt->cfgFile,argv[1]);
		return 2;
	} else if (strcmp(argv[0],"-input:obs")==0) {
		strcpy(sopt->obsFile,argv[1]);
		return 2;
	} else if (strcmp(argv[0],"-input:nav")==0) {
		strcpy(sopt->navFile[sopt->numNav],argv[1]);
		sopt->numNav++;
		return 2;
	} else if (strcasecmp(argv[0],"-input:sp3")==0) {
		strcpy(sopt->SP3File[sopt->numSP3],argv[1]);
		sopt->numSP3++;
		return 2;
	} else if (strcmp(argv[0],"-input:con")==0) {
		strcpy(sopt->conFile,argv[1]);
		return 2;
	} else if (strcmp(argv[0],"-input:orb")==0) {
		strcpy(sopt->orbFile[sopt->numOrb],argv[1]);
		sopt->numOrb++;
		return 2;
	} else if (strcmp(argv[0],"-input:clk")==0) {
		strcpy(sopt->clkFile[sopt->numClk],argv[1]);
		sopt->numClk++;
		return 2;
	} else if (strcmp(argv[0],"-input:ant")==0) {
		strcpy(sopt->antFile,argv[1]);
		return 2;
	} else if (strcmp(argv[0],"-input:klb")==0) {
		strcpy(sopt->klbFile,argv[1]);
		return 2;
	} else if (strcmp(argv[0],"-input:dcb")==0) {
		strcpy(sopt->tgdFile[sopt->numTGD],argv[1]);
		sopt->numTGD++;
		return 2;
	} else if (strcmp(argv[0],"-input:rec")==0) {
		strcpy(sopt->recFile,argv[1]);
		return 2;
	} else if (strcmp(argv[0],"-input:snx")==0) {
		strcpy(sopt->snxFile,argv[1]);
		return 2;
	} else if (strcmp(argv[0],"-filter:nav")==0) {
		if (argv[1]==NULL) return -3;
		if (strcmp(argv[1],"static")==0) {
			sopt->navKinematics = NKstatic;
		} else if (strcmp(argv[1],"kinematic")==0) {
			sopt->navKinematics = NKkinematic;
		} else return -2;
		return 2;
	} else if (strcmp(argv[0],"-pre:elevation")==0) {
		if (argv[1]==NULL) return -3;
		options->elevationMask = atof(argv[1])*d2r;
		return 2;
	} else if (strcmp(argv[0],"-pre:eclipse")==0) {
		sopt->discardEclipsedSatellites = 1;
		return 1;
	} else if (strcmp(argv[0],"--pre:eclipse")==0) {
		sopt->discardEclipsedSatellites = 0;
		return 1;
	} else if (strcmp(argv[0],"-pre:dec")==0) {
		if (argv[1]==NULL) return -3;
		options->decimate = atoi(argv[1]);
		return 2;
	} else if (strcmp(argv[0],"-pre:prealign")==0) {
		options->prealignCP = 1;
		return 1;
	} else if (strcmp(argv[0],"--pre:prealign")==0) {
		options->prealignCP = 0;
		return 1;
//	} else if (strcmp(argv[0],"-pre:flexibleC1P1")==0) {
//		options->flexibleC1P1 = 1;
//		return 1;
//	} else if (strcmp(argv[0],"--pre:flexibleC1P1")==0) {
//		options->flexibleC1P1 = 0;
//		return 1;
	} else if (strcmp(argv[0],"-pre:cs:li")==0) {
		sopt->csLi = 1;
		return 1;
	} else if (strcmp(argv[0],"--pre:cs:li")==0) {
		sopt->csLi = 0;
		return 1;
	} else if (strcmp(argv[0],"-pre:cs:li:min")==0) {
		if (argv[1]==NULL) return -3;
		options->csLimin = atof(argv[1]);
		return 2;
	} else if (strcmp(argv[0],"-pre:cs:li:max")==0) {
		if (argv[1]==NULL) return -3;
		options->csLimax = atof(argv[1]);
		return 2;
	} else if (strcmp(argv[0],"-pre:cs:li:t0")==0) {
		if (argv[1]==NULL) return -3;
		options->csLit = atof(argv[1]);
		return 2;
	} else if (strcmp(argv[0],"-pre:cs:bw")==0) {
		sopt->csBW = 1;
		return 1;
	} else if (strcmp(argv[0],"--pre:cs:bw")==0) {
		sopt->csBW = 0;
		return 1;
	} else if (strcmp(argv[0],"-pre:cs:bw:min")==0) {
		if (argv[1]==NULL) return -3;
		options->csBWmin = atof(argv[1]);
		return 2;
	} else if (strcmp(argv[0],"-pre:cs:bw:max")==0) {
		if (argv[1]==NULL) return -3;
		options->csBWmax = atof(argv[1]);
		return 2;
	} else if (strcmp(argv[0],"-pre:cs:bw:slope")==0) {
		if (argv[1]==NULL) return -3;
		options->csBWslope = atof(argv[1]);
		return 2;
	} else if (strcmp(argv[0],"-pre:cs:l1c1")==0) {
		sopt->csL1C1 = 1;
		return 1;
	} else if (strcmp(argv[0],"--pre:cs:l1c1")==0) {
		sopt->csL1C1 = 0;
		return 1;		
	} else if (strcmp(argv[0],"-pre:cs:l1c1:slope")==0) {
		if (argv[1]==NULL) return -3;
		options->csL1C1slope = atof(argv[1]);
		return 2;
	} else if (strcmp(argv[0],"-pre:cs:l1c1:window")==0) {
		if (argv[1]==NULL) return -3;
		options->csL1C1window = atoi(argv[1]);
		return 2;
	} else if (strcmp(argv[0],"-pre:cs:l1c1:max")==0) {
		if (argv[1]==NULL) return -3;
		options->csL1C1max = atof(argv[1]);
		return 2;
	} else if (strcmp(argv[0],"-pre:setrectype")==0) {
		if (argv[1]==NULL) return -3;
		if (strcmp(argv[1],"gpsrt")==0) {
			sopt->setRecType=rtNA;
		} else if (strcmp(argv[1],"0")==0) {
			sopt->setRecType=rtUNKNOWN;
		} else if (strcmp(argv[1],"1")==0) {
			sopt->setRecType=rtCROSS;
		} else if (strcmp(argv[1],"2")==0) {
			sopt->setRecType=rtNOP1;
		} else if (strcmp(argv[1],"3")==0) {
			sopt->setRecType=rtNOMINAL;
		} else return -2;
		return 2;
	} else if (strcmp(argv[0],"-filter:meas")==0) {
		if (argv[1]==NULL) return -3;
		if (strcmp(argv[1],"pseudorange")==0) {
			sopt->navObservables = NOpseudorange;
		} else if (strcmp(argv[1],"carrierphase")==0) {
			sopt->navObservables = NOcarrierphase;
		} else return -2;
		return 2;
	} else if (strcmp(argv[0],"-filter:select")==0) {
		if (argv[1]==NULL) return -3;
		sopt->totalFilterMeasurements = atoi(argv[1]);
		if (sopt->totalFilterMeasurements==0) return -2;
		for (i=0;i<sopt->totalFilterMeasurements;i++) {
			if (argv[2+i]==NULL) return -3;
			sopt->measurement[i] = measstr2meastype(argv[2+i]);
			if (sopt->measurement[i] == NA) {
				strcpy (argv[1],argv[2+i]); // This is for a proper error control
				return -2;
			}
			//sopt->smoothWith[i] = -1;
		}
		return 2+i;
	} else if (strcmp(argv[0],"-pre:sat")==0) {
		if (argv[1]==NULL) return -3;
		if (argv[1][0]=='G') {
			aux1 = GPS;
			aux2 = atoi(&argv[1][1]);
			if (aux2<=0) return -2;
			options->includeSatellite[aux1][aux2] = 0;
		} else return -2;
		return 2;
	} else if (strcmp(argv[0],"-pre:availf")==0) {
		if (argv[1]==NULL) return -3;
		if (argv[1][0]=='G') {
			aux1 = GPS;
			// If this option is set, only present frequencies will be set
			for (i=0;i<MAX_FREQUENCIES_PER_GNSS;i++) {
				options->usableFreq[aux1][i] = 0;
			}
			if (strlen(argv[1])<=1) return -2;
			for (i=1;i<strlen(argv[1]);i++) {
				auxstr[0] = argv[1][i];
				auxstr[1] = '\0';
				aux2 = atoi(auxstr);
				options->usableFreq[aux1][aux2] = 1;
			}
		} else return -2;
		return 2;
	} else if (strcmp(argv[0],"-pre:smooth")==0) {
		if (argv[1]==NULL) return -3;
		options->smoothEpochs = atoi(argv[1]);
		return 2;
	} else if (strcmp(argv[0],"-pre:smoothMeas")==0) {
		if (argv[2]==NULL) return -3;
		aux1 = atoi(argv[1]);
		if (aux1==0) {
			return -2;
		}
		sopt->smoothWith[aux1-1] = measstr2meastype(argv[2]);
		if (sopt->smoothWith[aux1-1] == NA) {
			strcpy (argv[1],argv[2]); // This is for a proper error control
			return -2;
		}
		return 3;
	} else if (strcmp(argv[0],"-pre:setrecpos")==0) {
		if (argv[1]==NULL) return -3;
		if (strcasecmp(argv[1],"RINEX")==0) {
			options->receiverPositionSource = rpRINEX;
			return 2;
		} else if (strcasecmp(argv[1],"SINEX")==0) {
			options->receiverPositionSource = rpSINEX;
			return 2;
		} else if (strcasecmp(argv[1],"calculate")==0) {
			options->receiverPositionSource = rpCALCULATE;
			return 2;
		} else {  // X Y Z
			options->receiverPositionSource = rpSET;
			for (i=0;i<3;i++) {
				if (argv[1+i]==NULL) return -3;
				auxd = atof(argv[1+i]);
				//if (auxd==0) {
				//	strcpy(argv[1],argv[1+i]);
				//	return -2;
				//}
				options->aprioriReceiverPosition[i] = auxd;
			}
			return 4;
		}
	} else if (strcmp(argv[0],"-model:iono")==0) {
		if (argv[1]==NULL) return -3;
		if (strcmp(argv[1],"no")==0) {
			sopt->ionoModel = NoIono;
		} else if (strcmp(argv[1],"Klobuchar")==0) {
			sopt->ionoModel = Klobuchar;
		} else return -2;
		return 2;
	} else if (strcmp(argv[0],"--model:iono")==0) {
		sopt->ionoModel = NoIono;
		return 1;


	} else if (strcmp(argv[0],"-model:trop")==0) {
		sopt->troposphericCorrection = 1;
		return 1;
	} else if (strcmp(argv[0],"--model:trop")==0) {
		sopt->troposphericCorrection = 0;
		return 1;

	} else if (strcmp(argv[0],"-model:trop:nominal")==0) {
		if (argv[1]==NULL) return -3;
		if (strcmp(argv[1],"Simple")==0) {
			sopt->tropNominal = SimpleNominal;
		} else if (strcmp(argv[1],"UNB3")==0) {
			sopt->tropNominal = UNB3Nominal;
		} else return -2;
		return 2;

	} else if (strcmp(argv[0],"-model:trop:mapping")==0) {
		if (argv[1]==NULL) return -3;
		if (strcmp(argv[1],"Simple")==0) {
			sopt->tropMapping = SimpleMapping;
		} else if (strcmp(argv[1],"Niell")==0) {
			sopt->tropMapping = NiellMapping;
		} else return -2;
		return 2;

	} else if (strcmp(argv[0],"-model:satclocks")==0) {
		options->satelliteClockCorrection = 1;
		return 1;
	} else if (strcmp(argv[0],"--model:satclocks")==0) {
		options->satelliteClockCorrection = 0;
		return 1;
	} else if (strcmp(argv[0],"-model:relclock")==0) {
		options->relativisticCorrection = 1;
		return 1;
	} else if (strcmp(argv[0],"--model:relclock")==0) {
		options->relativisticCorrection = 0;
		return 1;
	} else if (strcmp(argv[0],"-model:satmovinflight")==0) {
		options->timeTrans = 1;
		return 1;
	} else if (strcmp(argv[0],"--model:satmovinflight")==0) {
		options->timeTrans = 0;
		return 1;
	} else if (strcmp(argv[0],"-model:earthrotinflight")==0) {
		options->earthRotation = 1;
		return 1;
	} else if (strcmp(argv[0],"--model:earthrotinflight")==0) {
		options->earthRotation = 0;
		return 1;
	} else if (strcmp(argv[0],"-model:satphasecenter")==0) {
		options->satellitePhaseCenter = 1;
		return 1;
	} else if (strcmp(argv[0],"--model:satphasecenter")==0) {
		options->satellitePhaseCenter = 0;
		return 1;
	} else if (strcmp(argv[0],"-model:recphasecenter")==0) {
		if (argv[1]==NULL) return -3;
		if (strcmp(argv[1],"no")==0) {
			options->antennaData = adNONE;
			return 2;
		} else if (strcmp(argv[1],"ANTEX")==0) {
			options->antennaData = adANTEX;
			return 2;
		} else {
			options->antennaData = adSET;
			aux1 = atoi(argv[1]);
			if (aux1==0) return -2;
			for (i=0;i<3;i++) {
				if (argv[2+i]==NULL) return -3;
				auxd = atof(argv[2+i]);
				if (auxd==0 && *argv[2+i]!='0') {
					strcpy(argv[1],argv[2+i]);
					return -2;
				}
				options->receiverPCO[aux1][i] = auxd;
			}
			return 5;
		}
	} else if (strcmp(argv[0],"-model:strictradome")==0) {
		options->strictradome = 1;
		return 1;
	} else if (strcmp(argv[0],"--model:strictradome")==0) {
		options->strictradome = 0;
		return 1;
	} else if (strcmp(argv[0],"--model:recphasecenter")==0) {
		options->antennaData = adNONE;
		return 1;
	} else if (strcmp(argv[0],"-model:arp")==0) {
		if (argv[1]==NULL) return -3;
		if (strcmp(argv[1],"no")==0) {
			options->ARPData = arpNONE;
			return 2;
		} else if (strcmp(argv[1],"RINEX")==0) {
			options->ARPData = arpRINEX;
			return 2;
		} else {
			options->ARPData = arpSET;
			for (i=0;i<3;i++) {
				if (argv[1+i]==NULL) return -3;
				auxd = atof(argv[1+i]);
				options->receiverARP[i] = auxd;
			}
		}
		return 4;
	} else if (strcmp(argv[0],"--model:arp")==0) {
		options->ARPData = arpNONE;
		return 1;
	} else if (strcmp(argv[0],"-model:dcb:p1c1")==0) {
		if (argv[1]==NULL) return -3;
		if (strcmp(argv[1],"no")==0) {
			sopt->p1c1dcbModel = p1c1NONE;
	  		options->flexibleC1P1 = 0;
		} else if (strcmp(argv[1],"flexible")==0) {
			sopt->p1c1dcbModel = p1c1FLEXIBLE;
	  		options->flexibleC1P1 = 1;
		} else if (strcmp(argv[1],"strict")==0) {
			sopt->p1c1dcbModel = p1c1STRICT;
	  		options->flexibleC1P1 = 1;
		} else {
			return -2;
		}
		return 2;
	} else if (strcmp(argv[0],"--model:dcb:p1c1")==0) {
		sopt->p1c1dcbModel = p1c1NONE;
	  	options->flexibleC1P1 = 0;
		return 1;
	} else if (strcmp(argv[0],"-model:dcb:p1p2")==0) {
		if (argv[1]==NULL) return -3;
		if (strcmp(argv[1],"no")==0) {
			sopt->p1p2dcbModel = p1p2NONE;
		} else if (strcmp(argv[1],"RINEX")==0) {
			sopt->p1p2dcbModel = p1p2RINEX;
		} else if (strcmp(argv[1],"DCB")==0) {
			sopt->p1p2dcbModel = p1p2DCB;
		} else {
			return -2;
		}
		return 2;
	} else if (strcmp(argv[0],"--model:dcb:p1p2")==0) {
		options->p1p2dcbModel = p1p2NONE;
		return 1;
	} else if (strcmp(argv[0],"-model:windup")==0) {
		options->windUpCorrection = 1;
		return 1;
	} else if (strcmp(argv[0],"--model:windup")==0) {
		options->windUpCorrection = 0;
		return 1;
	} else if (strcmp(argv[0],"-model:solidtides")==0) {
		options->solidTidesCorrection = 1;
		return 1;
	} else if (strcmp(argv[0],"--model:solidtides")==0) {
		options->solidTidesCorrection = 0;
		return 1;
	} else if (strcmp(argv[0],"-model:relpath")==0) {
		options->gravitationalCorrection = 1;
		return 1;
	} else if (strcmp(argv[0],"--model:relpath")==0) {
		options->gravitationalCorrection = 0;
		return 1;
	} else if (strcmp(argv[0],"-model:orbit:deg")==0) {
		if (argv[1]==NULL) return -3;
		aux1 = atoi(argv[1]);
		options->orbitInterpolationDegree = aux1;
		return 2;
	} else if (strcmp(argv[0],"-model:clock:deg")==0) {
		if (argv[1]==NULL) return -3;
		aux1 = atoi(argv[1]);
		options->clockInterpolationDegree = aux1;
		return 2;
	} else if (strcmp(argv[0],"-model:satellitehealth")==0) {
		options->satelliteHealth = 1;
		return 1;
	} else if (strcmp(argv[0],"--model:satellitehealth")==0) {
		options->satelliteHealth = 0;
		return 1;
	} else if (strcmp(argv[0],"-filter:trop")==0) {
		sopt->estimateTroposphere = 1;
		return 1;
	} else if (strcmp(argv[0],"--filter:trop")==0) {
		sopt->estimateTroposphere = 0;
		return 1;
	} else if (strcmp(argv[0],"-filter:fixedweight")==0) {
		if (argv[1]==NULL || argv[2]==NULL) return -3;
		aux1 = atoi(argv[1]);
		if (aux1==0) {
			return -2;
		}
		auxd = atof(argv[2]);
		if (auxd==0) {
			strcpy (argv[1],argv[2]);
			return -2;
		}
		sopt->fixedMeasurementWeight[aux1-1] = auxd;
		return 3;
	} else if (strcmp(argv[0],"-filter:elevweight")==0) {
		for (i=0;i<4;i++) if (argv[i+1]==NULL) return -3;
		aux1 = atoi(argv[1]);
		if (aux1==0) return -2;
		auxd = atof(argv[2]);
		sopt->fixedMeasurementWeight[aux1-1] = auxd;
		auxd = atof(argv[3]);
		sopt->elevationMeasurementWeight[aux1-1][0] = auxd;
		auxd = atof(argv[4]);
		sopt->elevationMeasurementWeight[aux1-1][1] = auxd*d2r;
		return 5;
	} else if (strcmp(argv[0],"-filter:phi:dr")==0) {
		if (argv[1]==NULL) return -3;
		sopt->filterParams[PHI_PAR][DR_UNK] = atof(argv[1]);
		return 2;
	} else if (strcmp(argv[0],"-filter:phi:clk")==0) {
		if (argv[1]==NULL) return -3;
		sopt->filterParams[PHI_PAR][DT_UNK] = atof(argv[1]);
		return 2;
	} else if (strcmp(argv[0],"-filter:phi:trop")==0) {
		if (argv[1]==NULL) return -3;
		sopt->filterParams[PHI_PAR][TROP_UNK] = atof(argv[1]);
		return 2;
	} else if (strcmp(argv[0],"-filter:phi:amb")==0) {
		if (argv[1]==NULL) return -3;
		sopt->filterParams[PHI_PAR][BIAS_UNK] = atof(argv[1]);
		return 2;
	} else if (strcmp(argv[0],"-filter:q:dr")==0) {
		if (argv[1]==NULL) return -3;
		sopt->filterParams[Q_PAR][DR_UNK] = atof(argv[1]);		
		return 2;
	} else if (strcmp(argv[0],"-filter:q:clk")==0) {
		if (argv[1]==NULL) return -3;
		sopt->filterParams[Q_PAR][DT_UNK] = atof(argv[1]);
		return 2;
	} else if (strcmp(argv[0],"-filter:q:trop")==0) {
		if (argv[1]==NULL) return -3;
		sopt->filterParams[Q_PAR][TROP_UNK] = atof(argv[1]);
		return 2;
	} else if (strcmp(argv[0],"-filter:q:amb")==0) {
		if (argv[1]==NULL) return -3;
		sopt->filterParams[Q_PAR][BIAS_UNK] = atof(argv[1]);
		return 2;
	} else if (strcmp(argv[0],"-filter:p0:dr")==0) {
		if (argv[1]==NULL) return -3;
		sopt->filterParams[P0_PAR][DR_UNK] = atof(argv[1]);
		return 2;
	} else if (strcmp(argv[0],"-filter:p0:clk")==0) {
		if (argv[1]==NULL) return -3;
		sopt->filterParams[P0_PAR][DT_UNK] = atof(argv[1]);
		return 2;
	} else if (strcmp(argv[0],"-filter:p0:trop")==0) {
		if (argv[1]==NULL) return -3;
		sopt->filterParams[P0_PAR][TROP_UNK] = atof(argv[1]);
		return 2;
	} else if (strcmp(argv[0],"-filter:p0:amb")==0) {
		if (argv[1]==NULL) return -3;
		sopt->filterParams[P0_PAR][BIAS_UNK] = atof(argv[1]);
		return 2;
	} else if (strcmp(argv[0],"-filter:backward")==0) {
		options->filterIterations = 2;
		return 1;
	} else if (strcmp(argv[0],"--filter:backward")==0) {
		options->filterIterations = 1;
		return 1;
	} else if (strcmp(argv[0],"-print:info")==0) {
		options->printInfo = 1;
		return 1;
	} else if (strcmp(argv[0],"--print:info")==0) {
		options->printInfo = 0;
		return 1;
	} else if (strcmp(argv[0],"-print:model")==0) {
		options->printModel = 1;
		return 1;
	} else if (strcmp(argv[0],"--print:model")==0) {
		options->printModel = 0;
		return 1;
	} else if (strcmp(argv[0],"-print:prefit")==0) {
		options->printPrefit = 1;
		return 1;
	} else if (strcmp(argv[0],"--print:prefit")==0) {
		options->printPrefit = 0;
		return 1;
	} else if (strcmp(argv[0],"-print:postfit")==0) {
		options->printPostfit = 1;
		return 1;
	} else if (strcmp(argv[0],"--print:postfit")==0) {
		options->printPostfit = 0;
		return 1;
	} else if (strcmp(argv[0],"-print:filter")==0) {
		options->printFilterSolution = 1;
		return 1;
	} else if (strcmp(argv[0],"--print:filter")==0) {
		options->printFilterSolution = 0;
		return 1;
	} else if (strcmp(argv[0],"-print:cycleslips")==0) {
		options->printCycleslips = 1;
		return 1;
	} else if (strcmp(argv[0],"--print:cycleslips")==0) {
		options->printCycleslips = 0;
		return 1;
	} else if (strcmp(argv[0],"-print:satellites")==0) {
		options->printSatellites = 1;
		return 1;
	} else if (strcmp(argv[0],"--print:satellites")==0) {
		options->printSatellites = 0;
		return 1;
	} else if (strcmp(argv[0],"-print:input")==0) {
		options->printInput = 1;
		return 1;
	} else if (strcmp(argv[0],"--print:input")==0) {
		options->printInput = 0;
		return 1;
	} else if (strcmp(argv[0],"-print:meas")==0) {
		options->printMeas = 1;
		return 1;
	} else if (strcmp(argv[0],"--print:meas")==0) {
		options->printMeas = 0;
		return 1;
	} else if (strcmp(argv[0],"-print:output")==0) {
		options->printOutput = 1;
		return 1;
	} else if (strcmp(argv[0],"--print:output")==0) {
		options->printOutput = 0;
		return 1;
	} else if (strcmp(argv[0],"-print:satsel")==0) {
		options->printSatSel = 1;
		return 1;
	} else if (strcmp(argv[0],"--print:satsel")==0) {
		options->printSatSel = 0;
		return 1;
	} else if (strcmp(argv[0],"-print:satdiff")==0) {
		options->printSatDiff = 1;
		return 1;
	} else if (strcmp(argv[0],"--print:satdiff")==0) {
		options->printSatDiff = 0;
		return 1;
	} else if (strcmp(argv[0],"-print:satstat")==0) {
		options->printSatStat = 1;
		return 1;
	} else if (strcmp(argv[0],"--print:satstat")==0) {
		options->printSatStat = 0;
		return 1;
	} else if (strcmp(argv[0],"-print:satstattot")==0) {
		options->printSatStatTot = 1;
		return 1;
	} else if (strcmp(argv[0],"--print:satstattot")==0) {
		options->printSatStatTot = 0;
		return 1;
	} else if (strcmp(argv[0],"-print:satpvt")==0) {
		options->printSatPvt = 1;
		return 1;
	} else if (strcmp(argv[0],"--print:satpvt")==0) {
		options->printSatPvt = 0;
		return 1;
	} else if (strcmp(argv[0],"-print:satsel")==0) {
		options->printSatSel = 1;
		return 1;
	} else if (strcmp(argv[0],"--print:satsel")==0) {
		options->printSatSel = 0;
		return 1;
	} else if (strcmp(argv[0],"-print:all")==0) {

		options->printInfo = 1;
		options->printInput = 1;
		options->printMeas = 1;
		options->printCycleslips = 1;
		options->printModel = 1;
		options->printSatellites = 1;
		options->printSatSel = 1;
		options->printPrefit = 1;
		options->printFilterSolution = 1;
		options->printPostfit = 1;
		options->printOutput = 1;
		options->printSatDiff = 1;
		options->printSatStat = 1;
		options->printSatStatTot = 1;
		return 1;
	} else if (strcmp(argv[0],"-print:none")==0) {
		options->printInfo = 0;
		options->printInput = 0;
		options->printMeas = 0;
		options->printCycleslips = 0;
		options->printModel = 0;
		options->printSatellites = 0;
		options->printSatSel = 0;
		options->printPrefit = 0;
		options->printFilterSolution = 0;
		options->printPostfit = 0;
		options->printOutput = 0;
		options->printSatDiff = 0;
		options->printSatStat = 0;
		options->printSatStatTot = 0;
		return 1;
	}
	return -1;
}

/*****************************************************************************
 * Name        : parseAllOptions
 * Description : Parse all options from a vector of strings into the options or 
 *               support options
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TSupportOptions  *sopt          O  N/A  TSupportOptions structure
 * TOptions  *options              O  N/A  TOptions structure
 * int  argc                       I  N/A  Number of strings in argv[]
 * char  *argv[]                   I  N/A  Vector of strings to be parsed
 * Returned value (int)            O  N/A  Status of the function
 *                                         0 => Required program stop 
 *                                              (error, showHelp or showExamples)                                        
 *                                         1 => Normal options lecture
 *****************************************************************************/
int parseAllOptions (TSupportOptions *sopt,TOptions *options, int argc, char *argv[]) {
	char	**ptr;
	int		i;
	int		res;
	
	i = 1;
	
	if (argc==1) {
		showHelp();
		return 0;
	}
	
	while (i<argc) {
		ptr = &argv[i];
		res = parseOption(sopt,options,ptr);
		if (res==-3) { // Missing parameters
			sprintf(messagestr,"'%s' requires more parameters. Use '-help' for a set of valid options.",*ptr);
			printError(messagestr,options);
			return 0;
		} else if (res==-2) { // Value not valid
			sprintf(messagestr,"'%s' is not a valid value for '%s'. Use '-help' for a set of valid options.",ptr[1],*ptr);
			printError(messagestr,options);
			return 0;
		} else if (res==-1) { // Parameter unknown
			sprintf(messagestr,"'%s' option not recognized. Use '-help' for a set of valid options.",*ptr);
			printError(messagestr,options);
			return 0;
		} else if (res==0) { // Show help/examples/config/messages
			return 0;
		}
		// Increment by the number of parameters readed
		i+=res; 
	}
	
	return 1;
}


/*****************************************************************************
 * Name        : processOptions
 * Description : To be called after all the options have been parsed. Read 
 *               options from TSupportOptions and translates it to TOptions
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TSupportOptions  *sopt          I  N/A  TSupportOptions structure
 * TOptions  *options              IO N/A  TOptions structure
 * Returned value (int)            O  N/A  Status of the function
 *                                         0 => Some problems with the options chosen
 *                                              (compatibility or missing critical
 *                                              options)                                      
 *                                         1 => Processing properly done
 *****************************************************************************/
int processOptions (TSupportOptions *sopt, TOptions *options) {
	int		preciseProductsAvailable;
	int		broadcastProductsAvailable;
	int 	i,j;
	int		sources;
	int		nProd;
	enum fileType				auxft1;
	
	// Number of GNSS product sources availables
	nProd = sopt->numNav + sopt->numSP3 + sopt->numOrb;
	
	// Output redirection
	if (sopt->outFile[0]!='\0') {
		if (freopen(sopt->outFile,"w",stdout)==NULL) {
			sprintf(messagestr,"Problem with output file '%s' redirection",sopt->outFile);
			printError(messagestr,options);
		}
	}
	
	// Products availability check
	if (sopt->SP3File[0][0]=='\0' && (sopt->orbFile[0][0]=='\0' || sopt->clkFile[0][0]=='\0')) {
		preciseProductsAvailable = 0;
	} else {
		preciseProductsAvailable = 1;
	}
	
	if (sopt->navFile[0][0]=='\0') {
		broadcastProductsAvailable = 0;
	} else {
		broadcastProductsAvailable = 1;
	}
	
	// Sanity check
	if (sopt->obsFile[0]!='\0') {
		options->workMode = wmDOPROCESSING;
		if (broadcastProductsAvailable && sopt->navMode==NMND) sopt->navMode=NMstandalone;
		if (preciseProductsAvailable && sopt->navMode==NMND) sopt->navMode=NMppp;
		if (broadcastProductsAvailable && preciseProductsAvailable) {
			printError("Cannot determine if processing Standalone or PPP. Please use only one source of GNSS products",options);
		} else if (sopt->navMode==NMppp && broadcastProductsAvailable) {
			printError("PPP and input:nav were set",options);
		} else if (sopt->navMode==NMstandalone && preciseProductsAvailable) {
			printError("Standalone and input:SP3/input:orb/input:clk were set",options);
		} else if (sopt->SP3File[0][0]!='\0' && sopt->orbFile[0][0]!='\0' && sopt->clkFile[0][0]!='\0') {
			printError("input:SP3, input:orb and input:clk cannot be all set",options);
		} else if (!broadcastProductsAvailable && !preciseProductsAvailable) {
			printInfo("Disconnecting modelling and filtering (no GNSS products available)",options);
			options->workMode = wmSHOWINPUT;
		}
		if (options->receiverPositionSource == rpSINEX && sopt->snxFile[0]=='\0') {
			printError("Receiver position source set to SINEX, but no SINEX file has been included. Include one with the '-input:snx' option",options);
		}
	} else { // No observation file found
		if (nProd == 1) {
			options->workMode = wmSHOWPRODUCT;
			printInfo("Show satellite information mode",options);
		} else if (nProd == 2) {
			options->workMode = wmCOMPAREORBITS;
			printInfo("Compare orbits & clocks mode",options);
			if (sopt->numOrb!=sopt->numClk) {
				sprintf(messagestr,"Inconsistent number of orb and clk input files (%d vs %d)",sopt->numOrb,sopt->numClk);
				printError(messagestr,options);
			}
		} else {
			sprintf(messagestr,"No observation file is found, but there are %d different GNSS sources. To compare products, provide 2 different sources. To parse a products file, provide 1 single source",nProd);
			printError(messagestr,options);
		} 
	}
	if (sopt->antFile[0]!='\0' && sopt->conFile[0]!='\0') {
		printError("-input:ant and -input:con cannot be both set",options);
	}
	
	if (preciseProductsAvailable) {
		if (options->workMode==wmDOPROCESSING) {
			if (sopt->antFile[0]=='\0' && sopt->conFile[0]=='\0' && (options->satellitePhaseCenter || options->antennaData == adANTEX)) {
				printError("When positioning with precise products, an ANTEX or Constellation file is required (-input:ant or -input:con)",options);
			}	
		} else if (options->workMode==wmCOMPAREORBITS || options->workMode==wmSHOWPRODUCT) {
			if (sopt->antFile[0]=='\0' && sopt->conFile[0]=='\0' && options->satellitePhaseCenter) {
				printError("Precise orbits and 'Satellite Phase Center' corrections are specified, but no ANTEX or Constellation file is found",options);
			}
		}
	}
		
	if (options->workMode == wmDOPROCESSING) {
		// Navigation mode and observables
		if (sopt->navMode==NMppp) {
			options->discardEclipsedSatellites = 1;
			options->csBW = 1;
			options->csLi = 1;
			options->csL1C1 = 0;
			options->ionoModel = NoIono;
			options->troposphericCorrection = 1;
			options->tropNominal = SimpleNominal; 
			options->tropMapping = NiellMapping;
			options->satellitePhaseCenter = 1;
			options->antennaData = adANTEX;
			options->ARPData = arpRINEX;
			options->solidTidesCorrection = 1;
			options->gravitationalCorrection = 1;
			// Carrier phase + pseudorange by default
			if (sopt->navObservables==NOND) sopt->navObservables = NOcarrierphase;
			// Static by default
			if (sopt->navKinematics==NKND) sopt->navKinematics = NKstatic;
			printInfo("Processing with precise products",options);
		} else if (sopt->navMode==NMstandalone) {
			options->discardEclipsedSatellites = 0;
			options->csBW = 0;
			options->csLi = 0;
			options->csL1C1 = 1;
			options->ionoModel = Klobuchar;
			// Adria P1P2 DCB source for SPP			
			options->p1p2dcbModel = p1p2RINEX;
			options->troposphericCorrection = 1;
			options->tropNominal = UNB3Nominal; 
			options->tropMapping = SimpleMapping;
			options->satellitePhaseCenter = 0;
			options->antennaData = adNONE;
			options->ARPData = arpNONE;
			options->solidTidesCorrection = 0;
			options->gravitationalCorrection = 0;
			//options->usableFreq[GPS][2] = 0;
			// Pseudorange only
			if (sopt->navObservables==NOND) sopt->navObservables = NOpseudorange;
			// Dynamic by default
			if (sopt->navKinematics==NKND) sopt->navKinematics = NKkinematic;
			printInfo("Processing with broadcast products",options);
		}
		if (sopt->navMode==NMppp && sopt->navObservables==NOpseudorange) {
			options->estimateTroposphere = 0;
			options->totalFilterMeasurements = 1;
			options->measurement[0] = PC;
			options->fixedMeasurementWeight[0] = 1;
			options->usePhase = 0;
			options->windUpCorrection = 0;
		} else if (sopt->navMode==NMppp && sopt->navObservables==NOcarrierphase) {
			options->estimateTroposphere = 1;
			options->totalFilterMeasurements = 2;
			options->measurement[0] = PC;
			options->measurement[1] = LC;
			options->fixedMeasurementWeight[0] = 1;
			options->fixedMeasurementWeight[1] = 0.01;
			options->usePhase = 1;
			options->windUpCorrection = 1;
		} else if (sopt->navMode==NMstandalone && sopt->navObservables==NOpseudorange) { 
			options->estimateTroposphere = 0;
			options->totalFilterMeasurements = 1;
			options->measurement[0] = C1C;  // C/A
			options->usePhase = 0;
			options->fixedMeasurementWeight[0] = 5;
			options->windUpCorrection = 0;
		} else if (sopt->navMode==NMstandalone && sopt->navObservables==NOcarrierphase) {
			options->estimateTroposphere = 0;
			options->totalFilterMeasurements = 2;
			options->measurement[0] = C1C;  // C/A
			options->measurement[1] = L1P;  // L1
			options->usePhase = 1;
			options->fixedMeasurementWeight[0] = 5;
			options->fixedMeasurementWeight[1] = 0.10;
			options->windUpCorrection = 1;
		}

		// Navigation kinematics
		if (sopt->navKinematics==NKstatic) {
			options->filterParams[PHI_PAR][DR_UNK] = 1;
			options->filterParams[Q_PAR][DR_UNK] = 0;
		} else if (sopt->navKinematics==NKkinematic) {
			options->filterParams[PHI_PAR][DR_UNK] = 0;
			options->filterParams[Q_PAR][DR_UNK] = SIGMA_INF_DR*SIGMA_INF_DR;
		}

		// Filter measurements
		if (sopt->totalFilterMeasurements!=-1) {
			options->totalFilterMeasurements = sopt->totalFilterMeasurements;
			options->usePhase = 0;
			options->estimateTroposphere = 0;
			options->minArcLength = 0;
			for (i=0;i<options->totalFilterMeasurements;i++) {
				options->measurement[i] = sopt->measurement[i];
				if (whatIs(options->measurement[i]) == CarrierPhase) {
					options->usePhase = 1;
					options->estimateTroposphere = 1;
					options->minArcLength = 4;
				}
				if (sopt->fixedMeasurementWeight[i]!=-1) {
					options->fixedMeasurementWeight[i] = sopt->fixedMeasurementWeight[i];
				}
			}
		}
		
		// Filter weights
		for (i=0;i<options->totalFilterMeasurements;i++) {
			if (sopt->fixedMeasurementWeight[i]!=-1) {
				options->fixedMeasurementWeight[i] = sopt->fixedMeasurementWeight[i];
			}
			if (sopt->elevationMeasurementWeight[i][0]!=-1) {
				options->elevationMeasurementWeight[i][0] = sopt->elevationMeasurementWeight[i][0];
				options->elevationMeasurementWeight[i][1] = sopt->elevationMeasurementWeight[i][1];
				if (options->elevationMeasurementWeight[i][1]==0) {
					options->elevationMeasurementWeight[i][0] = 0;
					options->elevationMeasurementWeight[i][1] = 1e-5;
				}
			}
		}
		
		// Filter parameters change
		for (i=0;i<3;i++) {
			for (j=0;j<MAX_PAR;j++) {
				if (sopt->filterParams[i][j]!=-1) {
					options->filterParams[i][j] = sopt->filterParams[i][j];
				}
			}
		}
	
		// Smoothing
		if (options->smoothEpochs!=0) {
			for (i=0;i<MAX_FILTER_MEASUREMENTS;i++) {
				if (sopt->smoothWith[i]!=NA) {
					options->minArcLength = 4;
					options->smoothMeas[i] = sopt->smoothWith[i];
				} else {
					options->smoothMeas[i] = NA;
				}
			}
		}
		options->minArcLength = 4;
		
		// Ionospheric model
		if (sopt->ionoModel!=UNKNOWN_IM) {
			options->ionoModel = sopt->ionoModel; 
			printInfo("Forcing ionospheric model for this processing",options);
		}
		if (options->ionoModel==Klobuchar) {
			if (sopt->numNav==0 && sopt->klbFile[0]=='\0') { // No Klobuchar source
				printError("Unable to get ionospheric model Klobuchar from input data. Please use '-input:nav' or '-input:klb'",options);
				return 0;
			}
		}
	
		// Tropospheric estimation
		if (sopt->estimateTroposphere==1) {
			options->estimateTroposphere = 1;
			printInfo("Forcing enabled tropospheric estimation for this processing",options);
		} else if (sopt->estimateTroposphere==0) {
			options->estimateTroposphere = 0;
			printInfo("Forcing disabled tropospheric estimation for this processing",options);
		}
	
/*
		// Tropospheric model
		if (sopt->tropModel!=UNKNOWN_TM) {
			options->tropModel = sopt->tropModel;
			if (options->tropModel == NoTrop) {
				if (options->estimateTroposphere) {
					printInfo("Disconecting tropospheric estimation (no tropospheric nominal value computed)",options);
					options->estimateTroposphere = 0;
				}
			}
		}
*/

		// Tropospheric model: specific configuration check
		if (sopt->troposphericCorrection!=-1) options->troposphericCorrection = sopt->troposphericCorrection;

		// Tropospheric model: Nominal
		if (sopt->tropNominal!=UNKNOWN_TMNominal) {
			options->tropNominal = sopt->tropNominal;
			if (options->troposphericCorrection == 0) {
				printInfo("Warning: Tropospheric modelling is disabled although nominal values are given",options);
				if (options->estimateTroposphere) {
					printInfo("Disconnecting tropospheric estimation (no tropospheric nominal value computed)",options);
					options->estimateTroposphere = 0;
				}
			}
		}

		// Tropospheric model: Mapping
		if (sopt->tropMapping!=UNKNOWN_TMMapping) {
			options->tropMapping = sopt->tropMapping;
			if (options->troposphericCorrection == 0) {
				printInfo("Warning: Tropospheric modelling is disabled although a troposphere mapping is given",options);
			 }
		}

		
		// TGD Source
		if (sopt->p1c1dcbModel!=p1c1UNKNOWN) {
			options->p1c1dcbModel = sopt->p1c1dcbModel;
		}
		if (sopt->p1p2dcbModel!=p1p2UNKNOWN) {
			options->p1p2dcbModel = sopt->p1p2dcbModel;
		}
    		for (i=0;i<sopt->numTGD;i++) {
		      auxft1 = whatFileTypeIs(sopt->tgdFile[i]);
		      if (auxft1 == ftP1C1DCB) sopt->numP1C1DCB++;
		      else if (auxft1 == ftP1P2DCB) sopt->numP1P2DCB++;
		      else if (auxft1 == ftRINEXbroadcast) sopt->numP1P2RINEX++;
    		}
		
		// Orbit and clock interpolation degree
		if (options->orbitInterpolationDegree>MAX_INTERPOLATION_DEGREE) {
			sprintf(messagestr,"Max interpolation degree for orbits is %d (selected %d)",MAX_INTERPOLATION_DEGREE,options->orbitInterpolationDegree);
			printError(messagestr,options);
		}
		if (options->clockInterpolationDegree>MAX_INTERPOLATION_DEGREE) {
			sprintf(messagestr,"Max interpolation degree for clocks is %d (selected %d)",MAX_INTERPOLATION_DEGREE,options->clockInterpolationDegree);
			printError(messagestr,options);
		}
		
		// Cycle-slips depending on frequency availability
		if (options->usableFreq[GPS][2]==0) { // Unavailable GPS L2
			options->csL1C1 = 1;
		}
	
		// Usable frequency defaults
		if (sopt->navMode==NMstandalone) {
			options->usableFreq[GPS][2] = 0;
			options->usableFreq[GPS][5] = 0;
		}
	}
			
	// Check forcing a specific configuration for cycle-slips
	if (sopt->csLi!=-1) options->csLi = sopt->csLi;
	if (sopt->csBW!=-1) options->csBW = sopt->csBW;
	if (sopt->csL1C1!=-1) options->csL1C1 = sopt->csL1C1;
		
	// Anyway if dual-frequency data is not available, it will NOT use Li and BW cycle slip detectors...
	if (options->usableFreq[GPS][2]==0) { // Unavailable GPS L2
		if (options->csLi || options->csBW) {
			printInfo ("Forcing disconnection of Li and BW cycle-slip detectors. No dual-frequency data available...",options);
		}
		options->csLi = 0;
		options->csBW = 0;
	}
		
	// Discard eclipse satellites
	if (sopt->discardEclipsedSatellites!=-1) {
		options->discardEclipsedSatellites = sopt->discardEclipsedSatellites;
	}
	
	// Printing
	if (options->printInfo) {
		if (options->workMode==wmDOPROCESSING) {
			sprintf(messagestr,"RINEX observation input file: %s",sopt->obsFile);
			printInfo(messagestr,options);
			if (sopt->navMode==NMstandalone) {
				sprintf(messagestr,"RINEX navigation message input file: %s",sopt->navFile[0]);
				printInfo(messagestr,options);
			}
			if (sopt->navMode==NMppp) {
				if (sopt->SP3File[0][0]!='\0') {
					sprintf(messagestr,"SP3 products file: %s",sopt->SP3File[0]);
					printInfo(messagestr,options);
				} else {
					sprintf(messagestr,"SP3 orbits file: %s",sopt->orbFile[0]);
					printInfo(messagestr,options);
					sprintf(messagestr,"Clocks file:     %s",sopt->clkFile[0]);
					printInfo(messagestr,options);
				}
			}
			if (options->ionoModel==Klobuchar) {
				if (sopt->klbFile[0]!='\0') {
					sprintf(messagestr,"RINEX navigation file for Klobuchar corrections: %s",sopt->klbFile);
					printInfo(messagestr,options);
				} else {
					printInfo("File for Klobuchar corrections same as for navigation",options); 
				}
			}
			if (sopt->numP1P2RINEX==0 && (options->p1p2dcbModel==p1p2RINEX) && broadcastProductsAvailable) {
				printInfo("File for P1-P2 DCB corrections same as for navigation",options);
			}
			for (i=0;i<sopt->numTGD;i++) {
				sprintf(messagestr,"File for DCB corrections: %s",sopt->tgdFile[i]);
				printInfo(messagestr,options);
			}
			if (sopt->recFile[0]!='\0') {
				sprintf(messagestr,"GPS receiver types file: %s",sopt->recFile);
				printInfo(messagestr,options);
			}
			if (sopt->snxFile[0]!='\0') {
				sprintf(messagestr,"SINEX input file: %s",sopt->snxFile);
				printInfo(messagestr,options);
			}
		} else if (options->workMode==wmCOMPAREORBITS || options->workMode==wmSHOWPRODUCT) {
			sources = 0;
			for (i=0;i<sopt->numNav;i++) {
				sources++;
				sprintf(messagestr,"OC Product %d: RINEX navigation message input file: %s",sources,sopt->navFile[i]);
				printInfo(messagestr,options);
			}
			for (i=0;i<sopt->numSP3;i++) {
				sources++;
				sprintf(messagestr,"OC Product %d: SP3 products file: %s",sources,sopt->SP3File[i]);
				printInfo(messagestr,options);
			}
			for (i=0;i<sopt->numOrb;i++) {
				sources++;
				sprintf(messagestr,"OC Product %d: SP3 orbit file: %s",sources,sopt->orbFile[i]);
				printInfo(messagestr,options);
				sprintf(messagestr,"OC Product %d: clock file:     %s",sources,sopt->clkFile[i]);
				printInfo(messagestr,options);
			}
		}

		if (options->workMode==wmDOPROCESSING || options->workMode==wmCOMPAREORBITS || options->workMode==wmSHOWPRODUCT) {
			if (preciseProductsAvailable) {
				if (sopt->conFile[0]!='\0') {
					sprintf(messagestr,"Constellation status file: %s",sopt->conFile);
					printInfo(messagestr,options);
				}
				if (sopt->antFile[0]!='\0') {
					sprintf(messagestr,"ANTEX antenna input file: %s",sopt->antFile);
					printInfo(messagestr,options);
				}
			}
		}
	}
	return 1;
}

/*****************************************************************************
 * Name        : readConfig
 * Description : Reads the input parameters from a FILE descriptor
 * Parameters  :
 * Name                           |Da|Unit|Description
 * FILE  *fd                       I  N/A  Input file descriptor
 * TSupportOptions  *sopt          IO N/A  TSupportOptions structure
 * TOptions  *options              IO N/A  TOptions structure
 * Returned value (int)            O  N/A  Status of the function
 *                                         -1 => Some problems with the parameter
 *                                               parsing                                
 *                                         1  => Parsing properly done
 *****************************************************************************/
int readConfig (FILE *fd, TSupportOptions *sopt, TOptions *options) {
	char 		line[MAX_INPUT_LINE];
	char		word[MAX_WORD_LENGTH];
	char		*wordlst[MAX_WORDS];
	char		format[MAX_INPUT_LINE];
	char		formataux[MAX_INPUT_LINE];
	int			len = 0;
	int			ret = 0;
	int			wordindex = 1;
	int i;
	
	wordlst[0] = NULL;
	while(getL(line,&len,fd)!=-1) {
		word[0]='#';
		word[1]='\0';
		ret = sscanf(line,"%s",word);
		if (ret) {
			if (word[0]!='#') { // It is not a comment
				wordlst[wordindex] = malloc(sizeof(char)*(strlen(word)+2)); // +2 for the '\0' and the '-' char
				strcpy(wordlst[wordindex],word);
				wordindex++;
				strcpy(format,"%*s");
				sprintf(formataux,"%s %%s",format);
				while(sscanf(line,formataux,word)==1) {
					wordlst[wordindex] = malloc(sizeof(char)*(strlen(word)+1)); // +1 for the '\0'
					strcpy(wordlst[wordindex],word);
					wordindex++;
					sprintf(format,"%s %%*s",format);
					sprintf(formataux,"%s %%s",format);
				}
			}
		}
	}
	
	wordlst[wordindex] = NULL;
	
	// Parse options
	if (!parseAllOptions (sopt,options, wordindex, wordlst)) {
		return -1;
	}
	
	// Free memory
	for (i=0;i<wordindex;i++) {
		free (wordlst[i]);
	}
	
	return 1;
}

/*****************************************************************************
 * Name        : checkConfigFile
 * Description : Checks if a config file has been specified (cfgFile). If it has
 *               call readConfig function to parse the file
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TSupportOptions  *sopt          IO N/A  TSupportOptions structure
 * TOptions  *options              IO N/A  TOptions structure
 * Returned value (int)            O  N/A  Status of the function
 *                                         -1 => Some problems with the parameter
 *                                               parsing or file not found
 *                                         0  => Config file not used
 *                                         1  => Parsing properly done
 *****************************************************************************/
int checkConfigFile (TSupportOptions *sopt, TOptions *options) {
	FILE	*fd;
	int		ret;
	
	if (sopt->cfgFile[0]!='\0') {
		if ((fd = fopen(sopt->cfgFile,"r"))) {
			ret = readConfig (fd, sopt, options);
			fclose (fd);
			return ret;
		} else {
			// File does not exist
			return -1;
		}
	}
	return 0;
}

/*****************************************************************************
 * Name        : compareOrbits
 * Description : Perform an orbit comparison between two different products
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TGNSSproducts  products[2]      I  N/A  Products to be compared
 * TConstellation  *constellation  I  N/A  Constellation information on the products
 * TSupportOptions  *sopt          I  N/A  TSupportOptions structure
 * TOptions  *options              I  N/A  TOptions structure
 *****************************************************************************/
void compareOrbits (TGNSSproducts products[2], TConstellation *constellation, TOptions *options, TSupportOptions *sopt) {
	double			orientation[2][3][3];
	double			orientationACR[3][3];
	int				i,j,k;
	TTime			startTime;
	TTime			endTime;
	TTime			t;
	double			position[2][3];
	double			velocity[2][3];
	double			ITRFvelocity[2][3];
	double			posDiff[3];
	double			clock[2];
	double			SISREDiff;
	double			Orb3DDiff;
	double			OrbDiff;
	double			ClkDiff;
	double			dr[3];
	int				available;
	double 			tIncrease = 30;
	int				ret;
	double			sunPos[3];
	double			auxT;
	TSatellite		sat;
	TStat			statSISRE[MAX_GNSS][MAX_SATELLITES_PER_GNSS];
	TStat			statSISRETot;
	TStat			stat3D[MAX_GNSS][MAX_SATELLITES_PER_GNSS];
	TStat			stat3DTot;
	TStat			statOrb[MAX_GNSS][MAX_SATELLITES_PER_GNSS];
	TStat			statOrbTot;
	TStat			statClk[MAX_GNSS][MAX_SATELLITES_PER_GNSS];
	TStat			statClkTot;
	TStat			statRadOrb[MAX_GNSS][MAX_SATELLITES_PER_GNSS];
	TStat			statRadOrbTot;
	TStat			statAloOrb[MAX_GNSS][MAX_SATELLITES_PER_GNSS];
	TStat			statAloOrbTot;
	TStat			statCroOrb[MAX_GNSS][MAX_SATELLITES_PER_GNSS];
	TStat			statCroOrbTot;
	TSatInfo		satInfo[MAX_GNSS][MAX_SATELLITES_PER_GNSS];
	int				bothSP3;
	
	for (j=0;j<MAX_GNSS;j++) {
		for (i=0;i<MAX_SATELLITES_PER_GNSS;i++) {
			satInfo[j][i].lastEclipse.SoD = 0;
			satInfo[j][i].lastEclipse.MJDN = 0;
		}
	}
	
	// Initialize stats
	initStat(&statSISRETot);
	initStat(&stat3DTot);
	initStat(&statOrbTot);
	initStat(&statClkTot);
	initStat(&statRadOrbTot);
	initStat(&statAloOrbTot);
	initStat(&statCroOrbTot);
	for (k=0;k<MAX_GNSS;k++) {
		for (i=0;i<MAX_SATELLITES_PER_GNSS;i++) {
			initStat(&statSISRE[k][i]);
			initStat(&stat3D[k][i]);
			initStat(&statOrb[k][i]);
			initStat(&statClk[k][i]);
			initStat(&statRadOrb[k][i]);
			initStat(&statAloOrb[k][i]);
			initStat(&statCroOrb[k][i]);
		}
	}
	
	// Use decimate option to choose increase interval
	if (options->decimate != 0) {
		tIncrease = options->decimate;
	}
	
	// A start and end times will be obtained from the products. 
	// In genereal the second product will be priorized (it is supposed 
	// to be the reference), nevertheless SP3 products will be have 
	// higher priority if they exist, as they have stricter intervals.
	if (products[0].type == SP3 && products[1].type == BRDC) {
		startTime = getProductsFirstEpoch(&products[0]);
		endTime = getProductsLastEpoch(&products[0]);
	} else {
		startTime = getProductsFirstEpoch(&products[1]);
		endTime = getProductsLastEpoch(&products[1]);
	}	
	
	sprintf(messagestr,"Discard satellites under eclipse condition:    %3s",options->discardEclipsedSatellites?"ON":"OFF");
	printInfo(messagestr,options);
	
	sprintf(messagestr,"StartTime: %5d %8.2lf    EndTime: %5d %8.2lf    IntervalTime: %8.2lf",startTime.MJDN,startTime.SoD,endTime.MJDN,endTime.SoD,tIncrease);
	printInfo(messagestr,options);
	
	if (products[0].type == SP3 && products[1].type == SP3) bothSP3 = 1;
	else bothSP3 = 0;
	
	if (bothSP3 && options->satellitePhaseCenter) {
		printInfo ("WARNING both input orbits are SP3, and the satellite phase center corrections are active. These corrections will only be applied to the second SP3 products",options);
	}
	
	memcpy(&t,&startTime,sizeof(TTime));
	while(tdiff(&t,&endTime)<0) {
		if (options->discardEclipsedSatellites) {
			findSun(&t,sunPos);
		}
		for (k=0;k<MAX_GNSS;k++) {
			for (i=0;i<MAX_SATELLITES_PER_GNSS;i++) {
				available=1;
				for (j=0;j<2;j++) {
					// Check if satellite has been deselected
					if (!options->includeSatellite[k][i]) available = 0;
					// Satellite positions and velocities
					ret = getSatellitePVT(&products[j],&t,0.0,k,i,position[j],velocity[j],ITRFvelocity[j],NULL,options);
					if (!ret) {
						available = 0;
					} else {
						// Eclipsed satellite check
						memcpy(sat.position,position[j],3*sizeof(double));
						memcpy(sat.velocity,velocity[j],3*sizeof(double));
						sat.hasOrbitsAndClocks = 1;
						if (options->discardEclipsedSatellites) {
							if (isEclipsed(&t,&sat,&satInfo[k][i],sunPos,&auxT)) {
								if (auxT==0) {
									printSatSel(0,i,"Satellite under eclipse",options);
								} else {
									sprintf(messagestr,"Satellite was under eclipse %4.0lf seconds ago",auxT);
									printSatSel(0,i,messagestr,options);
								}
								available = 0;
							}
						}
					}
					// Satellite clock
					if (options->satelliteClockCorrection) {
						ret = getSatellitePVT(&products[j],&t,0.0,k,i,NULL,NULL,NULL,&clock[j],options);
						if (!ret) available = 0;
					}
					// Satellite orientation
					getSatelliteOrientation (&t, position[j], orientation[j]);
					// Satellite phase center correction
					if (products[j].type == SP3 && options->satellitePhaseCenter) {
						if (!bothSP3 || j==1) { // The contrary, will mean both SP3 input files, and first product
							ret = satellitePhaseCenterCorrection3D(&t,k,i,C1C,orientation[j],constellation,dr);
							if (!ret) available = 0;
							// Coordinate correction
							position[j][0] += dr[0];
							position[j][1] += dr[1];
							position[j][2] += dr[2];
						}
					}
				}
				if (available) {
					for (j=0;j<3;j++)
						posDiff[j] = position[1][j] - position[0][j];
	
					// Satellite orientation in the ACR system (along, cross, radial)
					getSatelliteOrientationACR (position[1], velocity[1], orientationACR);
					changeBase(orientationACR,posDiff);
	
					// Compute orbit and clock differences
					if (options->satelliteClockCorrection) {
						ClkDiff = clock[1] - clock[0];
						SISREDiff = sqrt((posDiff[2]-ClkDiff)*(posDiff[2]-ClkDiff) + (posDiff[0]*posDiff[0] + posDiff[1]*posDiff[1])/49);
					} else {
						ClkDiff = 0;
						SISREDiff = 0;
					}
					OrbDiff = sqrt((posDiff[2])*(posDiff[2]) + (posDiff[0]*posDiff[0] + posDiff[1]*posDiff[1])/49);
					Orb3DDiff = sqrt(posDiff[0]*posDiff[0] + posDiff[1]*posDiff[1] + posDiff[2]*posDiff[2]);
					
					addStat(&statSISRE[k][i],SISREDiff);
					addStat(&statSISRETot,SISREDiff);
					addStat(&statOrb[k][i],OrbDiff);
					addStat(&statOrbTot,OrbDiff);
					addStat(&stat3D[k][i],Orb3DDiff);
					addStat(&stat3DTot,Orb3DDiff);
					addStat(&statClk[k][i],ClkDiff);
					addStat(&statClkTot,ClkDiff);
					addStat(&statRadOrb[k][i],posDiff[2]);
					addStat(&statRadOrbTot,posDiff[2]);
					addStat(&statAloOrb[k][i],posDiff[0]);
					addStat(&statAloOrbTot,posDiff[0]);
					addStat(&statCroOrb[k][i],posDiff[1]);
					addStat(&statCroOrbTot,posDiff[1]);
					
					printSatDiff(&t,k,i,SISREDiff,OrbDiff,Orb3DDiff,ClkDiff,posDiff,options);
				}
			}
		}
		t = tdadd(&t,tIncrease);
	}
	
	if (options->printSatStat) {
		for (k=0;k<MAX_GNSS;k++) {
			for (i=0;i<MAX_SATELLITES_PER_GNSS;i++) {
				printSatStat(k,i,statSISRE[k][i],statOrb[k][i],stat3D[k][i],statClk[k][i],statRadOrb[k][i],statAloOrb[k][i],statCroOrb[k][i],options);
			}
		}
	}
	
	printSatStatTot(statSISRETot,statOrbTot,stat3DTot,statClkTot,statRadOrbTot,statAloOrbTot,statCroOrbTot,options);
}


void showProduct (TGNSSproducts *product, TConstellation *constellation, TOptions *options, TSupportOptions *sopt) {
	double		tIncrease = 30;
	TTime		startTime;
	TTime		endTime;
	TTime		t;
	int			i,k;
	int			available;
	int			ret;
	double		position[3];
	double		velocity[3];
	double		ITRFvelocity[3];
	double		clock;
	double		dr[3];
	double		orientation[3][3];
	
	sprintf(messagestr,"OUTPUT Satellite Velocity:     %s",options->satVel==svINERTIAL?"Inertial":"ITRF");
	printInfo(messagestr,options);
	
	// Use decimate option to choose increase interval
	if (options->decimate != 0) {
		tIncrease = options->decimate;
	}
	
	// Get start and end times
	startTime = getProductsFirstEpoch(product);
	endTime = getProductsLastEpoch(product);
	
	// Tune start time to start at least in an entire hour
	startTime.SoD -= modulo(startTime.SoD,3600);
	
	sprintf(messagestr,"StartTime: %5d %8.2lf    EndTime: %5d %8.2lf    IntervalTime: %8.2lf",startTime.MJDN,startTime.SoD,endTime.MJDN,endTime.SoD,tIncrease);
	printInfo(messagestr,options);
	
	memcpy(&t,&startTime,sizeof(TTime));
	while(tdiff(&t,&endTime)<0) {
		for (k=0;k<MAX_GNSS;k++) {
			for (i=0;i<MAX_SATELLITES_PER_GNSS;i++) {
				available=1;
				// Check if satellite has been deselected
				if (!options->includeSatellite[k][i]) available = 0;
				// Satellite positions and velocities
				ret = getSatellitePVT(product,&t,0.0,k,i,position,velocity,ITRFvelocity,NULL,options);
				if (!ret) available = 0;
				// Satellite clock
				ret = getSatellitePVT(product,&t,0.0,k,i,NULL,NULL,NULL,&clock,options);
				//if (!ret) available = 0;
				// Satellite phase center correction
				if (product->type == SP3 && options->satellitePhaseCenter) {
					// Satellite orientation
					getSatelliteOrientation (&t, position, orientation);
					ret = satellitePhaseCenterCorrection3D(&t,k,i,C1C,orientation,constellation,dr);
					if (!ret) available = 0;
					// Coordinate correction
					position[0] += dr[0];
					position[1] += dr[1];
					position[2] += dr[2];
				}
				if (available) {
					if (options->satVel == svINERTIAL) {
						printSatPvt(&t,k,i,position,velocity,clock,options);
					} else {
						printSatPvt(&t,k,i,position,ITRFvelocity,clock,options);
					}
				}
			}
		}
		t = tdadd(&t,tIncrease);
	}
	
}

/*****************************************************************************
 * Name        : productsProcessing
 * Description : Do the two different kind of processes, depending on TOptions:
 *                 + Performing an orbit comparison between two different products
 *                 + Parse a single product file
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TSupportOptions  *sopt          I  N/A  TSupportOptions structure
 * TOptions  *options              I  N/A  TOptions structure
 * TConstellation  *constellation  I  N/A  TConstellation structure
 *****************************************************************************/
void productsProcessing (TOptions *options, TSupportOptions *sopt, TConstellation *constellation) {
	TGNSSproducts	products[2];
	int				sources;
	int				ret;
	int				i;
	
	initGNSSproducts(&products[0]);
	initGNSSproducts(&products[1]);
	
	sources = 0;
	for (i=0;i<sopt->numNav;i++) {
		ret = readRinexNavFile(sopt->navFile[i],&products[sources]);
		if (ret!=1) {
			sprintf(messagestr,"Reading RINEX navigation message file '%s'",sopt->navFile[i]);
			printError(messagestr,options);
		}
		sources++;
	}
	for (i=0;i<sopt->numSP3;i++) {
		ret = readSP3File(sopt->SP3File[i],&products[sources],1,1);
		if (ret!=1) {
			sprintf(messagestr,"Reading SP3 orbit and clock products file '%s'",sopt->SP3File[i]);
			printError(messagestr,options);
		}
		sources++;
	}
	for (i=0;i<sopt->numOrb;i++) {
		ret = readSP3File(sopt->orbFile[i],&products[sources],1,0);
		if (ret!=1) {
			sprintf(messagestr,"Reading SP3 orbit products file '%s'",sopt->orbFile[i]);
			printError(messagestr,options);
		}
		ret = readRinexClocksFile(sopt->clkFile[i],&products[sources]);
		if (ret!=1) {
			sprintf(messagestr,"Reading clock products file '%s'",sopt->clkFile[i]);
			printError(messagestr,options);
		}
		sources++;
	}
	
	sprintf(messagestr,"Satellite Phase Center Modeling:     %3s",options->satellitePhaseCenter?"YES":"NO");
	printInfo(messagestr,options);
		
	if (options->workMode==wmCOMPAREORBITS) {
		compareOrbits (products, constellation, options, sopt);
	} else if (options->workMode==wmSHOWPRODUCT) {
		showProduct (&products[0], constellation, options, sopt);
	}
}

/*****************************************************************************
 * Name        : main
 * Description : Main executable function
 *****************************************************************************/
int main(int argc, char* argv[]) {
	TOptions					options;
	TSupportOptions				sopt;
	TConstellation				constellation;
	TAntennaList				antennaList;
	TStationList				stationList;
	TReceiverList 				recList;
	TAntenna					*ant;
	TReceiverData				*recD;
	TStation					*station;
	TGNSSproducts				products;
	TGNSSproducts				productsKlb;
	TGNSSproducts 		  		productsTgd;
	TTGDdata					tgdData;
	TEpoch						epoch;
	FILE						*fdRNX;
	int							ret;
	int							i,j;
	double						r;
	int							kalmanRes;
	enum fileType				auxft1,auxft2;
	int							processedEpochs;
	enum ProcessingDirection	direction;
	int							filterIterations;
	TTime						endProductsTime;
	TTime						startProductsTime;
	int							anyInsideProducts = 0;
  int             kalmanIterations;
	
	// Filter
	TFilterSolution	solution;
	TUnkinfo		prevUnkinfo;
	int				PRNlist[MAX_SATELLITES_PER_GNSS];  // Only GPS!
		
	// Initialize
	initSupportOptions(&sopt);
	initOptions(&options);
	initEpoch(&epoch);
	initGNSSproducts(&products);
	initGNSSproducts(&productsKlb);
	initGNSSproducts(&productsTgd);
	initConstellation(&constellation);
	initFilterSolution(&solution);
	initAntennaList(&antennaList);
	initTGD(&tgdData,p1c1NONE,p1p2NONE);
	initReceiverList(&recList);
	initStationList(&stationList);
	
	// Parse options
	if (!parseAllOptions (&sopt,&options, argc, argv)) {
		exit (0);
	}
	
	for (i=0;i<2;i++) { 
		// This loop is done twice in order to override defaults of PPP and 
		// standalone once options are processed
		sopt.numNav = 0;
		sopt.numSP3 = 0;
		sopt.numOrb = 0;
		sopt.numClk = 0;
		sopt.numTGD = 0;
		
		// Check if config file parameter exists
		ret = checkConfigFile(&sopt,&options);
		if (ret == -1) {
			sprintf(messagestr,"Problems with config file '%s'",sopt.cfgFile);
			printError(messagestr,&options);
		} else { // config file parameter found and file parsed properly
			// Parse all command-line options again
			if (!parseAllOptions (&sopt,&options, argc, argv)) {
				exit (0);
			}
		} 
		// Process options
		if (i==0) {
			if (!processOptions (&sopt,&options)) {
				exit (0);
			}
		}
	}
	
	// Read files
	if (sopt.obsFile[0] != '\0') {
		if (!(fdRNX = fopen(sopt.obsFile,"r")) || (whatFileTypeIs(sopt.obsFile)!=ftRINEXobservation)) {
			sprintf(messagestr,"Reading RINEX observation file '%s'",sopt.obsFile);
			printError(messagestr,&options);
		}
	}
	
	// Satellite Constellation Data
	// Priority is given to the ANTEX file
	if (sopt.antFile[0]!='\0'){ // ANTEX file
		if (readAntexFile(sopt.antFile,&constellation,&antennaList)!=1) {
			sprintf(messagestr,"Reading ANTEX file '%s'",sopt.antFile);
			printError(messagestr,&options);
		}
	} else if (sopt.conFile[0]!='\0') { // Constellation Status file
		if (readConstellationFile(sopt.conFile,&constellation)!=1) {
			sprintf(messagestr,"Reading constellation file '%s'",sopt.conFile);
			printError(messagestr,&options);
			exit (0);
		}
	} 
	
	// Processing data
	if (options.workMode==wmCOMPAREORBITS || options.workMode==wmSHOWPRODUCT) {
		productsProcessing(&options,&sopt,&constellation);
	} else if (options.workMode==wmDOPROCESSING || options.workMode==wmSHOWINPUT) {
		if (sopt.navFile[0][0]!='\0') {
			if (readRinexNavFile(sopt.navFile[0],&products)!=1) {
				sprintf(messagestr,"Reading RINEX navigation message file '%s'",sopt.navFile[0]);
				printError(messagestr,&options);
			}
		}
		if (sopt.SP3File[0][0]!='\0') {
			if (readSP3File(sopt.SP3File[0],&products,1,1)!=1) {
				sprintf(messagestr,"Reading SP3 orbit and clock products file '%s'",sopt.SP3File[0]);
				printError(messagestr,&options);
			}
		}
		if (sopt.orbFile[0][0]!='\0' && sopt.clkFile[0][0]!='\0') {
			if (readSP3File(sopt.orbFile[0],&products,1,0)!=1) {
				sprintf(messagestr,"Reading SP3 orbit products file '%s'",sopt.orbFile[0]);
				printError(messagestr,&options);
			}
			if (readRinexClocksFile(sopt.clkFile[0],&products)!=1) {
				sprintf(messagestr,"Reading clock products file '%s'",sopt.clkFile[0]);
				printError(messagestr,&options);
			}
		}
		if (options.ionoModel==Klobuchar) {
			if (sopt.klbFile[0]!='\0') {
				if (readRinexNavFile(sopt.klbFile,&productsKlb)!=1) {
					sprintf(messagestr,"Reading RINEX navigation message file for Klobuchar data '%s'",sopt.klbFile);
					printError(messagestr,&options);
				}
			} else if (products.type==BRDC) {
				productsKlb.BRDC = products.BRDC;
			} else {
				printError("Klobuchar model selected for ionospheric corrections, but no source is found",&options);
			}
		}
		
		// TGD file management
		initTGD(&tgdData,options.p1c1dcbModel,options.p1p2dcbModel);
		if (options.p1c1dcbModel==p1c1STRICT) {
		      if (sopt.numP1C1DCB==0) {
			printError("No P1-C1 DCB file provided, but P1-C1 model set to strict",&options);
		      }
		      if (sopt.recFile[0]=='\0') {
			printError("No 'GPS_Receiver_Types' file provided, but P1-C1 model set to strict, include it with '-input:rec'",&options);
		      }
		}
		
		if (options.p1p2dcbModel==p1p2RINEX) {
		      if (sopt.numP1P2RINEX==0) {
			if (products.type!=BRDC) {
			  printError("Navigation message is selected for P1-P2 DCB corrections, but no source is found, include it with '-input:dcb'",&options);
			} else {
			  tgdData.BRDC = products.BRDC;
			}
		      }
		} else if (options.p1p2dcbModel==p1p2DCB) {
		      if (sopt.numP1P2DCB==0) {
			printError("No P1-P2 DCB file provided, but P1-P2 model set to DCB",&options);
		      }
		}
		
		for (i=0;i<sopt.numTGD;i++) {
		      auxft1 = whatFileTypeIs(sopt.tgdFile[i]);
		      if ((auxft1 == ftP1C1DCB && options.p1c1dcbModel==p1c1STRICT) || (auxft1 == ftP1P2DCB && options.p1p2dcbModel==p1p2DCB)) {
			if (readDCBFile(sopt.tgdFile[i],tgdData.DCB)!=1) {
					    sprintf(messagestr,"Reading DCB file for TGD data '%s'",sopt.tgdFile[i]);
					    printError(messagestr,&options);
			  	  }
		      } else if (auxft1 == ftRINEXbroadcast && options.p1p2dcbModel==p1p2RINEX) {
			if (readRinexNavFile(sopt.tgdFile[i],&productsTgd)!=1) {
					    sprintf(messagestr,"Reading RINEX navigation message file for TGD data '%s'",sopt.tgdFile[i]);
					    printError(messagestr,&options);
			  	  }
			tgdData.BRDC = productsTgd.BRDC;
		      }
		}

		// GPS Receiver type file
		if (sopt.recFile[0]!='\0') {
			if (sopt.setRecType==rtNA) {
				if (readRecTypeFile(sopt.recFile,&recList)!=1) {
					sprintf(messagestr,"Reading GPS Receiver type file '%s'",sopt.recFile);
					printError(messagestr,&options);
				}
			} else {
				sprintf(messagestr,"WARNING setrectype specified, ignoring GPS receiver types file (%s)",sopt.recFile);
				printInfo(messagestr,&options);
			}
		}
		
		// SINEX file
		if (sopt.snxFile[0]!='\0') {
			if (readSINEXFile(sopt.snxFile,&stationList)!=1) {
				sprintf(messagestr,"Reading SINEX input file '%s'",sopt.snxFile);
				printError(messagestr,&options);
			}
		}
		
		// If receiver antenna phase center is set, dump it to the antenna structure
		if (options.antennaData==adSET) {
			for (j=0;j<MAX_FREQUENCIES_PER_GNSS;j++) {
				for (i=0;i<3;i++) {
					epoch.receiver.antenna.PCO[j][i] = options.receiverPCO[j][i];
				}
			}
		}
		
		// If receiver antenna reference point is set, dump it to the receiver structure
		if (options.ARPData==arpSET) {
			for (i=0;i<3;i++) {
				epoch.receiver.ARP[i] = options.receiverARP[i];
			}
		}
				
		// Read RINEX Observation header
		ret = readRinexObsHeader(fdRNX,&epoch,&options);
		
		// If position receiver is set, dump it to the receiver structure 
		if (options.receiverPositionSource==rpSET) {
			for (i=0;i<3;i++)
				epoch.receiver.aproxPosition[i] = options.aprioriReceiverPosition[i];
			epoch.receiver.aproxPositionError = 1e4; // 10 Km
		} else if (options.receiverPositionSource==rpSINEX) { // If the position is read from SINEX, get it
			station = getStation(epoch.receiver.name,&stationList);
			if (station==NULL) {
				sprintf(messagestr,"Receiver marker name not found in SINEX file [%s]. Use '-pre:setrecpos x y z' to set it",epoch.receiver.name);
				printError(messagestr,&options);
			} else {
				for (i=0;i<3;i++)
					epoch.receiver.aproxPosition[i] = station->pos[i];
				epoch.receiver.aproxPositionError = 0.1; // 1 cm
			}
		}
		
		// Abnormal a priori receiver position
		// Check if station is at least above than -1 km from the earth surface 
		r = sqrt(epoch.receiver.aproxPosition[0]*epoch.receiver.aproxPosition[0] + 
		         epoch.receiver.aproxPosition[1]*epoch.receiver.aproxPosition[1] + 
		         epoch.receiver.aproxPosition[2]*epoch.receiver.aproxPosition[2]);
		if (r<(MIN_EARTH_RADIUS-1000)) {
			sprintf(messagestr,"Receiver a priori position must be above Earth surface (%.2lf vs %.2lf).",r,MIN_EARTH_RADIUS-1000.0);
			printError(messagestr,&options);
		} 
		
		
		// If receiver antenna phase center is from ANTEX file, look for it
		if (options.antennaData==adANTEX) {
			ant = getAntenna(epoch.receiver.antenna.type,&antennaList,&options);
			if (ant==NULL) {
				sprintf(messagestr,"Receiver antenna phase center not found in ANTEX file [%s]. Use '-model:recphasecenter no' to disable it",epoch.receiver.antenna.type);
				printError(messagestr,&options);
			} else {
				for (j=0;j<MAX_FREQUENCIES_PER_GNSS;j++) {
					for (i=0;i<3;i++) {
						epoch.receiver.antenna.PCO[j][i] = ant->PCO[j][i];
					}
				}
			}
		}
		
		// If GPS Receiver types file was available, classify the receiver
		if (sopt.setRecType==rtNA) {
			if (sopt.recFile[0] != '\0') {
				recD = getReceiverType(epoch.receiver.type,&recList);
				if (recD==NULL) {
					// Added a further Check: Only print error is strict option enabled!
					if (options.p1c1dcbModel==p1c1STRICT) {
				  		sprintf(messagestr,"Strict P1-C1 DCB option enabled, but receiver type not found in GPS Receiver type file [%s]",epoch.receiver.type);
						printError(messagestr,&options); 
					}
				} else {
					epoch.receiver.recType = recD->type;
				}
			} 
		} else {
			epoch.receiver.recType = sopt.setRecType;
		}
		if (epoch.receiver.recType == rtUNKNOWN) {
			if (sopt.recFile[0]=='\0') {
				printInfo("WARNING Receiver type is unknown (no GPS_Receiver_type file provided). Assuming nominal behaviour.",&options);
			} else {
				printInfo("WARNING Receiver type is unknown. Assuming nominal behaviour",&options);
			}
		}
		
		// Flexible C1/P1 (C1C/C1P) handling
		if (options.flexibleC1P1) {
			if ((epoch.measOrder[GPS].meas2Ind[C1C]==-1 && epoch.measOrder[GPS].meas2Ind[C1P]!=-1) || 
			    (epoch.measOrder[GPS].meas2Ind[C1C]!=-1 && epoch.measOrder[GPS].meas2Ind[C1P]==-1)) {
				epoch.receiver.equivalentC1P1 = 1;
				printInfo("Making equivalence: P1==C1",&options);
			}
		}
		
		// Print input parameters
		printParameters(&options,&epoch,&products);

		// Set products end time
		if (options.workMode==wmDOPROCESSING) {
			endProductsTime = getProductsLastEpoch(&products);
			startProductsTime = getProductsFirstEpoch(&products);
		}
		processedEpochs = 0;
		direction = pFORWARD;
		filterIterations = 1;
		while (ret) {
			// Determination of Backward/Forward
			if (filterIterations%2==1) direction = pFORWARD;
			else direction = pBACKWARD;
			
			ret = readRinexObsEpoch(fdRNX,&epoch,&constellation,direction);
			if (ret) {
				if (tdiff(&epoch.t,&endProductsTime)>0 && options.filterIterations>1) ret = 0; 
			}
			if (ret) {
				preprocess(&epoch,&options);
				if (!isEpochDecimated(&epoch,&options)) {
					printInput(&epoch,&options);
					if (options.workMode==wmDOPROCESSING) {
						if (!anyInsideProducts) {
							if (tdiff(&epoch.t,&startProductsTime)>=0 && tdiff(&epoch.t,&endProductsTime)<=0) anyInsideProducts = 1;
						}
						kalmanRes=2;
            					kalmanIterations=0;
						while (kalmanRes==2) {
						        kalmanIterations++;
						        if (kalmanIterations > options.maxKalmanIterations) { printError ("No convergence to solution",&options);      }
							modelEpoch(&epoch,&options);
							for (i=0;i<epoch.numSatellites;i++) {
								modelSatellite(&epoch,i,&products,&productsKlb,&tgdData,&constellation,&options);
							}
							printMeas(&epoch,&options);
							kalmanRes = Kalman(&epoch,&solution,PRNlist,&prevUnkinfo,&options);
							if (kalmanRes==1) {
								processedEpochs++;
								printOutput(&epoch,&solution,&options,&prevUnkinfo);
							}
						}
					} else if (options.workMode==wmSHOWINPUT){
						 printMeas(&epoch,&options);
					}
				}
			} else { // ret == 0
				if (options.filterIterations >= (filterIterations+1)) {
					filterIterations++;
					if (filterIterations%2==0) { // Backward direction
						rewindEpochRinexObs(fdRNX,epoch.source);
					}
					ret = 1;
				}
			}
		}
		if (!anyInsideProducts && options.workMode==wmDOPROCESSING) 
			printError ("Orbit and clock products are of a different time than the observation RINEX",&options);
	}
	
	return 1;
}
