/****************************************************************************
   Copyright & License:
   ====================
   
   Copyright 2009 - 2010 gAGE/UPC & ESA
   
   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at
   
   http://www.apache.org/licenses/LICENSE-2.0
   
   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 ***************************************************************************/

/****************************************************************************
 * Copyright: gAGE/UPC & ESA
 * Project: EDUNAV GNSS Lab Tool 
 * Author: Pere Ramos-Bosch (group of Astronomy and GEomatics - gAGE/UPC)
 * Maintainer: Adria Rovira-Garcia ( gAGE/UPC )
 *             glab.gage @ upc.edu      
 * File: input.c
 * Code Management Tool File Version: 2.0  Revision: 0 
 * Date: 2010/11/30
 ***************************************************************************/

/****************************************************************************
 * MODULE DESCRIPTION
 *
 * Name: input
 * Language: C
 *
 * Purpose:
 *  The purpose of this module is to serve as "driver" from the different ways
 *  that the data can be stored in the form of files and the rest of the program
 *
 * Function:
 *  This class is in charge of reading the different data formats in which the 
 *  data can be stored, and save it to the internal data structures.
 *
 * Dependencies:
 *  dataHandling
 *
 * Files modified:
 *  None
 *
 * Files read:
 *  Depending on function
 *
 * Resources usage:
 *  See Design Document
 ****************************************************************************/

/****************************
 *       RELEASE_HISTORY
 * -------------
 *          gLAB v1.3.1
 * Release: 2010/01/22
 * Change Log: First public version.
 * -------------
 *          gLAB v1.3.2
 * Release: 2010/02/23
 * Change Log: Corrected problem of not properly reading years below 2000
 *                  for RINEX observation and navigation files (Y2K effect).
 *             Changed whatFileTypeIs to improve detection of SP3 files.
 *             Fine-tune in automatically detecting file types.
 *             Security check to better read RINEX clock files not following 
 *                  the standard.
 *             Included input files: SINEX, DCB and GPS_Receiver_Type.
 *             Forward/Backward processing included.
 * -------------
 *          gLAB v1.4.0
 * Release: 2010/06/21
 * Change Log: Included usable measurement index in TMeasOrder structure to select/deselect
 *                  individual measurements.
 *             Corrected a bug when reading Galileo observations in a RINEX v2.11 file.
 *             Corrected a bug in elevation/azimuth initialisation, which appeared when new 
 *                  satellites without ephemeris data appeared.
 * -------------
 *          gLAB v1.4.1
 * Release: 2010/07/09
 * Change Log: No changes in this file.
 * -------------
 *          gLAB v1.4.2
 * Release: 2010/07/31
 * Change Log: No changes in this file.
 * -------------
 *          gLAB v1.4.3
 * Release: 2010/08/31
 * Change Log: No changes in this file.
 * --------------
 *          gLAB v1.4.4
 * Release: 2010/09/22
 * Change Log: Included new DCB management.
 * -------------
 *          gLAB v1.4.5
 * Release: 2010/10/31
 * Change Log: No changes in this file.
 * -----------
 *          gLAB v2.0.0
 * Release: 2010/11/30
 * Change Log: gLAB version released together with the Educational Book:
 *             "Global Navigation Satellite Systems: Fundamentals, Algorithms and Laboratory Exercises"
 *             Authors: J. Sanz, JM. Juan and M. Hernandez-Pajares.
 * -----------
 *       END_RELEASE_HISTORY
 *****************************/

/* External classes */
#include "input.h"

/**************************************
 * Declarations of internal operations
 **************************************/

/*****************************************************************************
 * Name        : getL
 * Description : Standard multipurpose input reader from files. Valid for 
 *               windows and linux files.
 * Parameters  :
 * Name                           |Da|Unit|Description
 * char  *lineptr                  O  N/A  Buffer to write to
 * int  *n                         O  N/A  Length of buffer
 * FILE  *stream                   I  N/A  Input stream file
 * Returned value (int)            O  N/A  Return the number of charaters read
 *                                         -1 on failure to read
 *****************************************************************************/
int getL(char *lineptr, int *n, FILE *stream) {
	int		rd;
	
	*n = 0;
	
	while ((rd=fgetc(stream))!=EOF) {
		if (rd==13 || rd=='\0' || rd=='\n') { // Character 13 is ^M (for Windows)
			lineptr[*n] = rd;
			if (rd!='\0') {
				(*n)++;
				lineptr[*n]='\0';
			}
			break;
		} else {
			lineptr[*n] = rd;
			(*n)++;
		}
	}
	
	if (rd==13) fgetc(stream);
	if (rd==EOF && (*n)==0) return -1;
	else return (*n); 
}

/*****************************************************************************
 * Name        : getLback
 * Description : Standard multipurpose input reader from files. Valid for 
 *               windows and linux files. It reads a line backwards.
 * Parameters  :
 * Name                           |Da|Unit|Description
 * char  *lineptr                  O  N/A  Buffer to write to
 * int  *n                         O  N/A  Length of buffer
 * FILE  *stream                   I  N/A  Input stream file
 * Returned value (int)            O  N/A  Return the number of charaters read
 *                                         -1 on failure to read
 *****************************************************************************/
int getLback (char *lineptr, int *n, FILE *stream) {
	int  	res;
	int		enters = 0;
	int		rd;
	char	buffer[MAX_INPUT_LINE];
	int		i;
	long	pos;
	int		seek;
	char	*enter1;
	char	*enter2;
	
	pos = ftell(stream);
	if (pos<BACKWARD_READ_BATCH) seek = pos;
	else seek = BACKWARD_READ_BATCH;
	buffer[seek] = '\0';
	//res = fseek(stream,-seek,SEEK_CUR);
	res = fseek(stream,pos-seek,SEEK_SET);
	for (i=0;i<seek;i++) {
		rd = fgetc(stream);
		buffer[i] = rd;
		if (rd=='\n') {
			enter2 = enter1;
			enter1 = &buffer[i];
			enters++;
		}
	}
	if (enters>=2) {
		*n=(int)(enter1-enter2);
		strncpy(lineptr,enter2+1,*n);
	} else return -1;
	lineptr[*n]='\0';
	
	//res = fseek(stream,-*n,SEEK_CUR);
	res = fseek(stream,pos-*n,SEEK_SET);
	return *n;
}

/*****************************************************************************
 * Name        : whatFileTypeIs
 * Description : Identifies the file type
 * Parameters  :
 * Name                           |Da|Unit|Description
 * char  *filename                 I  N/A  Name of the file
 * Returned value (fileType)       O  N/A  Type of file
 *****************************************************************************/
enum fileType whatFileTypeIs (char *filename) {
	FILE			*fd = NULL;
	char 			line[MAX_INPUT_LINE];
	int 			len = 0;
	int				res;
	char			aux[100];
	enum fileType	ft;

	ft = ftUNKNOWN;	// Default value

	fd = fopen(filename,"r");
	
	if (fd==NULL) return ft;
	
	res = getL(line,&len,fd); 
	
	if (res!=-1) {
		getstr(aux,line,60,13);
		// RINEX check
		if (strcmp(aux,"RINEX VERSION")==0) {
			getstr(aux,line,20,11);
			if (strncmp(aux,"O",1)==0) ft = ftRINEXobservation;
			else if (strncmp(aux,"N",1)==0) ft = ftRINEXbroadcast;
			else if (strncmp(aux,"C",1)==0) ft = ftRINEXclocks;
			else if (strncmp(aux,"I",1)==0) ft = ftIONEX;
		// ANTEX check
		} else if (strcmp(aux,"ANTEX VERSION")==0) {
			ft = ftANTEX;
		// SP3 check
		} else if (line[0]=='#' &&  ((line[1]>='a' && line[1]<='z') || line[1]==' ') && (line[2]=='P' || line[2]=='V' || line[2]==' ')) {
			ft = ftSP3;
		// DCB check
		} else if (strstr(line,"DCB SOLUTION")!=NULL) {
      if (strstr(line,"P1-C1")!=NULL) {
        ft = ftP1C1DCB;
      } else if (strstr(line,"P1-P2")!=NULL) {
        ft = ftP1P2DCB;
      } else {
			  ft = ftGENDCB;
      }
		// GPS Receiver file type
		} else if (strncmp(line,"#$Id: GPS_Receiver_Types",24)==0) {
			ft = ftRECEIVERTYPE;
		// SINEX check
		} else if (strncmp(line,"%=SNX",5)==0) {
			ft = ftSINEX;
		} else {
		// GPS Constellation Status file
			res = getL(line,&len,fd); 
			if (res!=-1) {
				if (strstr(line,"Navstar GPS Constellation Status")!=NULL) ft = ftConstellation;
			}
		}
		if (fd!=NULL) fclose(fd);
		return ft;
	}

	if (fd!=NULL) fclose(fd);
	return ftINVALID;
}

/*****************************************************************************
 * Name        : readRinexObsHeader
 * Description : Read a RINEX observation header from a file descriptor
 * Parameters  :
 * Name                           |Da|Unit|Description
 * FILE  *fd                       I  N/A  File descriptor
 * TEpoch  *epoch                  O  N/A  Structure to save the data
 * TOptions  *options              I  N/A  TOptions structure
 * Returned value (int)            O  N/A  Status of the function
 *                                         1 => Properly read
 *                                         0 => Error
 *****************************************************************************/
int readRinexObsHeader (FILE *fd, TEpoch *epoch, TOptions *options) {
	char 		line[MAX_INPUT_LINE];
	char    	aux[100];
	char		*ca;
	int 		len = 0;
	double		rinexVersion;
	int			properlyRead = 0;
	int			error = 0;
	int			readingHeader = 1;
	int			obs,i,j,k;
	int			auxN;
	enum MeasurementType	meas;

	while(getL(line,&len,fd)!=-1 && readingHeader) {
		if (strncmp(&line[60],"RINEX VERSION / TYPE",20)==0) {
			getstr(aux,line,0,9);
			rinexVersion = atof(aux);
			if ((int)(rinexVersion)==2) epoch->source = RINEX2;
			else if ((int)(rinexVersion)==3) epoch->source = RINEX3;
			else epoch->source = UNKNOWN_SOURCE;
		} else if (strncmp(&line[60],"MARKER NAME",11)==0) {
			getstr(aux,line,0,60);
			strcpy(epoch->receiver.name,aux);
		} else if (strncmp(&line[60],"ANT # / TYPE",12)==0) {
			getstr(aux,line,20,20);
			strcpy(epoch->receiver.antenna.type,aux);
			// Separate the anntena.type into antenna and radome
			strcpy(epoch->receiver.antenna.type_ant,epoch->receiver.antenna.type);
			ca = strstr(epoch->receiver.antenna.type_ant," ");
			if (ca != NULL) {
				*ca = '\0';
				strcpy(epoch->receiver.antenna.type_radome,&ca[1]);
				trim(epoch->receiver.antenna.type_radome);
			}
		} else if (strncmp(&line[60],"REC # / TYPE / VERS",19)==0) {
			getstr(aux,line,20,20);
			strcpy(epoch->receiver.type,aux);
		} else if (strncmp(&line[60],"ANTENNA: DELTA H/E/N",20)==0) {
			if (options->ARPData == arpRINEX) {
			getstr(aux,line,0,14);
				epoch->receiver.ARP[2] = atof(aux);
				getstr(aux,line,14,14);
				epoch->receiver.ARP[1] = atof(aux);
				getstr(aux,line,28,14);
				epoch->receiver.ARP[0] = atof(aux);
			}
		} else if (strncmp(&line[60],"# / TYPES OF OBSERV",19)==0) {  // Only for RINEX v2.11
			//      5    L1    L2    P1    P2    C1                        # / TYPES OF OBSERV
			getstr(aux,line,0,6);
			if (aux[0]!='\0') { // if this is the first line of observations
				auxN = atoi(aux);
				epoch->measOrder[GPS].nDiffMeasurements = auxN;
				epoch->measOrder[GPS].GNSS = GPS;
				epoch->measOrder[Galileo].nDiffMeasurements = auxN;
				epoch->measOrder[Galileo].GNSS = Galileo;
				epoch->measOrder[GEO].nDiffMeasurements = auxN;
				epoch->measOrder[GEO].GNSS = GEO;
				epoch->measOrder[GLONASS].nDiffMeasurements = auxN;
				epoch->measOrder[GLONASS].GNSS = GLONASS;
				epoch->numGNSS = 4;
				obs = 0;
			} 
			obs = 0;
			for (i=0,j=0;i<auxN;i++,j++) {
				getstr(aux,line,10+j*6,2);
				meas = measstr2meastype(aux);
				if (meas != NA) {
					// Carrier-phase are stored in meters instead of cycles, as
					// they are on RINEX files.
					//    aux[0]=='L'   => Measurement is carrier phase
					for (k=0;k<epoch->numGNSS;k++) {
						if (aux[0]=='L') {
							if (epoch->measOrder[k].GNSS==GPS) {
								if (aux[1]=='1') epoch->measOrder[k].conversionFactor[meas] = GPSl1;
								else if (aux[1]=='2') epoch->measOrder[k].conversionFactor[meas] = GPSl2;
								else if (aux[1]=='5') epoch->measOrder[k].conversionFactor[meas] = GPSl5;
							} else if (epoch->measOrder[k].GNSS==Galileo) {
								if (aux[1]=='1') epoch->measOrder[k].conversionFactor[meas] = GALl1;
								else if (aux[1]=='5') epoch->measOrder[k].conversionFactor[meas] = GALl5a;
								else if (aux[1]=='7') epoch->measOrder[k].conversionFactor[meas] = GALl5b;
								else if (aux[1]=='8') epoch->measOrder[k].conversionFactor[meas] = GALl5;
								else if (aux[1]=='6') epoch->measOrder[k].conversionFactor[meas] = GALl6;
							} else if (epoch->measOrder[k].GNSS==GEO) {
								if (aux[1]=='1') epoch->measOrder[k].conversionFactor[meas] = SBASl1;
								else if(aux[1]=='5') epoch->measOrder[k].conversionFactor[meas] = SBASl5;
							} else if (epoch->measOrder[k].GNSS==GLONASS) {
								epoch->measOrder[k].conversionFactor[meas] = 1; // In GLONASS carrier phase is given in cycles
							}
						} else { // If measurement in not carrier-phase, conversionFactor is 1
							epoch->measOrder[k].conversionFactor[meas] = 1.0;
						}
						// Update dictionaries (if usableFreq is disabled for a specific frequency, the dictionary meas2Ind is avoided)
						if (options->usableFreq[k][getFrequencyInt(meas)]) {
							epoch->measOrder[k].usable[meas] = 1;
						} else {
							epoch->measOrder[k].usable[meas] = 0;
						}
						epoch->measOrder[k].meas2Ind[meas] = obs;
						epoch->measOrder[k].ind2Meas[obs] = meas;
					}
				} else {
					sprintf(messagestr,"Reading RINEX 2.11 file: Unrecognized measurement type: '%s'",aux);
					printError(messagestr,options);
					error = 1;
				}				
				obs++;
				if ((j+1)%9==0) { // Following observables are in the next lines (max 9 per line)
					j=-1; // j will be increased in the for loop
					getL(line,&len,fd);
				}
			}
		} else if (strncmp(&line[60],"SYS / # / OBS TYPES",19)==0) { // Rinex version 3.00 can have multiple systems
			// G    4 C1C L1C D1C S1C                                      SYS / # / OBS TYPES
			// E   36 C1B L1B D1B S1B C1C L1C D1C S1C C5I L5I D5I S5I C5Q  SYS / # / OBS TYPES
			//        L5Q D5Q S5Q C7I L7I D7I S7I C7Q L7Q D7Q S7Q C6A L6A  SYS / # / OBS TYPES
			//        D6A S6A C6B L6B D6B S6B C6C L6C D6C S6C              SYS / # / OBS TYPES
			getstr(aux,line,1,5);
			if (aux[0]!='\0') { // if this is the first line of a SYS
				epoch->measOrder[epoch->numGNSS].nDiffMeasurements = atoi(aux);
				epoch->measOrder[epoch->numGNSS].GNSS = gnsschar2gnsstype(line[0]);
				epoch->numGNSS++;
				obs = 0;
			}
			for (i=0,j=0;i<epoch->measOrder[epoch->numGNSS-1].nDiffMeasurements;j++,i++) {
				getstr(aux,line,7+4*j,3);
				meas = measstr2meastype(aux);
				if (meas != NA) {
					// Carrier-phase are stored in meters instead of cycles, as
					// they are on RINEX files.
					//    aux[0]=='L'   => Measurement is carrier phase
					if (aux[0]=='L') {
						if (epoch->measOrder[epoch->numGNSS-1].GNSS==GPS) {
							if (aux[1]=='1') epoch->measOrder[epoch->numGNSS-1].conversionFactor[meas] = GPSl1;
							else if (aux[1]=='2') epoch->measOrder[epoch->numGNSS-1].conversionFactor[meas] = GPSl2;
							else if (aux[1]=='5') epoch->measOrder[epoch->numGNSS-1].conversionFactor[meas] = GPSl5;
						} else if (epoch->measOrder[epoch->numGNSS-1].GNSS==Galileo) {
							if (aux[1]=='1') epoch->measOrder[epoch->numGNSS-1].conversionFactor[meas] = GALl1;
							else if (aux[1]=='5') epoch->measOrder[epoch->numGNSS-1].conversionFactor[meas] = GALl5a;
							else if (aux[1]=='7') epoch->measOrder[epoch->numGNSS-1].conversionFactor[meas] = GALl5b;
							else if (aux[1]=='8') epoch->measOrder[epoch->numGNSS-1].conversionFactor[meas] = GALl5;
							else if (aux[1]=='6') epoch->measOrder[epoch->numGNSS-1].conversionFactor[meas] = GALl6;
						} else if (epoch->measOrder[epoch->numGNSS-1].GNSS==GEO) {
							if (aux[1]=='1') epoch->measOrder[epoch->numGNSS-1].conversionFactor[meas] = SBASl1;
							else if(aux[1]=='5') epoch->measOrder[epoch->numGNSS-1].conversionFactor[meas] = SBASl5;
						}
					} else { // If measurement in not carrier-phase, conversionFactor is 1
						epoch->measOrder[epoch->numGNSS-1].conversionFactor[meas] = 1.0;
					}
					// Update dictionaries (if usableFreq is disabled for a specific frequency, the dictionary meas2Ind is avoided)
					if (options->usableFreq[epoch->numGNSS-1][getFrequencyInt(meas)]) {
					//if (epoch->measOrder[epoch->numGNSS-1].usableFreq[getFrequencyInt(epoch->measOrder[epoch->numGNSS-1].GNSS,meas)]) {
						epoch->measOrder[epoch->numGNSS-1].meas2Ind[meas] = obs;
					}
					epoch->measOrder[epoch->numGNSS-1].ind2Meas[obs] = meas;
				} else {
					sprintf(messagestr,"Reading RINEX 3.00 file: Unrecognized measurement type: '%s'",aux);
					printError(messagestr,options);
					error = 1;
				}				
				obs++;
				if ((j+1)%13==0) { // Following observables are in the next lines (max 13 per line)
					j=-1; // j will be increased in the for loop
					getL(line,&len,fd);
				}
			}
		} else if (strncmp(&line[60],"INTERVAL",8)==0) {
			getstr(aux,line,0,10);
			epoch->receiver.interval = atof(aux); 
		} else if (strncmp(&line[60],"APPROX POSITION XYZ",19)==0) {
			if (options->receiverPositionSource==rpRINEX) {
				getstr(aux,line,1,14);
				epoch->receiver.aproxPosition[0] = atof(aux);
				getstr(aux,line,14,14);
				epoch->receiver.aproxPosition[1] = atof(aux);
				getstr(aux,line,28,14);
				epoch->receiver.aproxPosition[2] = atof(aux);
				epoch->receiver.aproxPositionError = 1e4; // 10 Km
			}
		} else if (strncmp(&line[60],"END OF HEADER",13)==0) {
			readingHeader = 0;
			properlyRead = 1;
			break;
		}
	}

	return (properlyRead && !error); // properly read AND no error
}

/*****************************************************************************
 * Name        : readRinexObsEpoch
 * Description : Read a RINEX observation epoch from a file descriptor
 * Parameters  :
 * Name                           |Da|Unit|Description
 * FILE  *fd                       I  N/A  File descriptor
 * TEpoch  *epoch                  O  N/A  Structure to save the data
 * TConstellation  *constellation  I  N/A  TConstellation structure
 * enum ProcessingDirection        I  N/A  Direction of the read
 *                                         pFORWARD  => Forward
 *                                         pBACKWARD => Backward
 * Returned value (int)            O  N/A  Status of the function
 *                                         1 => Properly read
 *                                         0 => Error or EOF (no more epochs)
 *****************************************************************************/
int readRinexObsEpoch (FILE *fd, TEpoch *epoch, TConstellation *constellation, enum ProcessingDirection direction) {
	char 		line[MAX_INPUT_LINE];
	char    	aux[100];
	char		cons;
	int 		len = 0;
	int			properlyRead = 0;
	int			error = 0;
	int			i,j,k;
	int			n;
	struct tm	tm;
	double		seconds;
	double		value;
	int			nextRead;
	int			ret;

	// Rewind for backward processing
	if (direction==pBACKWARD) {
		ret = rewindEpochRinexObs(fd,epoch->source);
		if (ret==0) return 0;
		ret = rewindEpochRinexObs(fd,epoch->source);
		if (ret==0) return 0;
	}

	// Regular processing
	beforeReadEpochUpdate(epoch);

	if(getL(line,&len,fd)!=-1) {
		properlyRead = 1;
		if (epoch->source == RINEX2) {
			// RINEX 2.11				
			getstr(aux,line,27,2);
			epoch->flag = atoi(aux);
			
			if (epoch->flag==4) { // Unusable epoch
				getstr(aux,line,30,2);
				n = atoi(aux);
				for (i=0;i<=n;i++) {
					if (getL(line,&len,fd)==-1) {
						afterReadEpochUpdate(epoch, constellation);
						return 0;
					}
				}
			}
			
			getstr(aux,line,1,2);
			tm.tm_year = atoi(aux);
			if (tm.tm_year <= 70) tm.tm_year += 100;
			getstr(aux,line,4,2);
			tm.tm_mon  = atoi(aux)-1;
			getstr(aux,line,7,2);
			tm.tm_mday = atoi(aux);
			getstr(aux,line,10,2);
			tm.tm_hour = atoi(aux);
			getstr(aux,line,13,2);
			tm.tm_min  = atoi(aux);
			getstr(aux,line,16,10);
			tm.tm_sec  = atoi(aux);
			seconds = atof(aux);
			epoch->t.MJDN = MJDN(&tm);
			epoch->t.SoD = tm.tm_hour*3600 + tm.tm_min*60 + seconds;
				
			getstr(aux,line,30,2);
			epoch->numSatellites = atoi(aux);
			for (i=0,j=0;i<epoch->numSatellites;i++,j++) { // Get satellite PRNs
				if ((i%12)==0 && i!=0) {
					if (getL(line,&len,fd)==-1) error = 1; // The next satellitesare in the following line
					j = 0;
				}
				cons = line[32+3*j];
				getstr(aux,line,33+3*j,2);
				epoch->sat[i].PRN = atoi(aux);
				if (cons=='G' || cons==' ')
					epoch->sat[i].GNSS = GPS;
				else if (cons=='S')
					epoch->sat[i].GNSS = GEO;
				else if (cons=='R')
					epoch->sat[i].GNSS = GLONASS;
				else if (cons=='E')
					epoch->sat[i].GNSS = Galileo;
			}
			for (i=0;i<epoch->numSatellites;i++) { // Get satellite measurements
				if (getL(line,&len,fd)==-1) error = 1;
				nextRead = 0;
				for (j=0,k=0;j<epoch->measOrder[epoch->sat[i].GNSS].nDiffMeasurements;j++,k++) {
					if (nextRead) {
						if (getL(line,&len,fd)==-1) error = 1;
						nextRead = 0;
					}
					getstr(aux,line,16*k,14);
					value = atof(aux);
					if (value==0) value = -1;
					epoch->sat[i].meas[j].value = value;
					getstr(aux,line,14+16*k,1);
					epoch->sat[i].meas[j].LLI = atoi(aux);
					getstr(aux,line,15+16*k,1);
					epoch->sat[i].meas[j].SNR = atoi(aux);
					epoch->sat[i].meas[j].dataFlag = 0;
					if ((k+1)%5==0) { // Following observables are in the next lines (max 8 per line)
						nextRead = 1;
						k=-1; // j will be increased in the for loop
					}
 				}
			}
		} else if (epoch->source==RINEX3) {
			// RINEX 3.00
			getstr(aux,line,30,2);
			epoch->flag = atoi(aux);
			
			if (epoch->flag==4) { // Unusable epoch
				getstr(aux,line,33,2);
				n = atoi(aux);
				for (i=0;i<=n;i++) {
					if (getL(line,&len,fd)==-1) {
						afterReadEpochUpdate(epoch, constellation);
						return 0;
					}
				}
			}
			
			getstr(aux,line,2,4);
			tm.tm_year = atoi(aux)-1900;
			getstr(aux,line,7,2);
			tm.tm_mon  = atoi(aux)-1;
			getstr(aux,line,10,2);
			tm.tm_mday = atoi(aux);
			getstr(aux,line,13,2);
			tm.tm_hour = atoi(aux);
			getstr(aux,line,16,2);
			tm.tm_min  = atoi(aux);
			getstr(aux,line,19,10);
			tm.tm_sec  = atoi(aux);
			seconds = atof(aux);
			epoch->t.MJDN = MJDN(&tm);
			epoch->t.SoD = tm.tm_hour*3600 + tm.tm_min*60 + seconds;
			
			getstr(aux,line,33,2);
			epoch->numSatellites = atoi(aux);
			for (i=0;i<epoch->numSatellites;i++) { // Get satellite measurements and PRNs
				if (getL(line,&len,fd)==-1) error = 1;
				epoch->sat[i].GNSS = gnsschar2gnsstype(line[0]);
				getstr(aux,line,1,2);
				epoch->sat[i].PRN = atoi(aux);
				for (j=0;j<epoch->measOrder[epoch->sat[i].GNSS].nDiffMeasurements;j++) {
					getstr(aux,line,3+16*j,14);
					value = atof(aux);
					if (value==0) value = -1;
					epoch->sat[i].meas[j].value = value;
					getstr(aux,line,17+16*j,1);
					epoch->sat[i].meas[j].LLI = atoi(aux);
					getstr(aux,line,18+16*j,1);
					epoch->sat[i].meas[j].SNR = atoi(aux);
					epoch->sat[i].meas[j].dataFlag = 0;
				}
			}
		} // (epoch->source==RINEX3)
		afterReadEpochUpdate(epoch, constellation);
	}

	return (properlyRead && !error); // properly read AND no error
	
}

/*****************************************************************************
 * Name        : rewindEpochRinexObs
 * Description : Rewind one epoch backward in a RINEX observation file
 * Parameters  :
 * Name                           |Da|Unit|Description
 * FILE  *fd                       I  N/A  File descriptor
 * enum Source  src                I  N/A  Source of the RINEX file (RINEX2 or RINEX3)
 * Returned value (int)            O  N/A  Status of the function
 *                                         1 => Properly read
 *                                         0 => Error or EOF (no more epochs)
 *****************************************************************************/
int rewindEpochRinexObs (FILE *fd, enum Source src) {
	int			ret;
	int			end;
	int			yy=-1,mm=-1,dd=-1,hour=-1,min=-1;
	char 		line[MAX_INPUT_LINE];
	char    	aux[100];
	int 		len = 0;
	int			someLineRead = 0;
	
	end = 0;
	while (1) {
		ret = getLback(line,&len,fd);
		if (ret==-1) return 0;
		if (len>60 && strncmp(&line[60],"END OF HEADER",13)==0) {
			if (someLineRead) return 1;
			else return 0;
		} else {
			someLineRead = 1;
			// Check if date is within margins to detect init of epoch
			// RINEX 2.11
			if (src==RINEX2) {
				getstr(aux,line,0,4);
				yy = atoi(aux);
				getstr(aux,line,3,4);
				mm = atoi(aux);
				getstr(aux,line,6,4);
				dd = atoi(aux);
				getstr(aux,line,9,4);
				hour = atoi(aux);
				getstr(aux,line,12,4);
				min = atoi(aux);
			// RINEX 3.00
			} else if (src==RINEX3) {
				getstr(aux,line,1,6);
				yy = atoi(aux);
				if (yy>=2000) yy-=2000;
				else yy-=1900;
				getstr(aux,line,6,4);
				mm = atoi(aux);
				getstr(aux,line,9,4);
				dd = atoi(aux);
				getstr(aux,line,12,4);
				hour = atoi(aux);
				getstr(aux,line,15,4);
				min = atoi(aux);
			} else return 0;
			if (yy>=0 && yy<=99 && mm>=1 && mm<=12 && dd>=1 && dd<=31 && hour>=0 && hour<=23 && min>=0 &&min<=59) {
				return 1;
			}
		}
	}
}

/*****************************************************************************
 * Name        : beforeReadEpochUpdate
 * Description : This function should be called before reading a full epoch.
 *               It updates for all satellites:
 *                  - prevSat
 *                  - prevNumSatellites
 *                  - LiPrev
 *                  - tPrev
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  IO N/A  Structure to save the data
 *****************************************************************************/
void beforeReadEpochUpdate (TEpoch *epoch) {
	int		i,j;
	
	// Store previous sat[i] in prevSat[i] and numSatellites in prevNumSatellites
	if (epoch->numSatellites != 0) 
		memcpy(epoch->prevSat,epoch->sat,sizeof(TSatellite)*epoch->numSatellites);
	epoch->prevNumSatellites = epoch->numSatellites;
	
	// Store previous t and LI for cycle-slip detection
	for (i=0;i<MAX_CS_MEMORY-1;i++) {
		memcpy(&epoch->cycleslip.tPrev[i],&epoch->cycleslip.tPrev[i+1],sizeof(TTime));
		for (j=0;j<MAX_SATELLITES_IN_VIEW;j++) {
			epoch->cycleslip.LiPrev[j][i] = epoch->cycleslip.LiPrev[j][i+1];
		}
	}
	memcpy(&epoch->cycleslip.tPrev[MAX_CS_MEMORY-1],&epoch->t,sizeof(TTime));
	for (j=0;j<MAX_SATELLITES_IN_VIEW;j++) 
		epoch->cycleslip.LiPrev[j][MAX_CS_MEMORY-1] = getMeasurementValue(epoch,epoch->sat[j].GNSS,epoch->sat[j].PRN,LI);
}

/*****************************************************************************
 * Name        : afterReadEpochUpdate
 * Description : This function should be called after reading a full epoch.
 *               It updates several parameteres for all satellites:
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  IO N/A  Structure to save the data
 * TConstellation  *constellation  I  N/A  Satellite constellation information
 *****************************************************************************/
void afterReadEpochUpdate (TEpoch *epoch, TConstellation* constellation) {
	int						i;
	int						ind;
	double					LiPrevAux[MAX_SATELLITES_IN_VIEW][MAX_CS_MEMORY];
	double					nBwmeanAux[MAX_SATELLITES_IN_VIEW];
	double					nBwmean2Aux[MAX_SATELLITES_IN_VIEW];
	double					L1C1meanAux[MAX_SATELLITES_IN_VIEW];
	double					L1C1mean2Aux[MAX_SATELLITES_IN_VIEW];
	double					phasePrealignAux[MAX_SATELLITES_IN_VIEW][MAX_MEASUREMENTS_PER_SATELLITE];
	int						hasBeenCycleslipAux[MAX_SATELLITES_IN_VIEW];
	double					windUpRadAccumReceiverAux[MAX_SATELLITES_IN_VIEW];
	double					windUpRadAccumSatelliteAux[MAX_SATELLITES_IN_VIEW];
	int						arcLengthAux[MAX_SATELLITES_IN_VIEW];
	int						anyChange = 0;
	int						hasChanged[MAX_SATELLITES_IN_VIEW];
	int						arcLengthPrev[MAX_SATELLITES_IN_VIEW];
	double					windUpRadAccumReceiverPrev[MAX_SATELLITES_IN_VIEW];
	double					windUpRadAccumSatellitePrev[MAX_SATELLITES_IN_VIEW];
	double					smoothedMeasAux[MAX_SATELLITES_IN_VIEW][MAX_FILTER_MEASUREMENTS];
	TConstellationElement	*el;
	
	for (i=0;i<MAX_SATELLITES_IN_VIEW;i++) {
		hasChanged[i]=0;
		arcLengthPrev[i] = epoch->sat[i].arcLength;
		windUpRadAccumReceiverPrev[i] = epoch->sat[i].windUpRadAccumReceiver;
		windUpRadAccumSatellitePrev[i] = epoch->sat[i].windUpRadAccumSatellite;
		memcpy(smoothedMeasAux[i],epoch->sat[i].smoothedMeas,sizeof(double)*MAX_FILTER_MEASUREMENTS);
	}
	
	for (i=0;i<epoch->numSatellites;i++) {
		// Update SVN
		el = getConstellationElement(epoch->sat[i].GNSS,epoch->sat[i].PRN,&epoch->t,constellation);
		if (el!=NULL) epoch->sat[i].SVN = el->SVN;
		else epoch->sat[i].SVN = 0;
		// ind contains the index of the position of the satellite in the previous epoch
		ind = satWas(epoch->sat[i].GNSS,epoch->sat[i].PRN,epoch->prevSat,epoch->prevNumSatellites);
		if (ind!=-1) { // Satellite already in view
			if (i!=ind) { // Satellite is NOT in the same position as in the previous epoch
				anyChange = 1;
				hasChanged[i] = 1; // Change of satellite position
				memcpy(&LiPrevAux[i],&epoch->cycleslip.LiPrev[ind],sizeof(double)*MAX_CS_MEMORY);
				hasBeenCycleslipAux[i] = epoch->sat[ind].hasBeenCycleslip;
				nBwmeanAux[i] = epoch->cycleslip.nBwmean[ind];
				nBwmean2Aux[i] = epoch->cycleslip.nBwmean2[ind];
				L1C1meanAux[i] = epoch->cycleslip.L1C1mean[ind];
				L1C1mean2Aux[i] = epoch->cycleslip.L1C1mean2[ind];
				memcpy(&phasePrealignAux[i],&epoch->prevSat[ind].phasePrealign,sizeof(double)*MAX_MEASUREMENTS_PER_SATELLITE);
				arcLengthAux[i] = epoch->sat[ind].arcLength + 1;
				windUpRadAccumReceiverAux[i] =  epoch->sat[ind].windUpRadAccumReceiver;
				windUpRadAccumSatelliteAux[i] =  epoch->sat[ind].windUpRadAccumSatellite;
				memcpy(smoothedMeasAux[i],epoch->sat[ind].smoothedMeas,sizeof(double)*MAX_FILTER_MEASUREMENTS);
			} else { // Satellite is in the same position
				epoch->sat[i].arcLength += 1;
			}
		} else {
			anyChange = 1;
			hasChanged[i] = 2; // New satellite
		}
	}

	if (anyChange) {
		for (i=0;i<MAX_SATELLITES_IN_VIEW;i++) {
			if (hasChanged[i]==1) {
				memcpy(&epoch->cycleslip.LiPrev[i],&LiPrevAux[i],sizeof(double)*MAX_CS_MEMORY);
				epoch->cycleslip.nBwmean[i] = nBwmeanAux[i];
				epoch->cycleslip.nBwmean2[i] = nBwmean2Aux[i];
				epoch->cycleslip.L1C1mean[i] = L1C1meanAux[i];
				epoch->cycleslip.L1C1mean2[i] = L1C1mean2Aux[i];
				memcpy(&epoch->sat[i].phasePrealign,&phasePrealignAux[i],sizeof(double)*MAX_MEASUREMENTS_PER_SATELLITE);
				epoch->sat[i].hasBeenCycleslip = hasBeenCycleslipAux[i];
				epoch->sat[i].arcLength = arcLengthAux[i];
				epoch->sat[i].elevation = 0;
				epoch->sat[i].azimuth = 0;
				epoch->sat[i].windUpRadAccumReceiver = windUpRadAccumReceiverAux[i];
				epoch->sat[i].windUpRadAccumSatellite = windUpRadAccumSatelliteAux[i];
				memcpy(epoch->sat[i].smoothedMeas,smoothedMeasAux[i],sizeof(double)*MAX_FILTER_MEASUREMENTS);
			} else if (hasChanged[i]==2) {
				// New satellite
				initSatellite(epoch,i);
			}
		}
	}
}

/*****************************************************************************
 * Name        : readRinexNav
 * Description : Read a RINEX navigation file from a file descriptor
 * Parameters  :
 * Name                           |Da|Unit|Description
 * FILE  *fd                       I  N/A  File descriptor
 * TGNSSproducts  *products        O  N/A  Struct to save the data
 * Returned value (int)            O  N/A  Status of the function
 *                                         1 => Properly read
 *                                         0 => Error
 *****************************************************************************/
int readRinexNav (FILE *fd, TGNSSproducts *products) {
	char 		line[MAX_INPUT_LINE];
	char    	aux[100],a0[100],a1[100],a2[100],a3[100];
	int 		len = 0;
	double		rinexVersion;
	int			properlyRead = 0;
	int			error = 0;
	int			readingHeader = 1;
	int			i,j;
	TBRDCblock	block;
	struct tm	tm;
	double		seconds;
	int			row;
	int			indSat,indBlock;
	
	products->type = BRDC;
	products->BRDC = malloc(sizeof(TBRDCproducts));
	initBRDCproducts(products->BRDC);
	
	while(getL(line,&len,fd)!=-1) {
		if (readingHeader) {
			if (strncmp(&line[60],"RINEX VERSION / TYPE",20)==0) {
				getstr(aux,line,0,9);
				rinexVersion = atof(aux);
			} else if (strncmp(&line[60],"ION ALPHA",9)==0) {
				for (i=0;i<4;i++) {
					getstr(aux,line,2+12*i,12);
					for (j=0;j<strlen(aux);j++) {
						if (aux[j]=='D') aux[j]='E';
					}
					products->BRDC->ionA[i] = atof(aux);
				}
			} else if (strncmp(&line[60],"ION BETA",8)==0) {
				for (i=0;i<4;i++) {
					getstr(aux,line,2+12*i,12);
					for (j=0;j<strlen(aux);j++) {
						if (aux[j]=='D') aux[j]='E';
					}
					products->BRDC->ionB[i] = atof(aux);
				}
			} else if (strncmp(&line[60],"END OF HEADER",13)==0) {
				readingHeader = 0;
				row = 0;
				properlyRead = 1;
			}
		} else {  // !readingHeader
			if ((int)(rinexVersion)==2) {
				if (row==0) {
					getstr(aux,line,0,2);
					block.PRN = atoi(aux);
					getstr(aux,line,3,2);
					tm.tm_year = atoi(aux);
					if (tm.tm_year <= 70) tm.tm_year += 100;
					getstr(aux,line,6,2);
					tm.tm_mon  = atoi(aux)-1;
					getstr(aux,line,9,2);
					tm.tm_mday = atoi(aux);
					getstr(aux,line,12,2);
					tm.tm_hour = atoi(aux);
					getstr(aux,line,15,2);
					tm.tm_min  = atoi(aux);
					getstr(aux,line,18,4);
					tm.tm_sec  = atoi(aux);
					seconds = atof(aux);
					block.Ttoc.MJDN = MJDN(&tm);
					block.Ttoc.SoD = tm.tm_hour*3600 + tm.tm_min*60 + seconds;
				}
				getstr(a0,line,3,19);
				if(a0[15]=='D') a0[15]='E';
				if(a0[14]=='D') a0[14]='E'; // '-' sign can appear or not
				getstr(a1,line,22,19);
				if(a1[15]=='D') a1[15]='E';
				if(a1[14]=='D') a1[14]='E'; // '-' sign can appear or not
				getstr(a2,line,41,19);
				if(a2[15]=='D') a2[15]='E';
				if(a2[14]=='D') a2[14]='E'; // '-' sign can appear or not
				getstr(a3,line,60,19);
				if(a3[15]=='D') a3[15]='E';
				if(a3[14]=='D') a3[14]='E'; // '-' sign can appear or not
				
				switch (row) {
					case 0:		
						block.clockbias 		= atof(a1);
						block.clockdrift 		= atof(a2);
						block.clockdriftrate 	= atof(a3);
						break;
					case 1:
						block.IODE 				= (int)atof(a0);
						block.crs				= atof(a1);
						block.deltan			= atof(a2);
						block.M0				= atof(a3);
						break;
					case 2:
						block.cuc				= atof(a0);
						block.e					= atof(a1);
						block.cus				= atof(a2);
						block.sqrta				= atof(a3);
						break;
					case 3:
						block.toe				= atof(a0);
						block.cic				= atof(a1);
						block.OMEGA				= atof(a2);
						block.cis				= atof(a3);
						break;
					case 4:
						block.i0				= atof(a0);
						block.crc				= atof(a1);
						block.omega				= atof(a2);
						block.OMEGADOT			= atof(a3);
						break;
					case 5:
						block.IDOT				= atof(a0);
						block.codesOnL2			= (int)atof(a1);
						block.GPSweek			= (int)atof(a2);
						block.L2Pdataflag		= (int)atof(a3);
						break;
					case 6:
						block.SVaccuracy		= atof(a0);
						block.SVhealth			= (int)atof(a1);
						block.TGD				= atof(a2);
						block.IODC				= (int)atof(a3);
						break;
					case 7:
						block.transTime			= atof(a0);
						block.fitInterval		= (int)atof(a1);
				}
				row++;
				if (row==8) {
					row = 0;
					block.Ttoe = gpsws2ttime(block.GPSweek,block.toe);
					block.TtransTime = gpsws2ttime(block.GPSweek,block.transTime);
					indSat = products->BRDC->index[GPS][block.PRN];
					if (indSat == -1) {
						products->BRDC->index[GPS][block.PRN] = products->BRDC->numsats;
						indSat = products->BRDC->index[GPS][block.PRN];
						products->BRDC->block = realloc(products->BRDC->block,sizeof(TBRDCblock*)*(products->BRDC->numsats+1));
						products->BRDC->numblocks = realloc(products->BRDC->numblocks,sizeof(int)*(products->BRDC->numsats+1));
						products->BRDC->block[indSat] = NULL;
						products->BRDC->numblocks[indSat] = 0;
						products->BRDC->numsats++;
					}
					indBlock = products->BRDC->numblocks[indSat];
					products->BRDC->block[indSat] = realloc(products->BRDC->block[indSat],sizeof(TBRDCblock)*(indBlock+1));
					memcpy(&products->BRDC->block[indSat][indBlock], &block,sizeof(TBRDCblock));
					products->BRDC->numblocks[indSat]++;
				}
			}
		}
	}

	return (properlyRead && !error); // properly read AND no error
}

/*****************************************************************************
 * Name        : readRinexNavFile
 * Description : Read a RINEX navigation file from a filename
 * Parameters  :
 * Name                           |Da|Unit|Description
 * char  *filename                 I  N/A  Name of the file containing the RINEX
 * TGNSSproducts  *products        O  N/A  Struct to save the data
 * Returned value (int)            O  N/A  Status of the function
 *                                         1  => File open and properly read
 *                                         0  => File open and failed to read
 *                                         -1 => File not open
 *****************************************************************************/
int readRinexNavFile (char *filename, TGNSSproducts *products) {
	FILE	*fd;
	int		ret;
	
	if (whatFileTypeIs(filename)!=ftRINEXbroadcast) return 0;
	
	if ((fd = fopen(filename,"r"))) {
		ret = readRinexNav (fd, products);
		fclose (fd);
		return ret;
	} else {
		// File do not exist
		return -1;
	}
}

/*****************************************************************************
 * Name        : readSP3
 * Description : Read an SP3 precise products file from a file descriptor
 * Parameters  :
 * Name                           |Da|Unit|Description
 * FILE  *fd                       I  N/A  File descriptor
 * TGNSSproducts  *products        O  N/A  Struct to save the data
 * int  readOrbits                 I  N/A  Flag to read the position and 
 *                                         velocities (0 or 1)
 * int  readClocks                         Flag to read the clocks (0 or 1)
 * Returned value (int)            O  N/A  Status of the function
 *                                         1 => Properly read
 *                                         0 => Error
 *****************************************************************************/
int readSP3 (FILE *fd, TGNSSproducts *products, int readOrbits, int readClocks) {
	char 			line[MAX_INPUT_LINE];
	char    		aux[100];
	int 			len = 0;
	int				properlyRead = 0;
	int				error = 0;
	int				readingHeader = 1;
	int				i,j;
	int				PRN;
	TTime			t;
	int				ind;
	struct tm		tm;
	double			seconds;
	int				readedLine = 0;
	int				readedSatellites;
	char			versionId;
	enum GNSSystem	system;
	int				usedRecords;
	int				auxN;

	if (!readOrbits && !readClocks) return 0;

	products->type = SP3;
	products->SP3 = malloc(sizeof(TSP3products));
	initSP3products(products->SP3);

	while(getL(line,&len,fd)!=-1) {
		readedLine++;

		if (readingHeader) {
			switch(readedLine) {
				case 1:
					// #bP2008  8 27 00 00 00.00000000      96   d   ITR91 FIT  JPL
					if (line[0]!='#') error = 1;
					versionId = line[1];
					if (line[2]=='V') products->SP3->orbits.velocityAvailable = 1;
					else products->SP3->orbits.velocityAvailable = 0;
					getstr(aux,line,32,7);
					if (readOrbits) products->SP3->orbits.numRecords = atoi(aux);
					if (readClocks) products->SP3->clocks.numRecords = atoi(aux);
					break;
				case 2:
					// ## 1494 259200.00000000   900.00000000 54705 0.0000000000000
					getstr(aux,line,24,14);
					if (readOrbits) products->SP3->orbits.interval = atoi(aux);
					if (readClocks) products->SP3->clocks.interval = atoi(aux);
					getstr(aux,line,39,5);
					if (readOrbits) products->SP3->orbits.startTime.MJDN = atoi(aux);
					if (readClocks) products->SP3->clocks.startTime.MJDN = atoi(aux);
					getstr(aux,line,45,15);
					if (readOrbits) products->SP3->orbits.startTime.SoD = atof(aux)*86400.0;
					if (readClocks) products->SP3->clocks.startTime.SoD = atof(aux)*86400.0;
					break;
				case 3:
					// +   31   G02G03G04G05G06G07G08G09G10G11G12G13G14G15G16G17G18
					getstr(aux,line,4,2);
					if (readOrbits) products->SP3->orbits.numSatellites = atoi(aux);
					if (readClocks) products->SP3->clocks.numSatellites = atoi(aux);
					// Reserve memory for satellites
					if (readOrbits) {
						products->SP3->orbits.block = malloc(sizeof(TSP3orbitblock*)*products->SP3->orbits.numSatellites);
						products->SP3->orbits.accuracy = malloc(sizeof(double)*products->SP3->orbits.numSatellites);
					}
					if (readClocks) products->SP3->clocks.block = malloc(sizeof(TSP3clockblock*)*products->SP3->clocks.numSatellites);
					if (readOrbits) auxN = products->SP3->orbits.numSatellites;
					else auxN = products->SP3->clocks.numSatellites;
					for (i=0;i<auxN;i++) {
						if (readOrbits) {
							products->SP3->orbits.block[i] = malloc(sizeof(TSP3orbitblock)*products->SP3->orbits.numRecords);
							for (j=0;j<products->SP3->orbits.numRecords;j++)
								initSP3orbitblock(&products->SP3->orbits.block[i][j]);
						}
						if (readClocks) {
							products->SP3->clocks.block[i] = malloc(sizeof(TSP3clockblock)*products->SP3->clocks.numRecords);
							for (j=0;j<products->SP3->clocks.numRecords;j++) 
								initSP3clockblock(&products->SP3->clocks.block[i][j]);
						}
					}
					readedSatellites = 0;
				case 4:
					// +        G19G20G21G22G23G24G25G26G27G28G29G30G31G32  0  0  0
				case 5:
					// +          0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
				case 6:
					// +          0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
				case 7:
					// +          0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
					for (i=0;i<17 && readedSatellites<auxN;i++,readedSatellites++) { //auxN = products->SP3->XXXX.numSatellites;
						system = gnsschar2gnsstype(line[9+3*i]);
						getstr(aux,line,10+3*i,2);
						if (readOrbits) products->SP3->orbits.index[system][atoi(aux)] = readedSatellites;
						if (readClocks) products->SP3->clocks.index[system][atoi(aux)] = readedSatellites;
					}
					break;
				case 8:
					// ++         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
					readedSatellites = 0;
				case 9:
					// ++         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
				case 10:
					// ++         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
				case 11:
					// ++         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
				case 12:
					// ++         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
					for (i=0;i<17 && readedSatellites<auxN;i++,readedSatellites++) { //auxN = products->SP3->XXXX.numSatellites;
						// Accuracy is 2^val mm.
						getstr(aux,line,10+3*i,2);
						if (readOrbits) products->SP3->orbits.accuracy[readedSatellites] = pow(2,atoi(aux))*1e-3;
					}
					break;
				case 22:
					// /* CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
					properlyRead = 1;
					readingHeader = 0;
					usedRecords = 0;
					break;
				default:
					break;
			}
		} else { // !readingHeader
			if (line[0]=='*') { // Epoch starting
				// *  2008 08 27 00 00 00.00000000
				getstr(aux,line,3,4);
				tm.tm_year = atoi(aux)-1900;
				getstr(aux,line,8,2);
				tm.tm_mon  = atoi(aux)-1;
				getstr(aux,line,11,2);
				tm.tm_mday = atoi(aux);
				getstr(aux,line,14,2);
				tm.tm_hour = atoi(aux);
				getstr(aux,line,17,2);
				tm.tm_min  = atoi(aux);
				getstr(aux,line,20,11);
				tm.tm_sec  = atoi(aux);
				seconds = atof(aux);
				t.MJDN = MJDN(&tm);
				t.SoD = tm.tm_hour*3600 + tm.tm_min*60 + seconds;
				for (i=0;i<products->SP3->orbits.numSatellites;i++) {
					memcpy(&products->SP3->orbits.block[i][usedRecords].t,&t,sizeof(TTime));
				}
				usedRecords++;
			} else if (line[0]=='P') { // Position record
				system = gnsschar2gnsstype(line[1]);
				getstr(aux,line,2,2);
				PRN = atoi(aux);
				if (readOrbits) ind = products->SP3->orbits.index[system][PRN];
				else ind = products->SP3->clocks.index[system][PRN];
				if (readOrbits) {
					getstr(aux,line,4,14);
					products->SP3->orbits.block[ind][usedRecords-1].x[0] = atof(aux)*1e3;
					getstr(aux,line,18,14);
					products->SP3->orbits.block[ind][usedRecords-1].x[1] = atof(aux)*1e3;
					getstr(aux,line,32,14);
					products->SP3->orbits.block[ind][usedRecords-1].x[2] = atof(aux)*1e3;
					memcpy(&products->SP3->orbits.block[ind][usedRecords-1].t,&t,sizeof(TTime));
				}
				if (readClocks) {
					getstr(aux,line,46,14);
					if (atof(aux)==INVALID_CLOCK) {
						products->SP3->clocks.block[ind][usedRecords-1].clock = INVALID_CLOCK;
					} else {
						products->SP3->clocks.block[ind][usedRecords-1].clock = atof(aux)*(c/1e6);
					}
					memcpy(&products->SP3->clocks.block[ind][usedRecords-1].t,&t,sizeof(TTime));
				}
				
			} else if (line[0]=='V') { // Velocity record
				system = gnsschar2gnsstype(line[1]);
				getstr(aux,line,2,2);
				PRN = atoi(aux);
				if (readOrbits) ind = products->SP3->orbits.index[system][PRN];
				else ind = products->SP3->clocks.index[system][PRN];
				if (readOrbits) {
					getstr(aux,line,4,14);
					products->SP3->orbits.block[ind][usedRecords-1].v[0] = atof(aux)*1e-1;
					getstr(aux,line,18,14);
					products->SP3->orbits.block[ind][usedRecords-1].v[1] = atof(aux)*1e-1;
					getstr(aux,line,32,14);
					products->SP3->orbits.block[ind][usedRecords-1].v[2] = atof(aux)*1e-1;
					memcpy(&products->SP3->orbits.block[ind][usedRecords-1].t,&t,sizeof(TTime));
				}
				if (readClocks) {
					getstr(aux,line,46,14);
					products->SP3->clocks.block[ind][usedRecords-1].clockrate = atof(aux)*(c/1e7);
					memcpy(&products->SP3->clocks.block[ind][usedRecords-1].t,&t,sizeof(TTime));
				}
			} else if (strncmp(line,"EOF",3)==0) { // End of File
				break;
			} else { // Record unidentified
				error = 1;
			}
		}
	}
	
	return (properlyRead && !error); // properly read AND no error
}

/*****************************************************************************
 * Name        : readSP3File
 * Description : Read an SP3 precise products file from a filename
 * Parameters  :
 * Name                           |Da|Unit|Description
 * char  *filename                 I  N/A  Name of the file containing the SP3
 * TGNSSproducts  *products        O  N/A  Struct to save the data
 * int  readOrbits                 I  N/A  Flag to read the position and 
 *                                         velocities (0 or 1)
 * int  readClocks                         Flag to read the clocks (0 or 1)
 * Returned value (int)            O  N/A  Status of the function
 *                                         1  => File open and properly read
 *                                         0  => File open and failed to read
 *                                         -1 => File not open
 *****************************************************************************/
int readSP3File (char *filename, TGNSSproducts *products, int readOrbits, int readClocks) {
	FILE	*fd;
	int		ret;
	
	if (whatFileTypeIs(filename)!=ftSP3) return 0;
	
	if ((fd = fopen(filename,"r"))) {
		ret = readSP3 (fd, products, readOrbits, readClocks);
		fclose(fd);
		return ret;
	} else {
		// File do not exist
		return -1;
	}
}

/*****************************************************************************
 * Name        : readRinexClocks
 * Description : Read a RINEX Clocks file from a file descriptor
 * Parameters  :
 * Name                           |Da|Unit|Description
 * FILE  *fd                       I  N/A  File descriptor
 * TGNSSproducts  *products        O  N/A  Struct to save the data
 * Returned value (int)            O  N/A  Status of the function
 *                                         1 => Properly read
 *                                         0 => Error
 *****************************************************************************/
int readRinexClocks (FILE *fd, TGNSSproducts *products) {
	char 		line[MAX_INPUT_LINE];
	char    	aux[100];
	int 		len = 0;
	double		rinexVersion;
	int			properlyRead = 0;
	int			error = 0;
	int			readingHeader = 1;
	enum GNSSystem	system;
	int			PRN;
	struct tm	tm;
	int			i,j;
	int			ind;
	double		seconds;
	TTime		t, lastT;
	int			readedSatellites;
	const int	ini_vsize = 64;
	int 		found_prn_list = 0;
	
	lastT.MJDN = 0;
	lastT.SoD = 0;
	
	while(getL(line,&len,fd)!=-1) {
		if (readingHeader) {
			if (strncmp(&line[60],"RINEX VERSION / TYPE",20)==0) {
				getstr(aux,line,0,9);
				rinexVersion = atof(aux);
			} else if (strncmp(&line[60],"# OF SOLN SATS",14)==0) {
				getstr(aux,line,4,2);
				products->SP3->clocks.numSatellites = atoi(aux);
				// Reserve memory for satellites
				products->SP3->clocks.block = malloc(sizeof(TSP3clockblock*)*products->SP3->clocks.numSatellites);
				// RINEX clock standard does not provide any apriori information on the number of records that will 
				// appear on the file. A dynamic readjustment becomes necessary to overcome this issue.
				products->SP3->clocks.vsizeblock = ini_vsize;
				// Reserve initial memory for blocks
				for (i=0;i<products->SP3->clocks.numSatellites;i++) {
					products->SP3->clocks.block[i] = malloc(sizeof(TSP3clockblock)*products->SP3->clocks.vsizeblock);
					for (j=0;j<products->SP3->clocks.vsizeblock;j++) 
						initSP3clockblock(&products->SP3->clocks.block[i][j]);
				}
				readedSatellites = 0;
			} else if (strncmp(&line[60],"PRN LIST",8)==0) {
				found_prn_list = 1;
				for (i=0;i<15 && readedSatellites<products->SP3->clocks.numSatellites;i++,readedSatellites++) {
					system = gnsschar2gnsstype(line[4*i]);
					getstr(aux,line,1+4*i,2);
					products->SP3->clocks.index[system][atoi(aux)] = readedSatellites;
				}
			} else if (strncmp(&line[60],"END OF HEADER",13)==0) {
				readingHeader = 0;
				properlyRead = 1;
				products->SP3->clocks.numRecords = 0;
				// The standard forces that the field 'PRN LIST' is present in the header, so the next section would
				// never be executed...
				if (!found_prn_list) {
					products->SP3->clocks.numSatellites = MAX_SATELLITES_PER_GNSS;
					// Reserve memory for satellites
					products->SP3->clocks.block = malloc(sizeof(TSP3clockblock*)*products->SP3->clocks.numSatellites);
					products->SP3->clocks.vsizeblock = ini_vsize;
					for (i=0;i<products->SP3->clocks.numSatellites;i++) {
						products->SP3->clocks.block[i] = malloc(sizeof(TSP3clockblock)*products->SP3->clocks.vsizeblock);
						for (j=0;j<products->SP3->clocks.vsizeblock;j++) 
							initSP3clockblock(&products->SP3->clocks.block[i][j]);
					}
					readedSatellites = 0;
					for (j=0;j<MAX_GNSS;j++) {
						for (i=0;i<MAX_SATELLITES_PER_GNSS;i++) {
							products->SP3->clocks.index[j][i] = readedSatellites;
						}
					}
				}
			}
		} else { // !readingHeader
			getstr(aux,line,3,4);
			if (aux[0]=='G' && atoi(&aux[1])!=0) {
				// AS G02  2008 10 10 00 00  0.000000  2    1.813604156067e-04  3.113411106060e-11
				system = gnsschar2gnsstype(aux[0]);
				getstr(aux,line,4,2);
				PRN = atoi(aux);
				ind = products->SP3->clocks.index[system][PRN];

				getstr(aux,line,8,4);
				tm.tm_year = atoi(aux)-1900;
				getstr(aux,line,13,2);
				tm.tm_mon  = atoi(aux)-1;
				getstr(aux,line,16,2);
				tm.tm_mday = atoi(aux);
				getstr(aux,line,19,2);
				tm.tm_hour = atoi(aux);
				getstr(aux,line,22,2);
				tm.tm_min  = atoi(aux);
				getstr(aux,line,25,9);
				tm.tm_sec  = atoi(aux);
				seconds = atof(aux);
				t.MJDN = MJDN(&tm);
				t.SoD = tm.tm_hour*3600 + tm.tm_min*60 + seconds;
				
				if (lastT.MJDN != t.MJDN || lastT.SoD != t.SoD) { // New Epoch
					if (products->SP3->clocks.interval == 0 && lastT.MJDN!=0) { // If interval is still not known (and it is not the first epoch)
						products->SP3->clocks.interval = (int)tdiff(&t,&lastT);
					}
					memcpy(&lastT,&t,sizeof(TTime));
					products->SP3->clocks.numRecords++;
					if (products->SP3->clocks.numRecords == products->SP3->clocks.vsizeblock) {
						products->SP3->clocks.vsizeblock *= 2;
						// Reserve memory for blocks
						for (i=0;i<products->SP3->clocks.numSatellites;i++) {
							products->SP3->clocks.block[i] = realloc(products->SP3->clocks.block[i],sizeof(TSP3clockblock)*products->SP3->clocks.vsizeblock);
							for (j=products->SP3->clocks.vsizeblock/2;j<products->SP3->clocks.vsizeblock;j++) 
								initSP3clockblock(&products->SP3->clocks.block[i][j]);
						}
					}
				}
				
				getstr(aux,line,37,22);
				products->SP3->clocks.block[ind][products->SP3->clocks.numRecords-1].clock = atof(aux)*c;
				products->SP3->clocks.block[ind][products->SP3->clocks.numRecords-1].clockrate = 0;
				products->SP3->clocks.block[ind][products->SP3->clocks.numRecords-1].clockdriftrate = 0;
				
				getstr(aux,line,59,22);
				
				memcpy(&products->SP3->clocks.block[ind][products->SP3->clocks.numRecords-1].t,&t,sizeof(TTime));
			}
		}
	}

	return (properlyRead && !error); // properly read AND no error
}

/*****************************************************************************
 * Name        : readRinexClocksFile
 * Description : Read a RINEX Clocks file from a filename
 * Parameters  :
 * Name                           |Da|Unit|Description
 * char  *filename                 I  N/A  Name of the file
 * TGNSSproducts  *products        O  N/A  Struct to save the data
 * Returned value (int)            O  N/A  Status of the function
 *                                         1  => File open and properly read
 *                                         0  => File open and failed to read
 *                                         -1 => File not open
 *****************************************************************************/
int readRinexClocksFile (char *filename, TGNSSproducts *products) {
	FILE	*fd;
	int		ret;
	
	if (whatFileTypeIs(filename)!=ftRINEXclocks) return 0;
	
	if ((fd = fopen(filename,"r"))) {
		ret = readRinexClocks (fd, products);
		fclose(fd);
		return ret;
	} else {
		// File do not exist
		return -1;
	}
}

/*****************************************************************************
 * Name        : readIONEX
 * Description : Read a IONEX file from a file descriptor
 * Parameters  :
 * Name                           |Da|Unit|Description
 * FILE  *fd                       I  N/A  File descriptoy
 * TIONEX  *IONEX                  O  N/A  Struct to save the data
 * Returned value (int)            O  N/A  Status of the function
 *                                         1 => Properly read
 *                                         0 => Error
 *****************************************************************************/
int readIONEX (FILE *fd, TIONEX *IONEX) {
	char 		line[MAX_INPUT_LINE];
	char    	aux[100];
	int 		len = 0;
	double		ionexVersion;
	int			properlyRead = 0;
	int			error = 0;
	int			i;
	int			PRN;
	struct tm	tm;
	double		seconds;
	int			readingCells = 0;
	int			mapType; // 0->TEC 1->RMS
	TIonoMap	ionoMap;
	int			ind;
	double		lat;
	double		lon1,lon2,dlon,lon;
	int			nlon,ilon;
	double		h;
	
	initIonoMap(&ionoMap);
	while(getL(line,&len,fd)!=-1) {
		if (!readingCells) {
			if (strncmp(&line[60],"IONEX VERSION / TYPE",20)==0) {
				getstr(aux,line,0,9);
				ionexVersion = atof(aux);
			} else if (strncmp(&line[60],"EPOCH OF FIRST MAP",18)==0) {
				getstr(aux,line,2,4);
				tm.tm_year = atoi(aux)-1900;
				getstr(aux,line,10,2);
				tm.tm_mon  = atoi(aux)-1;
				getstr(aux,line,16,2);
				tm.tm_mday = atoi(aux);
				getstr(aux,line,22,2);
				tm.tm_hour = atoi(aux);
				getstr(aux,line,28,2);
				tm.tm_min  = atoi(aux);
				getstr(aux,line,34,11);
				tm.tm_sec  = atoi(aux);
				seconds = atof(aux);
				IONEX->startTime.MJDN = MJDN(&tm);
				IONEX->startTime.SoD = tm.tm_hour*3600 + tm.tm_min*60 + seconds;
			} else if (strncmp(&line[60],"INTERVAL",8)==0) {
				getstr(aux,line,0,6);
				IONEX->interval = atoi(aux);
			} else if (strncmp(&line[60],"# OF MAPS IN FILE",17)==0) {
				getstr(aux,line,0,6);
				IONEX->numMaps = atoi(aux);
				// Reserve memory for maps
				IONEX->map = malloc(sizeof(TIonoMap)*IONEX->numMaps);
			} else if (strncmp(&line[60],"EXPONENT",8)==0) {
				getstr(aux,line,0,6);
				IONEX->exponent = atoi(aux);
				IONEX->conversionFactor = E(IONEX->exponent);
			} else if (strncmp(&line[60],"HGT1 / HGT2 / DHGT",18)==0) {
				getstr(aux,line,2,6);			// HGT1
				ionoMap.hgt1 = atof(aux);
				getstr(aux,line,8,6);			// HGT2
				ionoMap.hgt2 = atof(aux);
				getstr(aux,line,14,6);			// DHGT
				ionoMap.dhgt = atof(aux);
				ionoMap.nhgt = numSteps(ionoMap.hgt1,ionoMap.hgt2,ionoMap.dhgt);
			} else if (strncmp(&line[60],"LAT1 / LAT2 / DLAT",18)==0) {
				getstr(aux,line,2,6);			// LAT1
				ionoMap.lat1 = atof(aux);
				getstr(aux,line,8,6);			// LAT2
				ionoMap.lat2 = atof(aux);
				getstr(aux,line,14,6);			// DLAT
				ionoMap.dlat = atof(aux);
				ionoMap.nlat = numSteps(ionoMap.lat1,ionoMap.lat2,ionoMap.dlat);
			} else if (strncmp(&line[60],"LON1 / LON2 / DLON",18)==0) {
				getstr(aux,line,2,6);			// LON1
				ionoMap.lon1 = atof(aux);
				getstr(aux,line,8,6);			// LON2
				ionoMap.lon2 = atof(aux);
				getstr(aux,line,14,6);			// DLON
				ionoMap.dlon = atof(aux);
				ionoMap.nlon = numSteps(ionoMap.lon1,ionoMap.lon2,ionoMap.dlon);
			} else if (strncmp(&line[60],"START OF TEC MAP",16)==0) {
				getstr(aux,line,0,6);
				ind = atoi(aux);
				mapType = 0;  // 0->TEC
			} else if (strncmp(&line[60],"START OF RMS MAP",16)==0) {
				getstr(aux,line,0,6);
				ind = atoi(aux);
				mapType = 1;  // 1->RMS
			} else if (strncmp(&line[60],"END OF TEC MAP",14)==0) {

			} else if (strncmp(&line[60],"END OF RMS MAP",14)==0) {
		
			} else if (strncmp(&line[60],"END OF HEADER",13)==0) {
				for (i=0;i<IONEX->numMaps;i++) {
					memcpy(&IONEX->map[i],&ionoMap,sizeof(TIonoMap));
					IONEX->map[i].TEC = malloc(sizeof(double)*getIonoMapTotalSize(&IONEX->map[i]));
					IONEX->map[i].RMS = malloc(sizeof(double)*getIonoMapTotalSize(&IONEX->map[i]));					
				}
			} else if (strncmp(&line[60],"EPOCH OF CURRENT MAP",20)==0) {
				getstr(aux,line,2,4);
				tm.tm_year = atoi(aux)-1900;
				getstr(aux,line,10,2);
				tm.tm_mon  = atoi(aux)-1;
				getstr(aux,line,16,2);
				tm.tm_mday = atoi(aux);
				getstr(aux,line,22,2);
				tm.tm_hour = atoi(aux);
				getstr(aux,line,28,2);
				tm.tm_min  = atoi(aux);
				getstr(aux,line,34,11);
				tm.tm_sec  = atoi(aux);
				seconds = atof(aux);
				IONEX->map[ind].t.MJDN = MJDN(&tm);
				IONEX->map[ind].t.SoD = tm.tm_hour*3600 + tm.tm_min*60 + seconds;
			} else if (strncmp(&line[60],"LAT/LON1/LON2/DLON/H",20)==0) {
				getstr(aux,line,2,6);
				lat = atof(aux);
				getstr(aux,line,8,6);
				lon1 = atof(aux);
				getstr(aux,line,14,6);
				lon2 = atof(aux);
				getstr(aux,line,20,6);
				dlon = atof(aux);
				getstr(aux,line,26,6);
				h = atof(aux);
				nlon = numSteps(lon1,lon2,dlon);
				ilon = 0;
				readingCells = 1;
			} else if (strncmp(&line[60],"PRN / BIAS / RMS",16)==0) {
				getstr(aux,line,4,2);
				PRN = atoi(aux);
				getstr(aux,line,7,9);
				IONEX->DCB[GPS][PRN] = atof(aux);
				getstr(aux,line,17,9);
				IONEX->DCBRMS[GPS][PRN] = atof(aux);
			}
		} else {  // readingCells
			for (i=0;i<16 && ilon<nlon;i++,ilon++) {
				getstr(aux,line,i*5,5);
				lon = ilon*dlon + lon1;
				setIonoValue(&IONEX->map[ind],mapType,h,lat,lon,atof(aux)*IONEX->conversionFactor);
			}
			if (ilon==nlon) readingCells = 0;
		}
		
	}

	return (properlyRead && !error); // properly read AND no error
}

/*****************************************************************************
 * Name        : readIONEX
 * Description : Read a IONEX file from a filename
 * Parameters  :
 * Name                           |Da|Unit|Description
 * char  *filename                 I  N/A  Name of the file
 * TIONEX  *IONEX                  O  N/A  Struct to save the data
 * Returned value (int)            O  N/A  Status of the function
 *                                         1  => File open and properly read
 *                                         0  => File open and failed to read
 *                                         -1 => File not open
 *****************************************************************************/
int readIONEXFile (char *filename, TIONEX *IONEX) {
	FILE	*fd;
	int		ret;
	
	if (whatFileTypeIs(filename)!=ftIONEX) return 0;
	
	if ((fd = fopen(filename,"r"))) {
		ret = readIONEX (fd, IONEX);
		fclose(fd);
		return ret;
	} else {
		// File do not exist
		return -1;
	}
}

/*****************************************************************************
 * Name        : readConstellation
 * Description : Read a constellation status file from a file descriptor
 *               File available at: 
 *                 http://gge.unb.ca/Resources/GPSConstellationStatus.txt
 *	             File compiled and maintaned by Richard B. Langley, Dept. of 
 *               Geodesy and Geomatics Engineering, University of New Brunswick.
 * Parameters  :
 * Name                           |Da|Unit|Description
 * FILE  *fd                       I  N/A  File descriptor
 * TConstellation  *constellation  O  N/A  Struct to save the data
 * Returned value (int)            O  N/A  Status of the function
 *                                         1 => Properly read
 *                                         0 => Error
 *****************************************************************************/
int readConstellation (FILE *fd, TConstellation *constellation) {
	char 					line[MAX_INPUT_LINE];
	char    				aux[100];
	int 					len = 0;
	int						properlyRead = 0;
	int						error = 0;
	struct tm				tm;
	int						readingInfo = 0;
	TConstellationElement	sat;
	int						i;
	
	tm.tm_hour = 0;
	tm.tm_min = 0;
	tm.tm_sec = 0;

	while(getL(line,&len,fd)!=-1) {
		if (line[strlen(line)-1]==13) line[strlen(line)-1] = '\n';
		getstr(aux,line,0,5);
		if (strcmp(aux,"Block")==0) {
			// Initialize
			initConstellationElement(&sat);
			sat.GNSS = GPS;		// Only prepared for GPS right now
						
			getstr(aux,line,6,10);
			if (strcmp(aux,"I")==0) {
				sat.block = GPS_BLOCK_I;
			} else if (strcmp(aux,"II")==0) {
				sat.block = GPS_BLOCK_II;
			} else if (strcmp(aux,"IIA")==0) {
				sat.block = GPS_BLOCK_IIA;
			} else if (strcmp(aux,"IIR")==0) {
				sat.block = GPS_BLOCK_IIR;
			} else if (strcmp(aux,"IIR-M")==0) {
				sat.block = GPS_BLOCK_IIRM;
			} else {
				sat.block = UNKNOWN_BLOCK;
			}
			if (sat.block == GPS_BLOCK_II || sat.block == GPS_BLOCK_IIA) {
				for (i=0;i<MAX_FREQUENCIES_PER_GNSS;i++) {
					sat.PCO[i][0] = 0.279;
					sat.PCO[i][1] = 0;
					sat.PCO[i][2] = 1.023;
				}
		 	} else if (sat.block == GPS_BLOCK_I)  {
        		for (i=0;i<MAX_FREQUENCIES_PER_GNSS;i++) {
          			sat.PCO[i][0] = 0.21;
          			sat.PCO[i][1] = 0;
          			sat.PCO[i][2] = 0.854;
        		}
			} else {
				for (i=0;i<MAX_FREQUENCIES_PER_GNSS;i++) {
					sat.PCO[i][0] = sat.PCO[i][1] = sat.PCO[i][2] = 0;
				}
			}
			properlyRead = 1;
			readingInfo = 1;
		} else {
			if (readingInfo) {
				if (strcmp(aux,"Notes")==0) {
					readingInfo = 0;
				} else { // Satellite Data
					getstr(aux,line,7,3);
					sat.SVN = atoi(aux);
					getstr(aux,line,57,14);
					if (sat.SVN!=0 && strcmp(aux,"Launch failure")!=0) { // Valid line
						getstr(aux,line,11,2);
						sat.PRN = atoi(aux);
						
						getstr(aux,line,41,2);
						tm.tm_year = atoi(aux);
						if (tm.tm_year < 70) tm.tm_year+=100;
						getstr(aux,line,44,2);
						tm.tm_mon = atoi(aux)-1;
						getstr(aux,line,47,2);
						tm.tm_mday = atoi(aux);
						sat.tLaunch.MJDN = MJDN(&tm);
						sat.tLaunch.SoD = 0;
						
						if (line[68]==':') { // Satellite not decomissioned
							tm.tm_year = 130;
							tm.tm_mon = 0;
							tm.tm_mday = 1;
						} else { // Satellite decomissioned
							getstr(aux,line,70,2);
							tm.tm_year = atoi(aux);
							if (tm.tm_year < 70) tm.tm_year+=100;
							getstr(aux,line,73,2);
							tm.tm_mon = atoi(aux)-1;
							getstr(aux,line,76,2);
							tm.tm_mday = atoi(aux);
						} 
						sat.tDecommissioned.MJDN = MJDN(&tm);
						sat.tDecommissioned.SoD = 0;
						
						// Allocate memory
						constellation->numSatellites ++;
						constellation->sat = realloc(constellation->sat,sizeof(TConstellationElement)*constellation->numSatellites);
						
						// Store satellite
						memcpy(&constellation->sat[constellation->numSatellites-1],&sat,sizeof(TConstellationElement));
					}
				}
			}
		}
	}

	return (properlyRead && !error); // properly read AND no error
}

/*****************************************************************************
 * Name        : readConstellationFile
 * Description : Read a constellation status file from a filename
 *               File available at: 
 *                 http://gge.unb.ca/Resources/GPSConstellationStatus.txt
 *	             File compiled and maintaned by Richard B. Langley, Dept. of 
 *               Geodesy and Geomatics Engineering, University of New Brunswick.
 * Parameters  :
 * Name                           |Da|Unit|Description
 * char  *filename                 I  N/A  Name of the file
 * TConstellation  *constellation  O  N/A  Struct to save the data
 * Returned value (int)            O  N/A  Status of the function
 *                                         1  => File open and properly read
 *                                         0  => File open and failed to read
 *                                         -1 => File not open
 *****************************************************************************/
int readConstellationFile (char *filename, TConstellation *constellation) {
	FILE	*fd;
	int		ret;
	
	if (whatFileTypeIs(filename)!=ftConstellation) return 0;
	
	if ((fd = fopen(filename,"r"))) {
		ret = readConstellation (fd, constellation);
		fclose(fd);
		return ret;
	} else {
		// File do not exist
		return -1;
	}
}

/*****************************************************************************
 * Name        : readAntex
 * Description : Read an ANTEX file from a file descriptor
 * Parameters  :
 * Name                           |Da|Unit|Description
 * FILE  *fd                       I  N/A  File descriptor
 * TConstellation  *constellation  O  N/A  Struct to save the data of GNSS satellites
 * TAntennaList *antennaList       O  N/A  Struct to save the data of receiver 
 *                                         antennas
 * Returned value (int)            O  N/A  Status of the function
 *                                         1 => Properly read
 *                                         0 => Error
 *****************************************************************************/
int readAntex (FILE *fd, TConstellation *constellation, TAntennaList *antennaList) {
	char 					line[MAX_INPUT_LINE];
	char    				aux[100];
	char					*ca;
	int 					len = 0;
	struct tm				tm;
	int						properlyRead = 0;
	TConstellationElement	sat;
	TAntenna				ant;
	int						i;
	int						freqInd;
	int						readingHeader = 1;
	int						error = 0;
	int						PRN;
	int						readingWhat; // =0 Satellite    =1 Receiver antenna
	double					seconds;
	
	tm.tm_hour = 0;
	tm.tm_min = 0;
	tm.tm_sec = 0;
	
	while(getL(line,&len,fd)!=-1) {
		if (readingHeader) {
			if (strncmp(&line[60],"ANTEX VERSION / SYST",20)==0) {
				properlyRead = 1;
			} else if (strncmp(&line[60],"PCV TYPE / REFANT",17)==0) {
				// Do nothing
			} else if (strncmp(&line[60],"END OF HEADER",13)==0) {
				readingHeader = 0;
			}
		} else { // readingHeader==0
			if (strncmp(&line[60],"TYPE / SERIAL NO",16)==0) {
				getstr(aux,line,20,20);
				PRN = atoi(&aux[1]);
				if (PRN!=0 && (aux[0]=='G' || aux[0]=='R' || aux[0]=='E')) {
					// GNSS Satellite
					initConstellationElement(&sat);
					sat.GNSS = gnsschar2gnsstype(aux[0]);
					sat.PRN = PRN;
					getstr(aux,line,40,10);
					sat.SVN = atoi(&aux[1]);
					getstr(aux,line,0,20);
					sat.block = satblockstr2satblock(aux);
					readingWhat = 0;
				} else {
					// Receiver antenna
					initAntenna(&ant);
					getstr(aux,line,0,20);
					strcpy(ant.type,aux);
					// Separate the anntena.type into antenna and radome
					strcpy(ant.type_ant,ant.type);
					ca = strstr(ant.type_ant," ");
					if (ca != NULL) {
						*ca = '\0';
						strcpy(ant.type_radome,&ca[1]);
						trim(ant.type_radome);
					}
					readingWhat = 1;
				}
			} else if (strncmp(&line[60],"VALID",5)==0) {
				if (readingWhat==0) { // Satellite Antenna
					getstr(aux,line,2,4);
					tm.tm_year = atoi(aux)-1900;
					getstr(aux,line,10,2);
					tm.tm_mon  = atoi(aux)-1;
					getstr(aux,line,16,2);
					tm.tm_mday = atoi(aux);
					getstr(aux,line,22,2);
					tm.tm_hour = atoi(aux);
					getstr(aux,line,28,2);
					tm.tm_min  = atoi(aux);
					getstr(aux,line,33,10);
					tm.tm_sec  = atoi(aux);
					seconds = atof(aux);
					if (strncmp(&line[66],"FROM",4)==0) {
						sat.tLaunch.MJDN = MJDN(&tm);
						sat.tLaunch.SoD = tm.tm_hour*3600 + tm.tm_min*60 + seconds;
					} else if (strncmp(&line[66],"UNTIL",5)==0) {
						sat.tDecommissioned.MJDN = MJDN(&tm);
						sat.tDecommissioned.SoD = tm.tm_hour*3600 + tm.tm_min*60 + seconds;
					}
				}
			} else if (strncmp(&line[60],"START OF FREQUENCY",18)==0) {
				getstr(aux,line,4,2);
				freqInd = atoi(aux);
			} else if (strncmp(&line[60],"NORTH / EAST / UP",17)==0) {
				for (i=0;i<3;i++) {
					getstr(aux,line,i*10,10);
					if (readingWhat==0) { // Satellite Antenna
						sat.PCO[freqInd][i] = atof(aux)/1e3; // (ANTEX in mm)
					} else if (readingWhat==1) { // Receiver Antenna
						ant.PCO[freqInd][i] = atof(aux)/1e3; // (ANTEX in mm)
					}
				}
			} else if (strncmp(&line[60],"END OF ANTENNA",14)==0) {
				if (readingWhat==0) { // Satellite Antenna
					// Allocate memory
					constellation->numSatellites ++;
					constellation->sat = realloc(constellation->sat,sizeof(TConstellationElement)*constellation->numSatellites);
						
					// Store satellite
					memcpy(&constellation->sat[constellation->numSatellites-1],&sat,sizeof(TConstellationElement));
				} else if (readingWhat==1) { // Receiver Antenna
					// Allocate memory
					antennaList->numAnt ++;
					antennaList->ant = realloc(antennaList->ant,sizeof(TAntenna)*antennaList->numAnt);
					
					// Store satellite
					memcpy(&antennaList->ant[antennaList->numAnt-1],&ant,sizeof(TAntenna));
				}
			}
		}
	}
	
	return (properlyRead && !error); // properly read AND no error
}

/*****************************************************************************
 * Name        : readAntexFile
 * Description : Read an ANTEX file from a filename
 * Parameters  :
 * Name                           |Da|Unit|Description
 * char  *filename                 I  N/A  Name of the file
 * TConstellation  *constellation  O  N/A  Struct to save the data of GNSS satellites
 * TAntennaList *antennaList       O  N/A  Struct to save the data of receiver 
 *                                         antennas
 * Returned value (int)            O  N/A  Status of the function
 *                                         1  => File open and properly read
 *                                         0  => File open and failed to read
 *                                         -1 => File not open
 *****************************************************************************/
int readAntexFile (char *filename, TConstellation *constellation, TAntennaList *antennaList) {
	FILE	*fd;
	int		ret;
	
	if (whatFileTypeIs(filename)!=ftANTEX) return 0;
	
	if ((fd = fopen(filename,"r"))) {
		ret = readAntex (fd, constellation, antennaList);
		fclose(fd);
		return ret;
	} else {
		// File do not exist
		return -1;
	}
}

/*****************************************************************************
 * Name        : readDCB
 * Description : Read a DCB file from a file descriptor
 * Parameters  :
 * Name                           |Da|Unit|Description
 * FILE  *fd                       I  N/A  File descriptor
 * TDCBdata  *DCB                  O  N/A  Struct to save the DCB data
 * Returned value (int)            O  N/A  Status of the function
 *                                         1 => Properly read
 *                                         0 => Error
 *****************************************************************************/
int readDCB (FILE *fd, TDCBdata *DCB) {
	char 					line[MAX_INPUT_LINE];
	char					*pt;
	char					*ptaux;
	char    				aux[100];
	enum MeasurementType	meas;
	enum GNSSystem			GNSS;
	double					value;
	double					rms;
	int 					len = 0;
	struct tm				tm;
	int						properlyRead = 0;
	int						readingHeader = 1;
	int						PRN;
	int						writtingAbs;
	double					factor;
	
	tm.tm_hour = 0;
	tm.tm_min = 0;
	tm.tm_sec = 0;
	tm.tm_mday = 1;
	
	while(getL(line,&len,fd)!=-1) {
		if (readingHeader) {
			if ((pt=strstr(line,"YEAR-MONTH"))!=NULL) {
				// CODE'S MONTHLY GPS P1-C1 DCB SOLUTION, YEAR-MONTH 10-01          04-FEB-10 20:11
				DCB->n++;
				pt += 11; // Pointing to '10-01 [...]'
				getstr(aux,pt,0,2);
				tm.tm_year = atoi(aux);
				if (tm.tm_year <= 70) tm.tm_year += 100;
				getstr(aux,pt,3,2);
				tm.tm_mon = atoi(aux) - 1;
				DCB->block[DCB->n-1].startTime.MJDN = MJDN(&tm);
				tm.tm_mon ++;
				DCB->block[DCB->n-1].endTime.MJDN = MJDN(&tm);

			} else if (strstr(line,"CODE BIASES")!=NULL) {
				// DIFFERENTIAL (P1-C1) CODE BIASES FOR SATELLITES AND RECEIVERS:
				pt = strstr(line,"(");
				if (pt==NULL) return 0;
				pt++;
				ptaux = strstr(pt,"-");
				if (ptaux==NULL) return 0;
				strncpy(aux,pt,ptaux-pt);
				aux[ptaux-pt]='\0';
				meas = measstr2meastype(aux);
				DCB->block[DCB->n-1].measTo = meas;
				pt=ptaux+1;
				ptaux = strstr(pt,")");
				if (ptaux==NULL) return 0;
				strncpy(aux,pt,ptaux-pt);
				aux[ptaux-pt]='\0';
				meas = measstr2meastype(aux);
				DCB->block[DCB->n-1].measFrom = meas;
				// Check to create the absolute measurement (from P1 to NA)
				if ((DCB->block[DCB->n-1].measFrom==C2P && DCB->block[DCB->n-1].measTo==C1P) || (DCB->block[DCB->n-1].measFrom==C1P && DCB->block[DCB->n-1].measTo==C2P)) {
					DCB->block[DCB->n].measFrom = C1P;
					DCB->block[DCB->n].measTo = NA;
					memcpy(&DCB->block[DCB->n].startTime,&DCB->block[DCB->n-1].startTime,sizeof(TTime));
					memcpy(&DCB->block[DCB->n].endTime,&DCB->block[DCB->n-1].endTime,sizeof(TTime));
					factor = 1.0/(1.0-(GPSf1*GPSf1)/(GPSf2*GPSf2)); // 1/(1-f1^2/f2^2)
					if (DCB->block[DCB->n-1].measFrom==C2P) factor*= -1.0;
					writtingAbs = 1;
				} else {
					writtingAbs = 0;
				}
			} else if (strncmp(line,"PRN / STATION NAME",18)==0) {
				// PRN / STATION NAME        VALUE (NS)  RMS (NS)
				// ***   ****************    *****.***   *****.***
				if (getL(line,&len,fd)==-1) {
					return 0;
				}
				readingHeader = 0;
			}
		} else { //readingHeader == 0
			GNSS = gnsschar2gnsstype(line[0]);
			getstr(aux,line,1,2);
			PRN = atoi(aux);
			if (PRN != 0) { // Satellite unidentified or blank line
				properlyRead = 1;
				getstr(aux,line,26,9);
				value = atof(aux);
				getstr(aux,line,38,9);
				rms = atof(aux);
				// From ns to meters
				value *= 1e-9*c;
				rms *= 1e-9*c;
				DCB->block[DCB->n-1].dcb[GNSS][PRN] = value;
				DCB->block[DCB->n-1].available[GNSS][PRN] = 1;
				if (writtingAbs) {
					DCB->block[DCB->n].dcb[GNSS][PRN] = value*factor;
					DCB->block[DCB->n].available[GNSS][PRN] = 1;
				}
			}
		}
	}
	if (writtingAbs) DCB->n++;
	
	return properlyRead;
}

/*****************************************************************************
 * Name        : readDCBFile
 * Description : Read a DCB file from a filename
 * Parameters  :
 * Name                           |Da|Unit|Description
 * char  *filename                 I  N/A  Name of the file
 * TDCBdata  *DCB                  O  N/A  Struct to save the DCB data
 * Returned value (int)            O  N/A  Status of the function
 *                                         1  => File open and properly read
 *                                         0  => File open and failed to read
 *                                         -1 => File not open
 *****************************************************************************/
int readDCBFile (char *filename, TDCBdata *DCB) {
	FILE	*fd;
	int 	ret;
	enum fileType				ft;
	
  ft = whatFileTypeIs(filename);

	if (ft!=ftP1C1DCB && ft!=ftP1P2DCB) return 0;

	if ((fd = fopen(filename,"r"))) {
		ret = readDCB (fd, DCB);
		fclose(fd);
		return ret;
	} else {
		// File do not exist
		return -1;
	}
}

/*****************************************************************************
 * Name        : readRecType
 * Description :  Read a GPS Receiver type file from a file descriptor
 * Parameters  :
 * Name                           |Da|Unit|Description
 * FILE  *fd                       I  N/A  File descriptor
 * TReceiverList  *recList         O  N/A  Struct to save the Receiver data
 * Returned value (int)            O  N/A  Status of the function
 *                                         1 => Properly read
 *                                         0 => Error
 *****************************************************************************/
int readRecType (FILE *fd, TReceiverList *recList) {
	char 					line[MAX_INPUT_LINE];
	char					cr;
	char    				aux[100];
	int 					len = 0;
	
	while(getL(line,&len,fd)!=-1) {
		if (line[0]!='#') { // Line is not a comment
			getstr(aux,line,0,20);
			if (aux[0]!='\0' && aux[0]!='\n') {
				recList->n++;
				recList->rec = realloc(recList->rec,sizeof(TReceiverData)*recList->n);
				strcpy(recList->rec[recList->n-1].name,aux);
				cr = line[21];
				switch (cr) {
					case '0':
						recList->rec[recList->n-1].type = rtUNKNOWN;
						break;
					case '1':
						recList->rec[recList->n-1].type = rtCROSS;
						break;
					case '2':
						recList->rec[recList->n-1].type = rtNOP1;
						break;
					case '3':
						recList->rec[recList->n-1].type = rtNOMINAL;
						break;
					default:
						recList->rec[recList->n-1].type = rtUNKNOWN;
				}
			}
		}
	}
	
	return 1;
}

/*****************************************************************************
 * Name        : readRecTypeFile
 * Description : Read a GPS Receiver type file from a filename
 * Parameters  :
 * Name                           |Da|Unit|Description
 * char  *filename                 I  N/A  Name of the file
 * TReceiverList  *recList         O  N/A  Struct to save the Receiver data
 * Returned value (int)            O  N/A  Status of the function
 *                                         1  => File open and properly read
 *                                         0  => File open and failed to read
 *                                         -1 => File not open
 *****************************************************************************/
int readRecTypeFile (char *filename, TReceiverList *recList) {
	FILE	*fd;
	int		ret;
	
	if (whatFileTypeIs(filename)!=ftRECEIVERTYPE) return 0;
	
	if ((fd = fopen(filename,"r"))) {
		ret = readRecType (fd, recList);
		fclose(fd);
		return ret;
	} else {
		// File do not exist
		return -1;
	}
}

/*****************************************************************************
 * Name        : readSINEX
 * Description :  Read a SINEX station positions file from a file descriptor
 * Parameters  :
 * Name                           |Da|Unit|Description
 * FILE  *fd                       I  N/A  File descriptor
 * TStationList  *stationList      O  N/A  Struct to save the station data
 * Returned value (int)            O  N/A  Status of the function
 *                                         1 => Properly read
 *                                         0 => Error
 *****************************************************************************/
int readSINEX (FILE *fd, TStationList *stationList) {
	char 	line[MAX_INPUT_LINE];
	char	aux[100];
	int 	len = 0;
	int		section;	// =0 => Unimportant stuff
						// =1 => SITE/ID section
						// =2 => SOLUTION/ESTIMATE section
	int 	properlyRead = 0;
	int		auxN;
	int		error = 0;
	int		i=0;
	char	prevName[5] = "";
	
	section = 0;
	while(getL(line,&len,fd)!=-1) {
		if (line[0]=='+') { // Change of section
			if (strncmp(line,"+SITE/ID",8)==0) section = 1;
			else if (strncmp(line,"+SOLUTION/ESTIMATE",18)==0) {
				section = 2;
				auxN = 0;
				// Reserve memory for detected stations
				stationList->station = malloc(sizeof(TStation)*stationList->n);
			} else section = 0;
		} else if (line[0]!='*' && line[0]!='-') {
			if (section==1) {
				stationList->n++;
			} else if (section==2) {
				getstr(aux,line,7,3);
				if (strcmp(aux,"STA")==0) { // It is a station
					properlyRead = 1; // Any station has been found
					getstr(aux,line,14,4);
					if (strcmp(prevName,aux)!=0) { // New station
						auxN++;
						strcpy(stationList->station[auxN-1].name,aux);
						strcpy(prevName,aux);
					}
					getstr(aux,line,46,23);
					if (line[10]=='X') i=0;
					else if (line[10]=='Y') i=1;
					else if (line[10]=='Z') i=2;
					else error = 1;
					stationList->station[auxN-1].pos[i] = atof(aux);
				}
			}
		}
	}
	
	if (auxN!=stationList->n) error = 1; // Stations number not properly parsed
	return (properlyRead && !error); // properly read AND no error
}

/*****************************************************************************
 * Name        : readSINEXFile
 * Description : Read a SINEX station positions file from a filename
 * Parameters  :
 * Name                           |Da|Unit|Description
 * char  *filename                 I  N/A  Name of the file
 * TStationList  *stationList      O  N/A  Struct to save the station data
 * Returned value (int)            O  N/A  Status of the function
 *                                         1  => File open and properly read
 *                                         0  => File open and failed to read
 *                                         -1 => File not open
 *****************************************************************************/
int readSINEXFile (char *filename, TStationList *stationList) {
	FILE	*fd;
	int		ret;
	
	if (whatFileTypeIs(filename)!=ftSINEX) return 0;
	
	if ((fd = fopen(filename,"r"))) {
		ret = readSINEX (fd, stationList);
		fclose(fd);
		return ret;
	} else {
		// File do not exist
		return -1;
	}
}
