/****************************************************************************
   Copyright & License:
   ====================
   
   Copyright 2009 - 2010 gAGE/UPC & ESA
   
   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at
   
   http://www.apache.org/licenses/LICENSE-2.0
   
   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 ***************************************************************************/

/****************************************************************************
 * Copyright: gAGE/UPC & ESA
 * Project: EDUNAV GNSS Lab Tool 
 * Author: Pere Ramos-Bosch (group of Astronomy and GEomatics - gAGE/UPC)
 * Maintainer: Adria Rovira-Garcia ( gAGE/UPC )
 *             glab.gage @ upc.edu       
 * File: output.c
 * Code Management Tool File Version: 2.0  Revision: 0 
 * Date: 2010/11/30
 ***************************************************************************/

/****************************************************************************
 * MODULE DESCRIPTION
 *
 * Name: output
 * Language: C
 *
 * Purpose:
 *  The purpose of this module is to act as output for the solution of the filter
 *
 * Function:
 *  This class is in charge of outputting the results of the Kalman filter done 
 *  by the filter module
 *
 * Dependencies:
 *  dataHandling, filter, model
 *
 * Files modified:
 *  None
 *
 * Files read:
 *  None
 *
 * Resources usage:
 *  See Design Document
 ****************************************************************************/

/****************************
 *       RELEASE_HISTORY
 * -------------
 *          gLAB v1.3.1
 * Release: 2010/01/22
 * Change Log: First public version
 * -------------
 *          gLAB v1.3.2
 * Release: 2010/02/02
 * Change Log: Imported SATDIFF, SATSTAT and SATSTATTOT printing from gLAB.c
 *             Added satellite SATPVT message
 *             Change in SATDIFF and CS messages to show the GNSS system that 
 *                  the satellite belongs to
 *             It is now possible to print in the MODEL and SATDIFF messages 
 *                  the inertial velocity of the satellite
 *             PREFIT message now includes the weight that will be applied to 
 *                  the filter
 *             ZTD and estimated residual tropospheric delay is now printed 
 *                  in the FILTER message
 *             Forward/Backward processing included
 *             SATSTAT message will now provide the GNSS system it refers to
 * -------------
 *          gLAB v1.4.0
 * Release: 2010/06/21
 * Change Log: Included prealignCP option to set/unset carrier-phase prealignment
 *             C1/P1 flexibility option is added to allow flexible handling of C1 and P1
 *             Receiver apriori position INFO message now provides the origin of the
 *                  position
 *             Changed printInput and printMeas function to show all the measurements 
 *                  independently that they have been set as available or unavailable
 *             Corrected a bug when computing the position in trajectories
 *             Cycle-slip detector changed to specify when it does not have enough measurements
 * -------------
 *          gLAB v1.4.0
 * Release: 2010/08/19
 * Change Log: No changes in this file.
 * -------------
 *          gLAB v1.4.1
 * Release: 2010/07/09
 * Change Log: No changes in this file.
 * -------------
 *          gLAB v1.4.2
 * Release: 2010/07/31
 * Change Log: No changes in this file.
 * -------------
 *          gLAB v1.4.3
 * Release: 2010/08/31
 * Change Log: Satellite elevation and azimuth addition in the print MODEL message.
 * -------------
 *          gLAB v1.4.4
 * Release: 2010/09/22
 * Change Log: Included new DCB management.
 * -------------
 *          gLAB v1.4.5
 * Release: 2010/10/31
 * Change Log: Tropospheric correction option is extended to select independently the nominal and mapping used.
 *             Dilution Of Precision extension: (GDOP, PDOP, TDOP, VDOP, HDOP) new variables.
 *             Some minor floating length adjustments. 
 *             Filter 'inf' value for clock and coordinates noise process and initial covariance substituted
 *                  to real floating quantities.
 * -----------
 *          gLAB v2.0.0
 * Release: 2010/11/30
 * Change Log: gLAB version released together with the Educational Book:
 *             "Global Navigation Satellite Systems: Fundamentals, Algorithms and Laboratory Exercises"
 *             Authors: J. Sanz, JM. Juan and M. Hernandez-Pajares.
 *
 *             SATSEL debugging message included in the output messages.
 * -----------
 *       END_RELEASE_HISTORY
 *****************************/

/* External classes */
#include "output.h"

/**************************************
 * Declarations of internal operations
 **************************************/

/*****************************************************************************
 * Name        : printParameters
 * Description : Print Input parameters
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TOptions  *options              I  N/A  TOptions structure
 *****************************************************************************/
void printParameters(TOptions *options, TEpoch *epoch, TGNSSproducts *products) {
	int		i,j;
	int		aux;
	char	auxstr[300];
	
	if (!options->printInfo) return;
	
	if (epoch->receiver.name[0] == '\0') {
		strcpy(auxstr,"Unknown");
	} else {
		strcpy(auxstr,epoch->receiver.name);
	}
	printf("INFO INPUT Station marker: %s\n",auxstr);
	if (epoch->receiver.antenna.type[0] == '\0') {
		strcpy(auxstr,"Unknown");
	} else {
		strcpy(auxstr,epoch->receiver.antenna.type);
	}
	printf("INFO INPUT Antenna type:   %s\n",auxstr);
	if (epoch->receiver.type[0] == '\0') {
		printf("INFO INPUT Receiver type:  0-Unknown\n");
	} else {
		if (epoch->receiver.recType == rtCROSS) strcpy(auxstr,"1-Cross-correlated");
		else if (epoch->receiver.recType == rtNOP1) strcpy(auxstr,"2-No P1");
		else if (epoch->receiver.recType == rtNOMINAL) strcpy(auxstr,"3-Consistent measurements");
		else strcpy(auxstr,"Unknown");
		printf("INFO INPUT Receiver type:  %s [%s]\n",epoch->receiver.type,auxstr);
	}
	printf("INFO PREPROCESSING Prealign carrier phase measurements: %3s\n",options->prealignCP?"ON":"OFF");
	printf("INFO PREPROCESSING Decimation %d\n",options->decimate);
	aux = 0;
	for (i=0;i<MAX_GNSS;i++) {
		for (j=0;j<MAX_SATELLITES_PER_GNSS;j++) {
			if (options->includeSatellite[i][j]==0) {
				if (!aux) {
					printf("INFO PREPROCESSING Excluding satellites:");
				}
				printf(" %c%02d",gnsstype2char(i),j);
				aux = 1;
			}
		}
	}
	if (aux) printf("\n");
	printf("INFO PREPROCESSING Usable frequencies [GPS]:");
	if (options->usableFreq[GPS][1]) printf(" F%1d",1);
	if (options->usableFreq[GPS][2]) printf(" F%1d",2);
	if (options->usableFreq[GPS][5]) printf(" F%1d",5);
	
	printf("\n");
	printf("INFO PREPROCESSING Elevation mask: %.2lf\n",options->elevationMask/d2r);
	printf("INFO PREPROCESSING Discard satellites under eclipse conditions: %3s\n",options->discardEclipsedSatellites?"YES":"NO");
	printf("INFO PREPROCESSING Receiver apriori position: %14.4lf %14.4lf %14.4lf  [",epoch->receiver.aproxPosition[0],epoch->receiver.aproxPosition[1],epoch->receiver.aproxPosition[2]);
	if (options->receiverPositionSource == rpRINEX) printf("RINEX");
	else if (options->receiverPositionSource == rpSINEX) printf("SINEX");
	else if (options->receiverPositionSource == rpCALCULATE) printf("TO_BE_CALCULATED");
	else if (options->receiverPositionSource == rpSET) printf("SET");
	printf("]\n");
	if (options->csLi) {
		printf("INFO PREPROCESSING CycleSlip Li:    ON  [Min:%.3lf  Max:%.3lf  t:%.3lf]\n",options->csLimin,options->csLimax,options->csLit);
	} else {
		printf("INFO PREPROCESSING CycleSlip Li:   OFF\n");
	}
	if (options->csBW) {
		printf("INFO PREPROCESSING CycleSlip BW:    ON  [Min:%.3lf  Max:%.3lf  Slope:%.3lf]\n",options->csBWmin,options->csBWmax,options->csBWslope);
	} else {
		printf("INFO PREPROCESSING CycleSlip BW:   OFF\n");
	}
	if (options->csL1C1) {
		printf("INFO PREPROCESSING CycleSlip L1C1:  ON  [Slope:%.3lf  Window:%d]\n",options->csL1C1slope,options->csL1C1window);
	} else {
		printf("INFO PREPROCESSING CycleSlip L1C1: OFF\n");
	}
	if (epoch->receiver.recType == rtNOP1) {
		printf("INFO PREPROCESSING Receiver type does not provide P1... I will use C1 instead of P1\n"); 
	}
	printf("INFO MODELING Satellite clock offset correction: %3s\n",options->satelliteClockCorrection?"ON":"OFF");
	printf("INFO MODELING Consider satellite movement during signal flight time: %3s\n",options->timeTrans?"ON":"OFF");
	printf("INFO MODELING Consider Earth rotation during signal flight time: %3s\n",options->earthRotation?"ON":"OFF");
	for (i=1;i<3;i++)
		printf("INFO MODELING Receiver Antenna Phase Center Offset (PCO): F%1d %8.5lf %8.5lf %8.5lf\n",i,epoch->receiver.antenna.PCO[i][0],epoch->receiver.antenna.PCO[i][1],epoch->receiver.antenna.PCO[i][2]);
	printf("INFO MODELING Receiver Antenna Reference Point (ARP): %7.4lf %7.4lf %7.4lf\n",epoch->receiver.ARP[0],epoch->receiver.ARP[1],epoch->receiver.ARP[2]);
	printf("INFO MODELING Relativistic clock correction: %3s\n",options->relativisticCorrection?"ON":"OFF");
	if (options->ionoModel == NoIono) strcpy(auxstr,"OFF");
	else if (options->ionoModel == Klobuchar) strcpy(auxstr,"Klobuchar");
	else if (options->ionoModel == IONEX) strcpy(auxstr,"IONEX");
	printf("INFO MODELING Ionosphere model: %s\n",auxstr);
	printf("INFO MODELING Troposphere model: %3s\n",options->troposphericCorrection?"ON":"OFF");
	if (options->troposphericCorrection  == 1) { 
 		if (options->tropNominal == SimpleNominal) printf("INFO MODELING Troposphere model: Simple Nominal\n");
		else if (options->tropNominal == UNB3Nominal) printf("INFO MODELING Troposphere model: UNB-3 Nominal\n");
		if (options->tropMapping == SimpleMapping) printf("INFO MODELING Troposphere model: Simple Mapping\n");
		else if (options->tropMapping == NiellMapping) printf("INFO MODELING Troposphere model: Niell Mapping\n");
	}
	if (options->p1c1dcbModel == p1c1NONE) strcpy(auxstr,"OFF");
	else if (options->p1c1dcbModel == p1c1FLEXIBLE) strcpy(auxstr,"Flexible");
	else if (options->p1c1dcbModel == p1c1STRICT) strcpy(auxstr,"Strict");
	printf("INFO MODELING P1-C1 DCB model: %s\n",auxstr);
	if (options->p1p2dcbModel == p1p2NONE) strcpy(auxstr,"OFF");
	else if (options->p1p2dcbModel == p1p2RINEX) strcpy(auxstr,"RINEX");
	else if (options->p1p2dcbModel == p1p2DCB) strcpy(auxstr,"DCB");
	printf("INFO MODELING P1-P2 DCB model: %s\n",auxstr);
	printf("INFO MODELING Wind up correction: %3s\n",options->windUpCorrection?"ON":"OFF");
	printf("INFO MODELING Solid tides correction: %3s\n",options->solidTidesCorrection?"ON":"OFF");
	printf("INFO MODELING Relativistic path range correction: %3s\n",options->gravitationalCorrection?"ON":"OFF");
	if (products->type == SP3) {
		printf("INFO MODELING Orbit interpolation degree: %2d\n",options->orbitInterpolationDegree);
		printf("INFO MODELING Clock interpolation degree: %2d\n",options->clockInterpolationDegree);
	}
	printf("INFO MODELING Use satellite 'SV Health' flag of navigation message: %s\n",options->satelliteHealth?"ON":"OFF");
	for (i=0;i<options->totalFilterMeasurements;i++) {
		printf("INFO FILTER Measurement: %1d %3s  StdDev:%.2lf",i+1,meastype2measstr(options->measurement[i]),options->fixedMeasurementWeight[i]);
		if (options->elevationMeasurementWeight[i][0]!=0) {
			printf("+%.2lf*e^(elev/%.2lf)",options->elevationMeasurementWeight[i][0],options->elevationMeasurementWeight[i][1]/d2r);
		}
		if (options->smoothMeas[i]!=NA) 
			printf("  SmoothWith:%3s [n=%d]",meastype2measstr(options->smoothMeas[i]),options->smoothEpochs);
		printf("\n");
	}
	printf("INFO FILTER Carrierphase is used: %3s\n",options->usePhase?"YES":"NO");
	printf("INFO FILTER Estimate troposphere: %3s\n",options->estimateTroposphere?"ON":"OFF");
	printf("INFO FILTER %s Processing\n",(options->filterIterations==1)?"Forward":"Backward"); 
	for (i=0;i<MAX_PAR;i++) {
		if (i==DR_UNK) strcpy (auxstr,"Position:");
		else if (i==DT_UNK) strcpy (auxstr,"Clock:");
		else if (i==TROP_UNK) strcpy (auxstr,"Troposphere:");
		else if (i==BIAS_UNK) strcpy (auxstr,"Ambiguity:"); 
		printf("INFO FILTER Parameters [Phi,Q,P0] %-14s",auxstr);
		for (j=0;j<3;j++) {
				printf(" %9.2e",options->filterParams[j][i]);
		}
		printf("\n");
	}
	printf("INFO OUTPUT Satellite Velocity:     %s\n",options->satVel==svINERTIAL?"Inertial":"ITRF");
	printf("INFO PRINT INFO:       %3s\n",options->printInfo?"ON":"OFF");
	printf("INFO PRINT CS:         %3s\n",options->printCycleslips?"ON":"OFF");
	printf("INFO PRINT INPUT:      %3s\n",options->printInput?"ON":"OFF");
	printf("INFO PRINT MEAS :      %3s\n",options->printMeas?"ON":"OFF");
	printf("INFO PRINT MODEL:      %3s\n",options->printModel?"ON":"OFF");
	printf("INFO PRINT EPOCHSAT:   %3s\n",options->printSatellites?"ON":"OFF");
	printf("INFO PRINT SATSEL:     %3s\n",options->printSatSel?"ON":"OFF");
	printf("INFO PRINT PREFIT:     %3s\n",options->printPrefit?"ON":"OFF");
	printf("INFO PRINT POSTFIT:    %3s\n",options->printPostfit?"ON":"OFF");
	printf("INFO PRINT FILTER:     %3s\n",options->printFilterSolution?"ON":"OFF");
	printf("INFO PRINT OUTPUT:     %3s\n",options->printOutput?"ON":"OFF");
	printf("INFO PRINT SATDIFF:    %3s\n",options->printSatDiff?"ON":"OFF");
	printf("INFO PRINT SATSTAT:    %3s\n",options->printSatStat?"ON":"OFF");
	printf("INFO PRINT SATSTATTOT: %3s\n",options->printSatStatTot?"ON":"OFF");
}

/*****************************************************************************
 * Name        : printError
 * Description : Print an ERROR message and exit with code -1
 * Parameters  :
 * Name                           |Da|Unit|Description
 * char  *message                  I  N/A  Message
 * TOptions  *options              I  N/A  TOptions structure 
 *****************************************************************************/
void printError (char *message, TOptions *options) {
	printf("ERROR %s\n",message);
	exit(-1);
}
 
/*****************************************************************************
 * Name        : printInfo
 * Description : Print an INFO message
 * Parameters  :
 * Name                           |Da|Unit|Description
 * char  *message                  I  N/A  Message
 * TOptions  *options              I  N/A  TOptions structure 
 *****************************************************************************/
void printInfo (char *message, TOptions *options) {
	if (options->printInfo)
		printf("INFO %s\n",message);
}

/*****************************************************************************
 * Name        : printInput
 * Description : Print the input data
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  TEpoch structure
 * TOptions  *options              I  N/A  TOptions structure 
 *****************************************************************************/
void printInput(TEpoch *epoch, TOptions *options) {
	int		i,j;
	double	measurements[10];
	int		res[10];
	
	if (options->printInput) {
		for (i=0;i<epoch->numSatellites;i++) {
			printf ("INPUT    %17s %3s %2d %6d",t2doystr(&epoch->t),gnsstype2gnssstr(epoch->sat[i].GNSS),epoch->sat[i].PRN,epoch->sat[i].arcLength);
			if (epoch->sat[i].GNSS == GPS) {
				res[0] = getMeasModelValue(epoch,epoch->sat[i].GNSS,epoch->sat[i].PRN,C1C,&measurements[0],NULL);
				res[1] = getMeasModelValue(epoch,epoch->sat[i].GNSS,epoch->sat[i].PRN,C1P,&measurements[1],NULL);
				res[2] = getMeasModelValue(epoch,epoch->sat[i].GNSS,epoch->sat[i].PRN,C2P,&measurements[2],NULL);
				res[3] = getMeasModelValue(epoch,epoch->sat[i].GNSS,epoch->sat[i].PRN,L1P,&measurements[3],NULL);
				res[4] = getMeasModelValue(epoch,epoch->sat[i].GNSS,epoch->sat[i].PRN,L2P,&measurements[4],NULL);
				for (j=0;j<5;j++) {
					if (measurements[j] ==-1 || res[j]==0) measurements[j] = 0;
					printf (" %14.4lf",measurements[j]);
				}
			} else if (epoch->sat[i].GNSS == Galileo) {
				res[0] = getMeasModelValue(epoch,epoch->sat[i].GNSS,epoch->sat[i].PRN,C1A,&measurements[0],NULL);
				res[1] = getMeasModelValue(epoch,epoch->sat[i].GNSS,epoch->sat[i].PRN,C1B,&measurements[1],NULL);
				res[2] = getMeasModelValue(epoch,epoch->sat[i].GNSS,epoch->sat[i].PRN,C1C,&measurements[2],NULL);
				res[3] = getMeasModelValue(epoch,epoch->sat[i].GNSS,epoch->sat[i].PRN,C7Q,&measurements[3],NULL);
				res[4] = getMeasModelValue(epoch,epoch->sat[i].GNSS,epoch->sat[i].PRN,C8Q,&measurements[4],NULL);
				res[5] = getMeasModelValue(epoch,epoch->sat[i].GNSS,epoch->sat[i].PRN,L1A,&measurements[5],NULL);
				res[6] = getMeasModelValue(epoch,epoch->sat[i].GNSS,epoch->sat[i].PRN,L1B,&measurements[6],NULL);
				res[7] = getMeasModelValue(epoch,epoch->sat[i].GNSS,epoch->sat[i].PRN,L1C,&measurements[7],NULL);
				res[8] = getMeasModelValue(epoch,epoch->sat[i].GNSS,epoch->sat[i].PRN,L7Q,&measurements[8],NULL);
				res[9] = getMeasModelValue(epoch,epoch->sat[i].GNSS,epoch->sat[i].PRN,L8Q,&measurements[9],NULL);
				for (j=0;j<10;j++) {
					if (measurements[j] ==-1 || res[j]==0) measurements[j] = 0;
					printf (" %14.4lf",measurements[j]);
				}
			} else if (epoch->sat[i].GNSS == GLONASS) {
				res[0] = getMeasModelValue(epoch,epoch->sat[i].GNSS,epoch->sat[i].PRN,C1C,&measurements[0],NULL);
				res[1] = getMeasModelValue(epoch,epoch->sat[i].GNSS,epoch->sat[i].PRN,C2C,&measurements[1],NULL);
				res[2] = getMeasModelValue(epoch,epoch->sat[i].GNSS,epoch->sat[i].PRN,L1P,&measurements[2],NULL);
				res[3] = getMeasModelValue(epoch,epoch->sat[i].GNSS,epoch->sat[i].PRN,L2P,&measurements[3],NULL);
				for (j=0;j<4;j++) {
					if (measurements[j] ==-1 || res[j]==0) measurements[j] = 0;
					printf (" %14.4lf",measurements[j]);
				}
			} else if (epoch->sat[i].GNSS == GEO) {
				res[0] = getMeasModelValue(epoch,epoch->sat[i].GNSS,epoch->sat[i].PRN,C1C,&measurements[0],NULL);
				for (j=0;j<1;j++) {
					if (measurements[j] ==-1 || res[j]==0) measurements[j] = 0;
					printf (" %14.4lf",measurements[j]);
				}
			}
			printf("\n");
		}
	}
}

/*****************************************************************************
 * Name        : printMeas
 * Description : Print information on the measurements of the input and azimuth
 *               and elevation of the satellite
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  TEpoch structure
 * TOptions  *options              I  N/A  TOptions structure 
 *****************************************************************************/
 void printMeas(TEpoch *epoch, TOptions *options) {
 	int						i,j;
 	int						n[MAX_GNSS];
 	int						maxN = 0;
 	char					measStr[MAX_GNSS][MAX_MEASUREMENTS_PER_SATELLITE*2*3+1];
 	int						len;
 	enum MeasurementType	measChain[MAX_GNSS][MAX_MEASUREMENTS_PER_SATELLITE*2];
 	double					measVal;
 	char					mS[2][4];
 	char					format[20];
 	int						res;
 	
 	if (options->printMeas) {
 		for (i=0;i<MAX_GNSS;i++) {
 			n[i] = 0;
 			measStr[i][0] = '\0';
 			len = 0;
 			for (j=1;j<ENDMEAS;j+=4) {
 				// j=1 is a pseudorange, and then pseudoranges come every 4 identifiers.
 				// See the "enum MeasurementType" enumerator at dataHandling.h file for
 				// more details 
 				if (epoch->measOrder[i].meas2Ind[j]!=-1 || epoch->measOrder[i].meas2Ind[j+1]!=-1) {
 					// See if the Pseudorange OR the associated Carrier Phase (j+1) are available
 					measChain[i][n[i]] = j;
 					measChain[i][n[i]+1] = j+1;
 					n[i]+=2;
 					strcpy(mS[0], meastype2measstr(j));
 					strcpy(mS[1], meastype2measstr(j+1));
 					if (len!=0) {
 						len += sprintf(&measStr[i][len],":");
 					}
 					len += sprintf(&measStr[i][len],"%s:%s",mS[0],mS[1]);
 				}
 			}
 			if (n[i]>maxN) maxN = n[i];
 		}
		for (i=0;i<epoch->numSatellites;i++) {
			printf ("MEAS     %17s %3s %2d %8.3lf %8.3lf %2d ",t2doystr(&epoch->t),gnsstype2gnssstr(epoch->sat[i].GNSS),epoch->sat[i].PRN,epoch->sat[i].elevation/d2r,epoch->sat[i].azimuth/d2r,n[epoch->sat[i].GNSS]);
			sprintf (format,"%%%ds",maxN*4);
			printf(format,measStr[epoch->sat[i].GNSS]);
			for (j=0;j<n[epoch->sat[i].GNSS];j++) {
				res = getMeasModelValue(epoch,epoch->sat[i].GNSS,epoch->sat[i].PRN,measChain[epoch->sat[i].GNSS][j],&measVal,NULL);
				if (measVal==-1 || res==0) measVal = 0;
				printf(" %14.4lf",measVal);
			}
			printf("\n");
		}
 	}
 }

/*****************************************************************************
 * Name        : printCS
 * Description : Prints the cycle-slip information when one is found
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  Reference epoch
 * int  satIndex                   I  N/A  Satellite index
 * int  LiCheck                    I  N/A  Li Flag (1-> cycle-slip found in Li)
 * double  LiLiestDiff             I    m  Absolute difference between measured 
 *                                         and estimated Li values
 * double  LiThreshold             I    m  Threshold for LiLiestDiff
 * int  BwCheck                    I  N/A  Bw Flag (1-> cycle-slip found in Bw)
 * double  nBwdiff                 I  cyc  Absolute difference between measured 
 *                                         and estimated Bw values
 * double  nBwdiffThreshold        I  cyc  Threshold for nBwdiff
 * int  L1C1Check                  I  N/A  L1C1 Flag (1->cycle-slip found in L1C1)
 * double  L1C1diff                I    m  Absolute difference between measured 
 *                                         and estimated L1C1 values
 * double  L1C1Threshold           I    m  Threshold for L1C1diff
 * TOptions  *options              I  N/A  TOptions structure
 *****************************************************************************/
void printCS (TEpoch *epoch, int satIndex, int LiCheck, double LiLiestDiff, double LiThreshold, int BwCheck, double nBwdiff, double nBwdiffThreshold, int L1C1Check, double L1C1diff, double L1C1Threshold, TOptions *options) {
	if (options->printCycleslips) {
		printf("CS       %17s %3s %2d",
				t2doystr(&epoch->t),
				gnsstype2gnssstr(epoch->sat[satIndex].GNSS),
				epoch->sat[satIndex].PRN
				);
		if (options->csLi) {
			if (LiThreshold!=-1) printf(" Li:%1d (%lf [%lf])",LiCheck,LiLiestDiff,LiThreshold);
			else printf(" Li:%1d (NOT_ENOUGH_MEASUREMENTS)",LiCheck);
		}
		if (options->csBW) {
			if (nBwdiffThreshold!=-1) printf(" Bw:%1d (%lf [%lf])",BwCheck,nBwdiff,nBwdiffThreshold);
			else printf(" Bw:%1d (NOT_ENOUGH_MEASUREMENTS)",BwCheck);
		}
		if (options->csL1C1) {
			if (L1C1Threshold!=-1) printf(" L1C1:%1d (%lf [%lf])",L1C1Check,L1C1diff,L1C1Threshold);
			else printf(" L1C1:%1d (NOT_ENOUGH_MEASUREMENTS)",L1C1Check);
		}
		printf ("\n");
	}
}
/*****************************************************************************
 * Name        : printModel
 * Description : Prints each of the components of the model for a specific 
 *               satellite and measurement
 * Parameters  :
 * Name                                    |Da|Unit|Description
 * TEpoch  *epoch                           I  N/A  Reference epoch
 * int  satIndex                            I  N/A  Selected satellite
 * int  measIndex                           I  N/A  Selected measurement
 * double  flightTime                       I    s  Signal flight time
 * double  geometricDistance                I    m  Geometric distance between satellite and receiver
 * double  satelliteClockModel              I    m  Satellite clock error
 * double  satellitePhaseCenterProjection   I    m  Projection of the satellite phase center into the LoS
 * double  receiverPhaseCenterProjection    I    m  Projection of the receiver phase center into the LoS 
 * double  receiverARPProjection            I    m  Projection of the receiver ARP into the LoS
 * double  relativityModel                  I    m  Relativity model
 * double  windUpMetModel                   I    m  Wind up model (for carrier phase only)
 * double  troposphereModel                 I    m  Troposphere model
 * double  ionosphereModel                  I    m  Ionosphere model
 * double  gravitationalDelayModel          I    m  Gravitational delay model
 * double  TGDModel                         I    m  TGD model
 * double  solidTidesModel                  I    m  Projection of the effect of solid tides into the LoS
 * TOptions  *options                       I  N/A  TOptions structure
 *****************************************************************************/
void printModel (TEpoch *epoch, int satIndex, int measIndex, double flightTime, double geometricDistance, double satelliteClockModel, double satellitePhaseCenterProjection, double receiverPhaseCenterProjection, double receiverARPProjection, double relativityModel, double windUpMetModel, double troposphereModel, double ionosphereModel, double gravitationalDelayModel, double TGDModel, double solidTidesModel, TOptions *options) {
	double 	velocity[3];
	double	measurement;
	double	model;
	int		res;
	
	if (options->satVel==svINERTIAL) {
		memcpy(velocity,epoch->sat[satIndex].velocity,3*sizeof(double));
	} else {
		memcpy(velocity,epoch->sat[satIndex].ITRFvel,3*sizeof(double));
	}	 
	if (options->printModel) {
		res = getMeasModelValue(epoch,epoch->sat[satIndex].GNSS,epoch->sat[satIndex].PRN,epoch->measOrder[epoch->sat[satIndex].GNSS].ind2Meas[measIndex],&measurement,&model);
		if (res==1) {
			printf("MODEL    %17s %3s %2d %3s %11.5lf %14.4lf %14.4lf %14.4lf %14.4lf %14.4lf %14.4lf %14.4lf %14.4lf %14.4lf %14.5lf %9.5lf %9.5lf %9.5lf %9.5lf %9.5lf %9.5lf %9.5lf %9.5lf %9.5lf %9.5lf %8.3lf %8.3lf\n",
					t2doystr(&epoch->t),
					gnsstype2gnssstr(epoch->sat[satIndex].GNSS),
					epoch->sat[satIndex].PRN,
					meastype2measstr(epoch->measOrder[epoch->sat[satIndex].GNSS].ind2Meas[measIndex]),
					flightTime,
					getMeasurementValue(epoch,epoch->sat[satIndex].GNSS,epoch->sat[satIndex].PRN,epoch->measOrder[epoch->sat[satIndex].GNSS].ind2Meas[measIndex]),
					getModelValue(epoch,epoch->sat[satIndex].GNSS,epoch->sat[satIndex].PRN,epoch->measOrder[epoch->sat[satIndex].GNSS].ind2Meas[measIndex]),
					epoch->sat[satIndex].position[0],
					epoch->sat[satIndex].position[1],
					epoch->sat[satIndex].position[2],
					velocity[0],
					velocity[1],
					velocity[2],
                                        //JAUME//
					geometricDistance,
                                        //GEOMETRIC RANGE BETWEEN RECEIVER AND ANTENNA PHASE CENTERS//
                                        //geometricDistance+satellitePhaseCenterProjection+receiverPhaseCenterProjection+receiverARPProjection+solidTidesModel,//
                                        //JAUME//
					satelliteClockModel,
					satellitePhaseCenterProjection,
					receiverPhaseCenterProjection,
					receiverARPProjection,
					relativityModel,
					//windUpRadModel,
					windUpMetModel,
					troposphereModel,
					ionosphereModel,
					gravitationalDelayModel,
					TGDModel,
					solidTidesModel,
                                        //JAUME
                                        epoch->sat[satIndex].elevation/d2r,
                                        epoch->sat[satIndex].azimuth/d2r);
                                        //JAUME
		}
	}
}

/*****************************************************************************
 * Name        : printSatSel
 * Description : Prints debug information of the reason why a satellite has 
 *               been discarded (or selected) for processing 
 * Parameters  :
 * Name                           |Da|Unit|Description
 * int  selected                   I  N/A  Flag to set if the satellite was 
 *                                         selected (1) or discarded (0) 
 * int  satellite                  I  N/A  PRN of the satellite
 * char  *message                  I  N/A  Reason of the discard (if selected==0)
 * TOptions  *options              I  N/A  TOptions structure
 *****************************************************************************/
void printSatSel (int selected, int satellite, char *message, TOptions *options) {
	if (options->printSatSel) {
		if (!selected) printf("SATSEL %2d discarded: %s\n",satellite,message);
		else printf("SATSEL %2d selected\n",satellite);
	}
}

/*****************************************************************************
 * Name        : printPrefit
 * Description : Prints the prefilter residuals for a specific satellite/measurement
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TTime  *t                       I  N/A  Reference time
 * TSatellite  *sat                I  N/A  Selected satellite
 * enum MeasurementType  measType  I  N/A  Selected measurement
 * double  prefit                  I    m  Computed prefit
 * double  measurement             I    m  Measurement value
 * double  model                   I    m  Modeled value
 * double  stdDev                  I    m  Assigned weight of the measurement 
 * TOptions  *options              I  N/A  TOptions structure
 *****************************************************************************/
void printPrefit (TTime *t, TSatellite *sat, enum MeasurementType measType, double prefit, double measurement, double model, double stdDev, TOptions *options) {	
	if (options->printPrefit) {
		printf("PREFIT%1s  %17s %3s %2d %3s %14.4lf %14.4lf %14.4lf %14.4lf %14.4lf %14.4lf %14.4lf %8.3lf %8.3lf",
				sat->available?"":"*",
				t2doystr(t),
				gnsstype2gnssstr(sat->GNSS),
				sat->PRN,
				meastype2measstr(measType),
				prefit,
				measurement,
				model,
				sat->LoS[0],
				sat->LoS[1],
				sat->LoS[2],
				sqrt(sat->LoS[0]*sat->LoS[0] + sat->LoS[1]*sat->LoS[1] + sat->LoS[2]*sat->LoS[2]),				
				sat->elevation/d2r,
				sat->azimuth/d2r
				);


		if (stdDev==SIGMA_INF) {
			printf(" 9999.999\n");
		} else {
			printf(" %8.3lf\n", stdDev);
		}
	}
}

/*****************************************************************************
 * Name        : printEpochsat
 * Description : Prints the satellites used for processing an epoch
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TTime  *t                       I  N/A  Reference time
 * enum MeasurementType  measType  I  N/A  Selected measurement
 * int  ns                         I  N/A  Number of satellites used
 * int  *sats                      I  N/A  Vector containing the satellites PRN
 * TOptions  *options              I  N/A  TOptions structure
 *****************************************************************************/
void printEpochsat (TTime *t, enum MeasurementType measType, int ns, int *sats, TOptions *options) {
	int i;
	
	if (options->printSatellites) {
		printf("EPOCHSAT %17s %3s %2d",t2doystr(t),meastype2measstr(measType),ns);
		for (i=0;i<ns;i++) {
			printf (" %2d",sats[i]);
		}
		printf("\n");
	}
}

/*****************************************************************************
 * Name        : printFilter
 * Description : Prints the parameters estimated from the filter
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TTime  *t                       I  N/A  Reference time
 * int  nunk                       I  N/A  Number of unknowns in the filter
 * TFilterSolution  *solution      I  N/A  Structure storing the estimated parameters
 * TOptions  *options              I  N/A  TOptions structure
 *****************************************************************************/
void printFilter (TTime *t, int nunk, TFilterSolution *solution, TOptions *options) {
	int i;

	if (options->printFilterSolution) {
		printf("FILTER   %17s",t2doystr(t));
		for (i=0;i<nunk;i++) {
			printf(" %14.4lf",solution->x[i]);
		}
		printf("\n");
	}
}

/*****************************************************************************
 * Name        : printPostfit
 * Description : Prints the postfilter residuals for a specific satellite/measurement
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TTime  *t                       I  N/A  Reference time
 * TSatellite  *sat                I  N/A  Selected satellite
 * enum MeasurementType  measType  I  N/A  Selected measurement
 * double  postfit                 I  N/A  Computed postfit
 * double  measurement             I  N/A  Measurement value
 * double  model                   I  N/A  Modeled value
 * double  estimatedParameters     I  N/A  Estimated parameters from the filter
 * double  ambiguity               I  N/A  Estimated ambiguity value (only for carrier phases)
 * TOptions  *options              I  N/A  TOptions structure
 *****************************************************************************/
void printPostfit (TTime *t, TSatellite *sat, enum MeasurementType measType, double postfit, double measurement, double model, double estimatedParameters, double ambiguity, TOptions *options) {
	if (options->printPostfit) {
		if (whatIs(measType)==CarrierPhase) {
			printf("POSTFIT%1s %17s %3s %2d %3s %14.4lf %14.4lf %14.4lf %8.3lf %8.3lf %14.4lf\n",
					sat->available?"":"*",
					t2doystr(t),
					gnsstype2gnssstr(sat->GNSS),
					sat->PRN,
					meastype2measstr(measType),
					postfit,
					measurement,
					model + estimatedParameters,
					sat->elevation/d2r,
					sat->azimuth/d2r,
					ambiguity);
		} else if (whatIs(measType)==Pseudorange) {
			printf("POSTFIT%1s %17s %3s %2d %3s %14.4lf %14.4lf %14.4lf %8.3lf %8.3lf\n",
					sat->available?"":"*",
					t2doystr(t),
					gnsstype2gnssstr(sat->GNSS),
					sat->PRN,
					meastype2measstr(measType),
					postfit,
					measurement,
					model + estimatedParameters,
					sat->elevation/d2r,
					sat->azimuth/d2r);
		}
	}
}
 
/*****************************************************************************
 * Name        : printOutput
 * Description : Prints the XYZ and NEU solutions for a station obtained by the
 *               filter, along with the dXYZ, dNEU in relation of the apriori
 *               position (if given), and formal errors of both
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  TEpoch structure
 * TFilterSolution  *solution      I  N/A  Solution of the filter
 * TOptions  *options              I  N/A  TOptions structure
 * TUnkinfo  *unkinfo              I  N/A  Information on the unknowns of the filter
 *****************************************************************************/
void printOutput (TEpoch *epoch, TFilterSolution *solution, TOptions *options, TUnkinfo *unkinfo) {
	double	positionNEU[3];
	double	dNEU[3];
	double  dXYZ[3];
	double  feNEU[3];
	int		i,j,k;
	int 	indx;
	int		indtrop;
	
	// First index of XYZ
	indx = unkinfo->par2unk[DR_UNK];
	
	// Index of troposphere
	if (options->estimateTroposphere) {
		indtrop = unkinfo->par2unk[TROP_UNK];
	}
	
	// Obtaing NEU coordinates
	XYZ2NEU(&solution->x[indx],positionNEU);
	
	// Compute:
	//		dNEU differences in relation to apriori position
	//		Formal errors in NEU system. For this:
	//					feNEU = R^-1*feXYZ*R
	//					Being, R the rotation matrix (orientation matrix)
	//						   feXYZ the formal error matrix in XYZ components (the correlation matrix)
	//					As we are only interested in the diagonal of feNEU, only these elements are computed
	for (i=0;i<3;i++) {
		dNEU[i] = 0;
		feNEU[i] = 0;
		for (j=0;j<3;j++) {
			dNEU[i] += (solution->x[indx+j] - epoch->receiver.aproxPosition[j])*epoch->receiver.orientation[i][j];
			for (k=0;k<3;k++) {
				feNEU[i] += epoch->receiver.orientation[j][i] * solution->correlations[m2v(indx+j,indx+k)] * epoch->receiver.orientation[k][i];
			}
		}
		feNEU[i] = sqrt(feNEU[i]);
		dXYZ[i] = solution->x[indx+i] - epoch->receiver.aproxPosition[i];
	}
	
	if (options->receiverPositionSource == rpCALCULATE) {
		for (i=0;i<3;i++) {
			dNEU[i] = 0;
			dXYZ[i] = 0;
		}
	}
	
	if (options->printOutput) {
		printf("OUTPUT %17s %9.4lf %14.4lf %14.4lf %14.4lf %10.4lf %10.4lf %10.4lf %10.4lf %10.4lf %10.4lf %14.9lf %14.9lf %14.4lf %10.4lf %10.4lf %10.4lf %10.4lf %10.4lf %10.4lf %9.4lf %9.4lf %9.4lf %9.4lf %9.4lf",
				t2doystr(&epoch->t),
				solution->dop,
				solution->x[indx+0],
				solution->x[indx+1],
				solution->x[indx+2],
				dXYZ[0],
				dXYZ[1],
				dXYZ[2],
				sqrt(solution->correlations[m2v(indx+0,indx+0)]),
				sqrt(solution->correlations[m2v(indx+1,indx+1)]),
				sqrt(solution->correlations[m2v(indx+2,indx+2)]),
				positionNEU[0]/d2r, // Latitude
				positionNEU[1]/d2r, // Longitude
				positionNEU[2],     // Height
				dNEU[0],			// dNorth
				dNEU[1],			// dEast
				dNEU[2],			// dUp
				feNEU[0],
				feNEU[1],
				feNEU[2],
				solution->GDOP,
				solution->PDOP,
				solution->TDOP,				
				solution->HDOP,
				solution->VDOP
				);
		if (options->estimateTroposphere) {
			printf(" %10.4lf %10.4lf %10.4lf\n",
					solution->x[indtrop],
					solution->x[indtrop] - epoch->ZTD,
					sqrt(solution->correlations[m2v(indtrop,indtrop)])
					);
		}else {
			printf("\n");
		}
	}
}

/*****************************************************************************
 * Name        : printSatDiff
 * Description : Prints the difference between two different satellites at a 
 *               specific time. Provides the difference in SISRE, SISRE 
 *               orbit-only, 3D orbit, clock and radial, along and cross-track
 *               components
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TTime  *t                       I  N/A  Reference time
 * enum GNSSystem  GNSS            I  N/A  GNSS system of the satellite
 * int  PRN                        I  N/A  PRN satellite identifier
 * double  SISREDiff               I    m  Signal-in-Space Range Error difference
 * double  OrbDiff                 I    m  SISRE orbit-only difference
 * double  Orb3DDiff               I    m  3D orbit difference
 * double  ClkDiff                 I    m  Clock difference
 * double  posDiff[3]              I    m  Radial, along- and cross-track position difference 
 * TOptions  *options              I  N/A  TOptions structure
 *****************************************************************************/
void printSatDiff (TTime *t, enum GNSSystem GNSS, int PRN, double SISREDiff, double OrbDiff, double Orb3DDiff, double ClkDiff, double posDiff[3], TOptions *options) {
	if (options->printSatDiff) {
		printf("SATDIFF %17s %3s %2d %10.4lf %10.4lf %10.4lf %10.4lf %10.4lf %10.4lf %10.4lf\n",
				t2doystr(t),
				gnsstype2gnssstr(GNSS),
				PRN,
				SISREDiff,
				OrbDiff,
				Orb3DDiff,
				ClkDiff,
				posDiff[2],
				posDiff[0],
				posDiff[1]
				);
	}
}

/*****************************************************************************
 * Name        : printSatStat
 * Description : Prints the statistics of differences between two different 
 *               satellites. Provides the difference in SISRE, SISRE orbit-only, 
 *               3D orbit, clock and radial, along and cross-track components
 * Parameters  :
 * Name                           |Da|Unit|Description
 * enum GNSSystem  GNSS            I  N/A  GNSS system of the satellite
 * int  PRN                        I  N/A  PRN satellite identifier
 * TStat  statSISRE                I  N/A  Signal-in-Space Range Error statistics
 * TStat  statOrb                  I  N/A  SISRE orbit-only statistics
 * TStat  stat3D                   I  N/A  3D orbit statistics
 * TStat  statClk                  I  N/A  Clock statistics
 * TStat  statRadOrb               I  N/A  Radial position statistics 
 * TStat  statAloOrb               I  N/A  Along-track position statistics 
 * TStat  statCroOrb               I  N/A  Cross-track position statistics 
 * TOptions  *options              I  N/A  TOptions structure
 *****************************************************************************/
void printSatStat (enum GNSSystem GNSS, int PRN, TStat statSISRE, TStat statOrb, TStat stat3D, TStat statClk, TStat statRadOrb, TStat statAloOrb, TStat statCroOrb, TOptions *options) {
	if (options->printSatStat && (statSISRE.n != 0)) {
		printf("SATSTAT %3s %2d %4d %10.4lf %10.4lf %10.4lf %10.4lf %10.4lf %10.4lf %10.4lf %10.4lf %10.4lf %10.4lf %10.4lf %10.4lf %10.4lf %10.4lf %10.4lf %10.4lf %10.4lf %10.4lf %10.4lf %10.4lf %10.4lf\n",
				gnsstype2gnssstr(GNSS),
				PRN,
				statSISRE.n,
				statSISRE.mean,
				sqrt(statSISRE.mean2),
				sqrt(statSISRE.mean2 - statSISRE.mean*statSISRE.mean),
				statOrb.mean,
				sqrt(statOrb.mean2),
				sqrt(statOrb.mean2 - statOrb.mean*statOrb.mean),
				stat3D.mean,
				sqrt(stat3D.mean2),
				sqrt(stat3D.mean2 - stat3D.mean*stat3D.mean),
				statClk.mean,
				sqrt(statClk.mean2),
				sqrt(statClk.mean2 - statClk.mean*statClk.mean),
				statRadOrb.mean,
				sqrt(statRadOrb.mean2),
				sqrt(statRadOrb.mean2 - statRadOrb.mean*statRadOrb.mean),
				statAloOrb.mean,
				sqrt(statAloOrb.mean2),
				sqrt(statAloOrb.mean2 - statAloOrb.mean*statAloOrb.mean),
				statCroOrb.mean,
				sqrt(statCroOrb.mean2),
				sqrt(statCroOrb.mean2 - statCroOrb.mean*statCroOrb.mean)
				); 
	}
}

/*****************************************************************************
 * Name        : printSatStatTot
 * Description : Prints the statistics of differences between two products. 
 *               Provides the difference in SISRE, SISRE orbit-only, 3D orbit, 
 *               clock and radial, along and cross-track components
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TStat  statSISRE                I  N/A  Signal-in-Space Range Error statistics
 * TStat  statOrb                  I  N/A  SISRE orbit-only statistics
 * TStat  stat3D                   I  N/A  3D orbit statistics
 * TStat  statClk                  I  N/A  Clock statistics
 * TStat  statRadOrb               I  N/A  Radial position statistics 
 * TStat  statAloOrb               I  N/A  Along-track position statistics 
 * TStat  statCroOrb               I  N/A  Cross-track position statistics 
 * TOptions  *options              I  N/A  TOptions structure
 *****************************************************************************/
void printSatStatTot (TStat statSISRETot, TStat statOrbTot, TStat stat3DTot, TStat statClkTot, TStat statRadOrbTot, TStat statAloOrbTot, TStat statCroOrbTot, TOptions *options) {
	if (options->printSatStatTot && statSISRETot.n != 0) {
		printf("SATSTATTOT %2d %10.4lf %10.4lf %10.4lf %10.4lf %10.4lf %10.4lf %10.4lf %10.4lf %10.4lf %10.4lf %10.4lf %10.4lf %10.4lf %10.4lf %10.4lf %10.4lf %10.4lf %10.4lf %10.4lf %10.4lf %10.4lf\n",
				statSISRETot.n,
				statSISRETot.mean,
				sqrt(statSISRETot.mean2),
				sqrt(statSISRETot.mean2 - statSISRETot.mean*statSISRETot.mean),
				statOrbTot.mean,
				sqrt(statOrbTot.mean2),
				sqrt(statOrbTot.mean2 - statOrbTot.mean*statOrbTot.mean),
				stat3DTot.mean,
				sqrt(stat3DTot.mean2),
				sqrt(stat3DTot.mean2 - stat3DTot.mean*stat3DTot.mean),
				statClkTot.mean,
				sqrt(statClkTot.mean2),
				sqrt(statClkTot.mean2 - statClkTot.mean*statClkTot.mean),
				statRadOrbTot.mean,
				sqrt(statRadOrbTot.mean2),
				sqrt(statRadOrbTot.mean2 - statRadOrbTot.mean*statRadOrbTot.mean),
				statAloOrbTot.mean,
				sqrt(statAloOrbTot.mean2),
				sqrt(statAloOrbTot.mean2 - statAloOrbTot.mean*statAloOrbTot.mean),
				statCroOrbTot.mean,
				sqrt(statCroOrbTot.mean2),
				sqrt(statCroOrbTot.mean2 - statCroOrbTot.mean*statCroOrbTot.mean)
				);
	}
}

/*****************************************************************************
 * Name        : printSatPvt
 * Description : Prints the statistics of differences between two products. 
 *               Provides the difference in SISRE, SISRE orbit-only, 3D orbit, 
 *               clock and radial, along and cross-track components
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TTime  *t                       I  N/A  Reference time
 * enum GNSSystem  GNSS            I  N/A  GNSS system of the satellite
 * int  PRN                        I  N/A  PRN satellite identifier
 * double  position[3]             I    m  XYZ position of the satellite
 * double  velocity[3]             I  m/s  VXVYVZ velocity of the satellite
 * double  clock                   I    m  Clock offset of the satellite
 * TOptions  *options              I  N/A  TOptions structure
 *****************************************************************************/
void printSatPvt (TTime *t, enum GNSSystem GNSS, int PRN, double position[3], double velocity[3], double clock, TOptions *options) {
	if (options->printSatPvt) {
		printf("SATPVT %17s %3s %2d %14.4lf %14.4lf %14.4lf %14.4lf %14.4lf %14.4lf %17.6lf\n",
				t2doystr(t),
				gnsstype2gnssstr(GNSS),
				PRN,
				position[0],
				position[1],
				position[2],
				velocity[0],
				velocity[1],
				velocity[2],
				clock
				);
	}
}
